@@ -77,6 +77,7 @@ public class RubyArray extends RubyObject implements List {
         runtime.setArray(arrayc);
         arrayc.index = ClassIndex.ARRAY;
         arrayc.kindOf = new RubyModule.KindOf() {
+            @Override
             public boolean isKindOf(IRubyObject obj, RubyModule type) {
                 return obj instanceof RubyArray;
             }
@@ -94,6 +95,7 @@ public class RubyArray extends RubyObject implements List {
         }
     };
 
+    @Override
     public int getNativeTypeIndex() {
         return ClassIndex.ARRAY;
     }
@@ -147,6 +149,10 @@ public class RubyArray extends RubyObject implements List {
         return new RubyArray(runtime, new IRubyObject[] { obj });
     }
 
+    public static RubyArray newArrayLight(Ruby runtime, IRubyObject obj) {
+        return new RubyArray(runtime, new IRubyObject[] { obj }, false);
+    }
+
     /** rb_assoc_new
      *
      */
@@ -210,6 +216,15 @@ public class RubyArray extends RubyObject implements List {
         realLength = vals.length;
     }
 
+    /* 
+     * plain internal array assignment
+     */
+    private RubyArray(Ruby runtime, IRubyObject[] vals, boolean objectSpace) {
+        super(runtime, runtime.getArray(), objectSpace);
+        values = vals;
+        realLength = vals.length;
+    }
+
     /* 
      * plain internal array assignment
      */
@@ -454,6 +469,7 @@ public class RubyArray extends RubyObject implements List {
      * 
      */
     @JRubyMethod(name = {"initialize_copy"}, required = 1, visibility=Visibility.PRIVATE)
+    @Override
     public IRubyObject initialize_copy(IRubyObject orig) {
         return this.replace(orig);
     }
@@ -483,6 +499,7 @@ public class RubyArray extends RubyObject implements List {
      *
      */
     @JRubyMethod(name = "to_s")
+    @Override
     public IRubyObject to_s() {
         if (realLength == 0) return RubyString.newEmptyString(getRuntime());
 
@@ -555,7 +572,7 @@ public class RubyArray extends RubyObject implements List {
      *
      */
     private final IRubyObject elt(long offset) {
-        if (realLength == 0 || offset < 0 || offset >= realLength) return getRuntime().getNil();
+        if (emptyOrOutsideRange(offset)) return getRuntime().getNil();
 
         return values[begin + (int) offset];
     }
@@ -564,16 +581,20 @@ public class RubyArray extends RubyObject implements List {
      *
      */
     private final IRubyObject elt(int offset) {
-        if (realLength == 0 || offset < 0 || offset >= realLength) return getRuntime().getNil();
+        if (emptyOrOutsideRange(offset)) return getRuntime().getNil();
 
         return values[begin + offset];
     }
+    
+    private boolean emptyOrOutsideRange(long offset) {
+        return realLength == 0 || offset < 0 || offset >= realLength;
+    }
 
     /** rb_ary_elt - faster
      *
      */
     private final IRubyObject elt_f(long offset) {
-        if (realLength == 0 || offset >= realLength) return getRuntime().getNil();
+        if (emptyOrTooHigh(offset)) return getRuntime().getNil();
 
         return values[begin + (int) offset];
     }
@@ -582,10 +603,14 @@ public class RubyArray extends RubyObject implements List {
      *
      */
     private final IRubyObject elt_f(int offset) {
-        if (realLength == 0 || offset >= realLength) return getRuntime().getNil();
+        if (emptyOrTooHigh(offset)) return getRuntime().getNil();
 
         return values[begin + offset];
     }
+    
+    private boolean emptyOrTooHigh(long offset) {
+        return realLength == 0 || offset >= realLength;
+    }
 
     /** rb_ary_entry
      *
@@ -1061,15 +1086,16 @@ public class RubyArray extends RubyObject implements List {
      */
     @JRubyMethod(name = "include?", required = 1)
     public RubyBoolean include_p(ThreadContext context, IRubyObject item) {
-        return getRuntime().newBoolean(includes(context, item));
+        return context.getRuntime().newBoolean(includes(context, item));
     }
 
     /** rb_ary_frozen_p
      *
      */
     @JRubyMethod(name = "frozen?")
-    public RubyBoolean frozen_p() {
-        return getRuntime().newBoolean(isFrozen() || (flags & TMPLOCK_ARR_F) != 0);
+    @Override
+    public RubyBoolean frozen_p(ThreadContext context) {
+        return context.getRuntime().newBoolean(isFrozen() || (flags & TMPLOCK_ARR_F) != 0);
     }
 
     /**
@@ -1210,6 +1236,7 @@ public class RubyArray extends RubyObject implements List {
     *
     */
     @JRubyMethod(name = "inspect")
+    @Override
     public IRubyObject inspect() {
         if (realLength == 0) return getRuntime().newString("[]");
         if (getRuntime().isInspecting(this)) return  getRuntime().newString("[...]");
@@ -1421,6 +1448,7 @@ public class RubyArray extends RubyObject implements List {
      *
      */
     @JRubyMethod(name = "to_a")
+    @Override
     public RubyArray to_a() {
         if(getMetaClass() != getRuntime().getArray()) {
             RubyArray dup = new RubyArray(getRuntime(), getRuntime().isObjectSpaceEnabled());
@@ -1441,10 +1469,12 @@ public class RubyArray extends RubyObject implements List {
     	return this;
     }
 
+    @Override
     public RubyArray convertToArray() {
         return this;
     }
     
+    @Override
     public IRubyObject checkArrayType(){
         return this;
     }
@@ -1453,6 +1483,7 @@ public class RubyArray extends RubyObject implements List {
      *
      */
     @JRubyMethod(name = "==", required = 1)
+    @Override
     public IRubyObject op_equal(ThreadContext context, IRubyObject obj) {
         if (this == obj) return getRuntime().getTrue();
 
@@ -1550,6 +1581,10 @@ public class RubyArray extends RubyObject implements List {
             isShared = true;
         } else if (values.length > ARRAY_DEFAULT_SIZE << 1){
             alloc(ARRAY_DEFAULT_SIZE << 1);
+        } else {
+            for (int i = 0; i < realLength; i++) {
+                values[begin + i] = null;
+            }
         }
 
         begin = 0;
@@ -1604,16 +1639,24 @@ public class RubyArray extends RubyObject implements List {
             // TODO: In MRI 1.9, an explicit check for negative length is
             // added here. IndexError is raised when length is negative.
             // See [ruby-core:12953] for more details.
+            //
+            // New note: This is actually under re-evaluation,
+            // see [ruby-core:17483].
             break;
         }
 
         modify();
 
-        end = beg + len;
-        // apparently, that's how MRI does overflow check
-        if (end < 0) {
+        // See [ruby-core:17483]
+        if (len < 0) {
+            return this;
+        }
+
+        if (len > Integer.MAX_VALUE - beg) {
             throw getRuntime().newArgumentError("argument too big");
         }
+
+        end = beg + len;
         if (end > realLength) {
             if (end >= values.length) realloc(end);
 
@@ -1632,7 +1675,7 @@ public class RubyArray extends RubyObject implements List {
         } else {
             if(len > 0) Arrays.fill(values, beg, beg + len, item);
         }
-        
+
         return this;
     }
 
@@ -1779,7 +1822,7 @@ public class RubyArray extends RubyObject implements List {
             if (i1 != i2) store(i2, e);
             i2++;
         }
-        
+
         if (realLength == i2) {
             if (block.isGiven()) return block.yield(context, item);
 
@@ -1789,6 +1832,9 @@ public class RubyArray extends RubyObject implements List {
         modify();
 
         if (realLength > i2) {
+            for (int i = i2; i < realLength; i++) {
+                values[begin + i] = null;
+            }
             realLength = i2;
             if (i2 << 1 < values.length && values.length > ARRAY_DEFAULT_SIZE) realloc(i2 << 1);
         }
@@ -1813,6 +1859,7 @@ public class RubyArray extends RubyObject implements List {
         System.arraycopy(values, pos + 1, values, pos, len - (pos + 1));
 
         realLength--;
+        values[realLength] = null;
 
         return obj;
     }
@@ -1852,7 +1899,12 @@ public class RubyArray extends RubyObject implements List {
         }
         if (realLength == i2) return getRuntime().getNil();
 
-        if (i2 < realLength) realLength = i2;
+        if (i2 < realLength) {
+            for (int i = i2; i < realLength; i++) {
+                values[i] = null;
+            }
+            realLength = i2;
+        }
 
         return this;
     }
@@ -2289,22 +2341,31 @@ public class RubyArray extends RubyObject implements List {
     static final class DefaultComparator implements Comparator {
         public int compare(Object o1, Object o2) {
             if (o1 instanceof RubyFixnum && o2 instanceof RubyFixnum) {
-                long a = ((RubyFixnum) o1).getLongValue();
-                long b = ((RubyFixnum) o2).getLongValue();
-                if (a > b) return 1;
-                if (a < b) return -1;
-                return 0;
+                return compareFixnums(o1, o2);
             }
             if (o1 instanceof RubyString && o2 instanceof RubyString) {
                 return ((RubyString) o1).op_cmp((RubyString) o2);
             }
+            //TODO: ary_sort_check should be done here
+            return compareOthers((IRubyObject)o1, (IRubyObject)o2);
+        }
 
-            IRubyObject obj1 = (IRubyObject) o1;
-            IRubyObject obj2 = (IRubyObject) o2;
+        private int compareFixnums(Object o1, Object o2) {
+            long a = ((RubyFixnum) o1).getLongValue();
+            long b = ((RubyFixnum) o2).getLongValue();
+            if (a > b) {
+                return 1;
+            }
+            if (a < b) {
+                return -1;
+            }
+            return 0;
+        }
 
-            ThreadContext context = obj1.getRuntime().getCurrentContext();
-            IRubyObject ret = obj1.callMethod(context, MethodIndex.OP_SPACESHIP, "<=>", obj2);
-            int n = RubyComparable.cmpint(context, ret, obj1, obj2);
+        private int compareOthers(IRubyObject o1, IRubyObject o2) {
+            ThreadContext context = o1.getRuntime().getCurrentContext();
+            IRubyObject ret = o1.callMethod(context, MethodIndex.OP_SPACESHIP, "<=>", o2);
+            int n = RubyComparable.cmpint(context, ret, o1, o2);
             //TODO: ary_sort_check should be done here
             return n;
         }
@@ -2337,6 +2398,7 @@ public class RubyArray extends RubyObject implements List {
         return Pack.pack(getRuntime(), this, iFmt.getByteList());
     }
 
+    @Override
     public Class getJavaClass() {
         return List.class;
     }
