@@ -62,6 +62,7 @@ public class RubyClass extends RubyModule {
     public static void createClassClass(Ruby runtime, RubyClass classClass) {
         classClass.index = ClassIndex.CLASS;
         classClass.kindOf = new RubyModule.KindOf() {
+            @Override
             public boolean isKindOf(IRubyObject obj, RubyModule type) {
                 return obj instanceof RubyClass;
             }
@@ -104,18 +105,22 @@ public class RubyClass extends RubyModule {
         return obj;
     }    
 
+    @Override
     public int getNativeTypeIndex() {
         return ClassIndex.CLASS;
     }
     
+    @Override
     public boolean isModule() {
         return false;
     }
 
+    @Override
     public boolean isClass() {
         return true;
     }
 
+    @Override
     public boolean isSingleton() {
         return false;
     }
@@ -203,6 +208,7 @@ public class RubyClass extends RubyModule {
     /** rb_make_metaclass
      *
      */
+    @Override
     public RubyClass makeMetaClass(RubyClass superClass) {
         if (isSingleton()) { // could be pulled down to RubyClass in future
             MetaClass klass = new MetaClass(getRuntime(), superClass); // rb_class_boot
@@ -227,8 +233,7 @@ public class RubyClass extends RubyModule {
     public IRubyObject invoke(ThreadContext context, IRubyObject self, String name,
             IRubyObject[] args, CallType callType, Block block) {
         assert args != null;
-        DynamicMethod method = null;
-        method = searchMethod(name);
+        DynamicMethod method = searchMethod(name);
         
 
         if (method.isUndefined() || (!name.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
@@ -240,28 +245,25 @@ public class RubyClass extends RubyModule {
     
     public IRubyObject invoke(ThreadContext context, IRubyObject self, String name,
             IRubyObject arg, CallType callType, Block block) {
-        DynamicMethod method = null;
-        method = searchMethod(name);
-        
+        DynamicMethod method = searchMethod(name);
 
         if (method.isUndefined() || (!name.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-            return RuntimeHelpers.callMethodMissing(context, self, method, name, new IRubyObject[] {arg}, context.getFrameSelf(), callType, block);
+            return RuntimeHelpers.callMethodMissing(context, self, method, name, 
+                    new IRubyObject[] {arg}, context.getFrameSelf(), callType, block);
         }
 
         return method.call(context, self, this, name, arg, block);
     }
     
-    public IRubyObject invokeInherited(ThreadContext context, IRubyObject self,
-            IRubyObject subclass) {
-        String name = "inherited";
-        DynamicMethod method = getMetaClass().searchMethod(name);
-        IRubyObject[] args = new IRubyObject[] {subclass};
+    public IRubyObject invokeInherited(ThreadContext context, IRubyObject self, IRubyObject subclass) {
+        DynamicMethod method = getMetaClass().searchMethod("inherited");
 
         if (method.isUndefined()) {
-            return RuntimeHelpers.callMethodMissing(context, self, method, name, args, context.getFrameSelf(), CallType.FUNCTIONAL, Block.NULL_BLOCK);
+            return RuntimeHelpers.callMethodMissing(context, self, method, "inherited", 
+                    new IRubyObject[] {subclass}, context.getFrameSelf(), CallType.FUNCTIONAL, Block.NULL_BLOCK);
         }
 
-        return method.call(context, self, getMetaClass(), name, args, Block.NULL_BLOCK);
+        return method.call(context, self, getMetaClass(), "inherited", subclass, Block.NULL_BLOCK);
     }
 
     /** rb_class_new_instance
@@ -280,7 +282,6 @@ public class RubyClass extends RubyModule {
     @JRubyMethod(name = "initialize", optional = 1, frame = true, visibility = Visibility.PRIVATE)
     public IRubyObject initialize(IRubyObject[] args, Block block) {
         if (superClass != null) {
-            System.out.println(classId);
             throw getRuntime().newTypeError("already initialized class");
         }
  
@@ -313,6 +314,7 @@ public class RubyClass extends RubyModule {
      * 
      */
     @JRubyMethod(name = "initialize_copy", required = 1)
+    @Override
     public IRubyObject initialize_copy(IRubyObject original){
         if (superClass != null) throw runtime.newTypeError("already initialized class");
         if (original instanceof MetaClass) throw getRuntime().newTypeError("can't copy singleton class");        
@@ -324,17 +326,17 @@ public class RubyClass extends RubyModule {
     
     // TODO: Someday, enable.
     // @JRubyMethod(name = "subclasses", optional = 1)
-    public IRubyObject subclasses(IRubyObject[] args) {
+    public IRubyObject subclasses(ThreadContext context, IRubyObject[] args) {
         boolean recursive = false;
         if (args.length == 1) {
             if (args[0] instanceof RubyBoolean) {
                 recursive = args[0].isTrue();
             } else {
-                getRuntime().newTypeError(args[0], getRuntime().fastGetClass("Boolean"));
+                context.getRuntime().newTypeError(args[0], context.getRuntime().fastGetClass("Boolean"));
             }
         }
         
-        return RubyArray.newArray(getRuntime(), subclasses(recursive)).freeze();
+        return RubyArray.newArray(context.getRuntime(), subclasses(recursive)).freeze(context);
     }
     
     public Collection subclasses(boolean includeDescendants) {
@@ -366,8 +368,8 @@ public class RubyClass extends RubyModule {
     }    
 
     @JRubyMethod(name = "inherited", required = 1)
-    public IRubyObject inherited(IRubyObject arg) {
-        return getRuntime().getNil();
+    public IRubyObject inherited(ThreadContext context, IRubyObject arg) {
+        return context.getRuntime().getNil();
     }
 
     /** rb_class_inherited (reversed semantics!)
@@ -375,9 +377,8 @@ public class RubyClass extends RubyModule {
      */
     public void inherit(RubyClass superClazz) {
         if (superClazz == null) superClazz = getRuntime().getObject();
-        superClazz.invokeInherited(
-                getRuntime().getCurrentContext(), superClazz,
-                this);
+        
+        superClazz.invokeInherited(getRuntime().getCurrentContext(), superClazz, this);
     }
 
     /** Return the real super class of this class.
@@ -386,15 +387,15 @@ public class RubyClass extends RubyModule {
      *
      */
     @JRubyMethod(name = "superclass")
-    public IRubyObject superclass() {
+    public IRubyObject superclass(ThreadContext context) {
         RubyClass superClazz = superClass;
 
-        if (superClazz == null) throw runtime.newTypeError("uninitialized class");
+        if (superClazz == null) throw context.getRuntime().newTypeError("uninitialized class");
         
-        if(isSingleton()) superClazz = metaClass;
+        if (isSingleton()) superClazz = metaClass;
         while (superClazz != null && superClazz.isIncluded()) superClazz = superClazz.superClass;
 
-        return superClazz != null ? superClazz : getRuntime().getNil();
+        return superClazz != null ? superClazz : context.getRuntime().getNil();
     }
 
     /** rb_check_inheritable
