@@ -63,6 +63,7 @@ public class RubyFixnum extends RubyInteger {
         runtime.setFixnum(fixnum);
         fixnum.index = ClassIndex.FIXNUM;
         fixnum.kindOf = new RubyModule.KindOf() {
+                @Override
                 public boolean isKindOf(IRubyObject obj, RubyModule type) {
                     return obj instanceof RubyFixnum;
                 }
@@ -108,6 +109,7 @@ public class RubyFixnum extends RubyInteger {
         this.value = value;
     }
     
+    @Override
     public int getNativeTypeIndex() {
         return ClassIndex.FIXNUM;
     }
@@ -115,26 +117,32 @@ public class RubyFixnum extends RubyInteger {
     /** 
      * short circuit for Fixnum key comparison
      */
+    @Override
     public final boolean eql(IRubyObject other) {
         return other instanceof RubyFixnum && value == ((RubyFixnum)other).value;
     }
     
+    @Override
     public boolean isImmediate() {
     	return true;
     }
     
+    @Override
     public RubyClass getSingletonClass() {
         throw getRuntime().newTypeError("can't define singleton");
     }
 
+    @Override
     public Class<?> getJavaClass() {
         return Long.TYPE;
     }
 
+    @Override
     public double getDoubleValue() {
         return value;
     }
 
+    @Override
     public long getLongValue() {
         return value;
     }
@@ -142,11 +150,15 @@ public class RubyFixnum extends RubyInteger {
     private static final int CACHE_OFFSET = 128;
     
     public static RubyFixnum newFixnum(Ruby runtime, long value) {
-        if (value <= 127 && value >= -128) {
+        if (isInCacheRange(value)) {
             return runtime.fixnumCache[(int) value + CACHE_OFFSET];
         }
         return new RubyFixnum(runtime, value);
     }
+    
+    private static boolean isInCacheRange(long value) {
+        return value <= 127 && value >= -128;
+    }
 
     public RubyFixnum newFixnum(long newValue) {
         return newFixnum(getRuntime(), newValue);
@@ -180,14 +192,17 @@ public class RubyFixnum extends RubyInteger {
         return runtime.fixnumCache[CACHE_OFFSET - 1];
     }
 
+    @Override
     public RubyFixnum hash() {
         return newFixnum(hashCode());
     }
 
+    @Override
     public final int hashCode() {
         return (int)(value ^ value >>> 32);
     }
 
+    @Override
     public boolean equals(Object other) {
         if (other == this) {
             return true;
@@ -212,14 +227,29 @@ public class RubyFixnum extends RubyInteger {
     /** fix_to_s
      * 
      */
-    @JRubyMethod(optional = 1)
     public RubyString to_s(IRubyObject[] args) {
-        int base = args.length == 0 ? 10 : num2int(args[0]);
+        switch (args.length) {
+        case 0: return to_s();
+        case 1: return to_s(args[0]);
+        default: throw getRuntime().newArgumentError(args.length, 1);
+        }
+    }
+    
+    @JRubyMethod
+    @Override
+    public RubyString to_s() {
+        int base = 10;
+        return getRuntime().newString(Convert.longToByteList(value, base));
+    }
+    
+    @JRubyMethod
+    public RubyString to_s(IRubyObject arg0) {
+        int base = num2int(arg0);
         if (base < 2 || base > 36) {
             throw getRuntime().newArgumentError("illegal radix " + base);
-            }
-        return getRuntime().newString(Convert.longToByteList(value, base));
         }
+        return getRuntime().newString(Convert.longToByteList(value, base));
+    }
 
     /** fix_id2name
      * 
@@ -260,18 +290,38 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "+")
     public IRubyObject op_plus(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            long otherValue = ((RubyFixnum) other).value;
-            long result = value + otherValue;
-            if ((~(value ^ otherValue) & (value ^ result) & SIGN_BIT) != 0) {
-                return RubyBignum.newBignum(getRuntime(), value).op_plus(context, other);
-            }
-		return newFixnum(result);
+            return addFixnum(context, (RubyFixnum)other);
+        }
+        return addOther(context, other);
     }
+    
+    private IRubyObject addFixnum(ThreadContext context, RubyFixnum other) {
+        long otherValue = other.value;
+        long result = value + otherValue;
+        if (additionOverflowed(value, otherValue, result)) {
+            return addAsBignum(context, other);
+        }
+        return newFixnum(context.getRuntime(), result);
+    }
+    
+    private static boolean additionOverflowed(long original, long other, long result) {
+        return (~(original ^ other) & (original ^ result) & SIGN_BIT) != 0;
+    }
+    
+    private static boolean subtractionOverflowed(long original, long other, long result) {
+        return (~(original ^ ~other) & (original ^ result) & SIGN_BIT) != 0;
+    }
+    
+    private IRubyObject addAsBignum(ThreadContext context, RubyFixnum other) {
+        return RubyBignum.newBignum(context.getRuntime(), value).op_plus(context, other);
+    }
+    
+    private IRubyObject addOther(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyBignum) {
             return ((RubyBignum) other).op_plus(context, this);
-    }
+        }
         if (other instanceof RubyFloat) {
-            return getRuntime().newFloat((double) value + ((RubyFloat) other).getDoubleValue());
+            return context.getRuntime().newFloat((double) value + ((RubyFloat) other).getDoubleValue());
         }
         return coerceBin(context, "+", other);
     }
@@ -282,16 +332,29 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "-")
     public IRubyObject op_minus(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            long otherValue = ((RubyFixnum) other).value;
-            long result = value - otherValue;
-            if ((~(value ^ ~otherValue) & (value ^ result) & SIGN_BIT) != 0) {
-                return RubyBignum.newBignum(getRuntime(), value).op_minus(context, other);
+            return subtractFixnum(context, (RubyFixnum)other);
+        }
+        return subtractOther(context, other);
     }
-            return newFixnum(result);
-        } else if (other instanceof RubyBignum) {
-            return RubyBignum.newBignum(getRuntime(), value).op_minus(context, other);
+    
+    private IRubyObject subtractFixnum(ThreadContext context, RubyFixnum other) {
+        long otherValue = other.value;
+        long result = value - otherValue;
+        if (subtractionOverflowed(value, otherValue, result)) {
+            return subtractAsBignum(context, other);
+        }
+        return newFixnum(context.getRuntime(), result);
+    }
+    
+    private IRubyObject subtractAsBignum(ThreadContext context, RubyFixnum other) {
+        return RubyBignum.newBignum(context.getRuntime(), value).op_minus(context, other);
+    }
+    
+    private IRubyObject subtractOther(ThreadContext context, IRubyObject other) {
+        if (other instanceof RubyBignum) {
+            return RubyBignum.newBignum(context.getRuntime(), value).op_minus(context, other);
         } else if (other instanceof RubyFloat) {
-            return getRuntime().newFloat((double) value - ((RubyFloat) other).getDoubleValue());
+            return context.getRuntime().newFloat((double) value - ((RubyFloat) other).getDoubleValue());
         }
         return coerceBin(context, "-", other);
     }
@@ -304,18 +367,18 @@ public class RubyFixnum extends RubyInteger {
         if (other instanceof RubyFixnum) {
             long otherValue = ((RubyFixnum) other).value;
             if (value == 0) {
-                return RubyFixnum.zero(getRuntime());
+                return RubyFixnum.zero(context.getRuntime());
             }
             long result = value * otherValue;
-            IRubyObject r = newFixnum(getRuntime(),result);
+            IRubyObject r = newFixnum(context.getRuntime(),result);
             if(RubyNumeric.fix2long(r) != result || result/value != otherValue) {
-                return (RubyNumeric) RubyBignum.newBignum(getRuntime(), value).op_mul(context, other);
+                return (RubyNumeric) RubyBignum.newBignum(context.getRuntime(), value).op_mul(context, other);
             }
             return r;
         } else if (other instanceof RubyBignum) {
             return ((RubyBignum) other).op_mul(context, this);
         } else if (other instanceof RubyFloat) {
-            return getRuntime().newFloat((double) value * ((RubyFloat) other).getDoubleValue());
+            return context.getRuntime().newFloat((double) value * ((RubyFloat) other).getDoubleValue());
         }
         return coerceBin(context, "*", other);
     }
@@ -367,7 +430,7 @@ public class RubyFixnum extends RubyInteger {
             long y = ((RubyFixnum) other).value;
             
             if (y == 0) {
-                throw getRuntime().newZeroDivisionError();
+                throw context.getRuntime().newZeroDivisionError();
             }
             
             long div = x / y;
@@ -377,7 +440,7 @@ public class RubyFixnum extends RubyInteger {
                 div -= 1;
             }
 
-            return getRuntime().newFixnum(div);
+            return context.getRuntime().newFixnum(div);
         } 
         return coerceBin(context, method, other);
     }
@@ -393,17 +456,17 @@ public class RubyFixnum extends RubyInteger {
             long y = ((RubyFixnum) other).value;
 
             if (y == 0) {
-            	throw getRuntime().newZeroDivisionError();
-    }
+                throw context.getRuntime().newZeroDivisionError();
+            }
 
             long mod = x % y;
 
             if (mod < 0 && y > 0 || mod > 0 && y < 0) {
                 mod += y;
             }
-                
-            return getRuntime().newFixnum(mod);
-	            }
+
+            return context.getRuntime().newFixnum(mod);
+        }
         return coerceBin(context, "%", other);
     }
                 
@@ -411,15 +474,16 @@ public class RubyFixnum extends RubyInteger {
      * 
      */
     @JRubyMethod
+    @Override
     public IRubyObject divmod(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
             long x = value;
             long y = ((RubyFixnum) other).value;
-            final Ruby runtime = getRuntime();
+            final Ruby runtime = context.getRuntime();
 
             if (y == 0) {
                 throw runtime.newZeroDivisionError();
-                }
+            }
 
             long div = x / y;
             long mod = x % y;
@@ -429,12 +493,12 @@ public class RubyFixnum extends RubyInteger {
                 mod += y;
             }
 
-            IRubyObject fixDiv = RubyFixnum.newFixnum(getRuntime(), div);
-            IRubyObject fixMod = RubyFixnum.newFixnum(getRuntime(), mod);
+            IRubyObject fixDiv = RubyFixnum.newFixnum(runtime, div);
+            IRubyObject fixMod = RubyFixnum.newFixnum(runtime, mod);
 
             return RubyArray.newArray(runtime, fixDiv, fixMod);
 
-    	}
+        }
         return coerceBin(context, "divmod", other);
     }
     	
@@ -442,13 +506,13 @@ public class RubyFixnum extends RubyInteger {
      * 
      */
     @JRubyMethod
+    @Override
     public IRubyObject quo(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return RubyFloat.newFloat(getRuntime(), (double) value
-                    / (double) ((RubyFixnum) other).value);
-    }
+            return RubyFloat.newFloat(context.getRuntime(), (double) value / (double) ((RubyFixnum) other).value);
+        }
         return coerceBin(context, "quo", other);
-	            }
+    }
 
     /** fix_pow 
      * 
@@ -458,17 +522,17 @@ public class RubyFixnum extends RubyInteger {
         if(other instanceof RubyFixnum) {
             long b = ((RubyFixnum) other).value;
             if (b == 0) {
-                return RubyFixnum.one(getRuntime());
+                return RubyFixnum.one(context.getRuntime());
             }
             if (b == 1) {
                 return this;
             }
             if (b > 0) {
-                return RubyBignum.newBignum(getRuntime(), value).op_pow(context, other);
+                return RubyBignum.newBignum(context.getRuntime(), value).op_pow(context, other);
             }
-            return RubyFloat.newFloat(getRuntime(), Math.pow(value, b));
+            return RubyFloat.newFloat(context.getRuntime(), Math.pow(value, b));
         } else if (other instanceof RubyFloat) {
-            return RubyFloat.newFloat(getRuntime(), Math.pow(value, ((RubyFloat) other)
+            return RubyFloat.newFloat(context.getRuntime(), Math.pow(value, ((RubyFloat) other)
                     .getDoubleValue()));
         }
         return coerceBin(context, "**", other);
@@ -494,9 +558,10 @@ public class RubyFixnum extends RubyInteger {
      * 
      */
     @JRubyMethod(name = "==")
+    @Override
     public IRubyObject op_equal(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return RubyBoolean.newBoolean(getRuntime(), value == ((RubyFixnum) other).value);
+            return RubyBoolean.newBoolean(context.getRuntime(), value == ((RubyFixnum) other).value);
         }
         return super.op_num_equal(context, other);
     }
@@ -507,17 +572,21 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "<=>")
     public IRubyObject op_cmp(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            long otherValue = ((RubyFixnum) other).value;
-            if (value == otherValue) {
-                return RubyFixnum.zero(getRuntime());
-    }
-            if (value > otherValue) {
-                return RubyFixnum.one(getRuntime());
-            }
-            return RubyFixnum.minus_one(getRuntime());
+            return compareFixnum(context, (RubyFixnum)other);
         }
         return coerceCmp(context, "<=>", other);
     }
+    
+    private IRubyObject compareFixnum(ThreadContext context, RubyFixnum other) {
+        long otherValue = ((RubyFixnum) other).value;
+        if (value == otherValue) {
+            return RubyFixnum.zero(context.getRuntime());
+        }
+        if (value > otherValue) {
+            return RubyFixnum.one(context.getRuntime());
+        }
+        return RubyFixnum.minus_one(context.getRuntime());
+    }
 
     /** fix_gt
      * 
@@ -525,7 +594,7 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = ">")
     public IRubyObject op_gt(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return RubyBoolean.newBoolean(getRuntime(), value > ((RubyFixnum) other).value);
+            return RubyBoolean.newBoolean(context.getRuntime(), value > ((RubyFixnum) other).value);
     }
         return coerceRelOp(context, ">", other);
     }
@@ -536,7 +605,7 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = ">=")
     public IRubyObject op_ge(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return RubyBoolean.newBoolean(getRuntime(), value >= ((RubyFixnum) other).value);
+            return RubyBoolean.newBoolean(context.getRuntime(), value >= ((RubyFixnum) other).value);
             }
         return coerceRelOp(context, ">=", other);
     }
@@ -547,8 +616,9 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "<")
     public IRubyObject op_lt(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return RubyBoolean.newBoolean(getRuntime(), value < ((RubyFixnum) other).value);
+            return RubyBoolean.newBoolean(context.getRuntime(), value < ((RubyFixnum) other).value);
         }
+        
         return coerceRelOp(context, "<", other);
     }
         
@@ -558,8 +628,9 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "<=")
     public IRubyObject op_le(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return RubyBoolean.newBoolean(getRuntime(), value <= ((RubyFixnum) other).value);
-    }
+            return RubyBoolean.newBoolean(context.getRuntime(), value <= ((RubyFixnum) other).value);
+        }
+        
         return coerceRelOp(context, "<=", other);
     }
 
@@ -569,7 +640,7 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "~")
     public IRubyObject op_neg() {
         return newFixnum(~value);
-    	}
+    }
     	
     /** fix_and
      * 
@@ -577,7 +648,7 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "&")
     public IRubyObject op_and(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum || (other = fixCoerce(other)) instanceof RubyFixnum) {
-            return newFixnum(value & ((RubyFixnum) other).value);
+            return newFixnum(context.getRuntime(), value & ((RubyFixnum) other).value);
         }
         return ((RubyBignum) other).op_and(context, this);
     }
@@ -588,7 +659,7 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "|")
     public IRubyObject op_or(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum || (other = fixCoerce(other)) instanceof RubyFixnum) {
-            return newFixnum(value | ((RubyFixnum) other).value);
+            return newFixnum(context.getRuntime(), value | ((RubyFixnum) other).value);
         }
         return ((RubyBignum) other).op_or(context, this);
     }
@@ -599,7 +670,7 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod(name = "^")
     public IRubyObject op_xor(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum || (other = fixCoerce(other)) instanceof RubyFixnum) {
-            return newFixnum(value ^ ((RubyFixnum) other).value);
+            return newFixnum(context.getRuntime(), value ^ ((RubyFixnum) other).value);
         }
         return ((RubyBignum) other).op_xor(context, this); 
     }
@@ -682,7 +753,7 @@ public class RubyFixnum extends RubyInteger {
     @JRubyMethod
     public IRubyObject size() {
         return newFixnum((long) ((BIT_SIZE + 7) / 8));
-        }
+    }
 
     /** fix_zero_p 
      * 
@@ -693,21 +764,27 @@ public class RubyFixnum extends RubyInteger {
     }
 
     @JRubyMethod
+    @Override
     public IRubyObject id() {
-        if (value <= Long.MAX_VALUE / 2 && value >= Long.MIN_VALUE / 2)
+        if (value <= Long.MAX_VALUE / 2 && value >= Long.MIN_VALUE / 2) {
             return newFixnum(2 * value + 1);
+        }
+        
         return super.id();
     }
 
-    public IRubyObject taint() {
+    @Override
+    public IRubyObject taint(ThreadContext context) {
         return this;
     }
 
-    public IRubyObject freeze() {
+    @Override
+    public IRubyObject freeze(ThreadContext context) {
         return this;
     }
     
     // Piece of mri rb_to_id
+    @Override
     public String asJavaString() {
         getRuntime().getWarnings().warn(ID.FIXNUMS_NOT_SYMBOLS, "do not use Fixnums as Symbols");
         
