@@ -47,6 +47,7 @@ import org.jruby.internal.runtime.ValueAccessor;
 import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.runtime.Constants;
 import org.jruby.runtime.GlobalVariable;
+import org.jruby.runtime.IAccessor;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ReadonlyGlobalVariable;
 import org.jruby.runtime.ThreadContext;
@@ -77,14 +78,12 @@ public class RubyGlobal {
             return hash;
         }
 
+        @Override
         public IRubyObject op_aref(ThreadContext context, IRubyObject key) {
-            if (!key.respondsTo("to_str")) {
-                throw getRuntime().newTypeError("can't convert " + key.getMetaClass() + " into String");
-            }
-
-            return super.op_aref(context, RuntimeHelpers.invoke(context, key, MethodIndex.TO_STR, "to_str", IRubyObject.NULL_ARRAY));
+            return super.op_aref(context, key.convertToString());
         }
 
+        @Override
         public IRubyObject op_aset(ThreadContext context, IRubyObject key, IRubyObject value) {
             if (!key.respondsTo("to_str")) {
                 throw getRuntime().newTypeError("can't convert " + key.getMetaClass() + " into String");
@@ -103,12 +102,13 @@ public class RubyGlobal {
         }
         
         @JRubyMethod
+        @Override
         public IRubyObject to_s(){
             return getRuntime().newString("ENV");
         }
     }
     
-    public static void createGlobals(Ruby runtime) {
+    public static void createGlobals(ThreadContext context, Ruby runtime) {
         runtime.defineGlobalConstant("TOPLEVEL_BINDING", runtime.newBinding());
         
         runtime.defineGlobalConstant("TRUE", runtime.getTrue());
@@ -119,19 +119,24 @@ public class RubyGlobal {
         RubyArray argvArray = runtime.newArray();
         String[] argv = runtime.getInstanceConfig().getArgv();
         for (int i = 0; i < argv.length; i++) {
-            argvArray.add(runtime.newString(argv[i]));
+            argvArray.append(RubyString.newString(runtime, argv[i].getBytes()));
         }
         runtime.defineGlobalConstant("ARGV", argvArray);
         runtime.getGlobalVariables().defineReadonly("$*", new ValueAccessor(argvArray));
-        
+
+        IAccessor d = new ValueAccessor(runtime.newString(
+                runtime.getInstanceConfig().displayedFileName()));
+        runtime.getGlobalVariables().define("$PROGRAM_NAME", d);
+        runtime.getGlobalVariables().define("$0", d);
+
         // Version information:
-        IRubyObject version = runtime.newString(Constants.RUBY_VERSION).freeze();
-        IRubyObject release = runtime.newString(Constants.COMPILE_DATE).freeze();
-        IRubyObject platform = runtime.newString(Constants.PLATFORM).freeze();
-        IRubyObject engine = runtime.newString(Constants.ENGINE).freeze();
+        IRubyObject version = runtime.newString(Constants.RUBY_VERSION).freeze(context);
+        IRubyObject release = runtime.newString(Constants.COMPILE_DATE).freeze(context);
+        IRubyObject platform = runtime.newString(Constants.PLATFORM).freeze(context);
+        IRubyObject engine = runtime.newString(Constants.ENGINE).freeze(context);
 
         runtime.defineGlobalConstant("RUBY_VERSION", version);
-        runtime.defineGlobalConstant("RUBY_PATCHLEVEL", runtime.newString(Constants.RUBY_PATCHLEVEL).freeze());
+        runtime.defineGlobalConstant("RUBY_PATCHLEVEL", runtime.newString(Constants.RUBY_PATCHLEVEL).freeze(context));
         runtime.defineGlobalConstant("RUBY_RELEASE_DATE", release);
         runtime.defineGlobalConstant("RUBY_PLATFORM", platform);
         runtime.defineGlobalConstant("RUBY_ENGINE", engine);
@@ -140,13 +145,13 @@ public class RubyGlobal {
         runtime.defineGlobalConstant("RELEASE_DATE", release);
         runtime.defineGlobalConstant("PLATFORM", platform);
         
-        IRubyObject jrubyVersion = runtime.newString(Constants.VERSION).freeze();
+        IRubyObject jrubyVersion = runtime.newString(Constants.VERSION).freeze(context);
         runtime.defineGlobalConstant("JRUBY_VERSION", jrubyVersion);
 		
         GlobalVariable kcodeGV = new KCodeGlobalVariable(runtime, "$KCODE", runtime.newString("NONE"));
         runtime.defineVariable(kcodeGV);
         runtime.defineVariable(new GlobalVariable.Copy(runtime, "$-K", kcodeGV));
-        IRubyObject defaultRS = runtime.newString(runtime.getInstanceConfig().getRecordSeparator()).freeze();
+        IRubyObject defaultRS = runtime.newString(runtime.getInstanceConfig().getRecordSeparator()).freeze(context);
         GlobalVariable rs = new StringGlobalVariable(runtime, "$/", defaultRS);
         runtime.defineVariable(rs);
         runtime.setRecordSeparatorVar(rs);
@@ -193,10 +198,11 @@ public class RubyGlobal {
         runtime.defineVariable(new InputGlobalVariable(runtime, "$stdin", stdin));
 
         runtime.defineVariable(new OutputGlobalVariable(runtime, "$stdout", stdout));
+        runtime.getGlobalVariables().alias("$>", "$stdout");
+        runtime.getGlobalVariables().alias("$defout", "$stdout");
+
         runtime.defineVariable(new OutputGlobalVariable(runtime, "$stderr", stderr));
-        runtime.defineVariable(new OutputGlobalVariable(runtime, "$>", stdout));
-        runtime.defineVariable(new OutputGlobalVariable(runtime, "$defout", stdout));
-        runtime.defineVariable(new OutputGlobalVariable(runtime, "$deferr", stderr));
+        runtime.getGlobalVariables().alias("$deferr", "$stderr");
 
         runtime.defineGlobalConstant("STDIN", stdin);
         runtime.defineGlobalConstant("STDOUT", stdout);
@@ -274,11 +280,13 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             runtime.getWarnings().warn(ID.INEFFECTIVE_GLOBAL, "warning: variable " + name + " is no longer effective; ignored", name);
             return value;
         }
 
+        @Override
         public IRubyObject get() {
             runtime.getWarnings().warn(ID.INEFFECTIVE_GLOBAL, "warning: variable " + name + " is no longer effective", name);
             return runtime.getFalse();
@@ -290,11 +298,13 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
             IRubyObject lastExitStatus = runtime.getCurrentContext().getLastExitStatus();
             return lastExitStatus == null ? runtime.getNil() : lastExitStatus;
         }
         
+        @Override
         public IRubyObject set(IRubyObject lastExitStatus) {
             runtime.getCurrentContext().setLastExitStatus(lastExitStatus);
             
@@ -307,6 +317,7 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
             return RubyRegexp.last_match(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
@@ -317,6 +328,7 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
             return RubyRegexp.match_pre(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
@@ -327,6 +339,7 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
             return RubyRegexp.match_post(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
@@ -337,6 +350,7 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
             return RubyRegexp.match_last(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
@@ -347,10 +361,12 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
             return RuntimeHelpers.getBackref(runtime, runtime.getCurrentContext());
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             RuntimeHelpers.setBackref(runtime, runtime.getCurrentContext(), value);
             return value;
@@ -364,6 +380,7 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             RubyArgsFile.setCurrentLineNumber(runtime.getGlobalVariables().get("$<"),RubyNumeric.fix2int(value));
             return super.set(value);
@@ -376,6 +393,7 @@ public class RubyGlobal {
             set(value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (!value.isNil() && ! runtime.getException().isInstance(value)) {
                 throw runtime.newTypeError("assigning non-exception to $!");
@@ -384,6 +402,7 @@ public class RubyGlobal {
             return runtime.getCurrentContext().setErrorInfo(value);
         }
 
+        @Override
         public IRubyObject get() {
             return runtime.getCurrentContext().getErrorInfo();
         }
@@ -395,6 +414,7 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (!value.isNil() && ! (value instanceof RubyString)) {
                 throw runtime.newTypeError("value of " + name() + " must be a String");
@@ -408,10 +428,12 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject get() {
             return runtime.getKCode().kcode(runtime);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             runtime.setKCode(KCode.create(runtime, value.convertToString().toString()));
             return value;
@@ -423,10 +445,12 @@ public class RubyGlobal {
             super(runtime, name, null);
         }
 
+        @Override
         public IRubyObject get() {
             return runtime.newFixnum(runtime.getSafeLevel());
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
 //            int level = RubyNumeric.fix2int(value);
 //            if (level < runtime.getSafeLevel()) {
@@ -446,10 +470,12 @@ public class RubyGlobal {
             set(initialValue);
         }
         
+        @Override
         public IRubyObject get() {
             return runtime.getVerbose();
         }
 
+        @Override
         public IRubyObject set(IRubyObject newValue) {
             if (newValue.isNil()) {
                 runtime.setVerbose(newValue);
@@ -467,10 +493,12 @@ public class RubyGlobal {
             set(initialValue);
         }
 
+        @Override
         public IRubyObject get() {
             return runtime.getDebug();
         }
 
+        @Override
         public IRubyObject set(IRubyObject newValue) {
             if (newValue.isNil()) {
                 runtime.setDebug(newValue);
@@ -487,6 +515,7 @@ public class RubyGlobal {
             super(runtime, name, null);
         }
 
+        @Override
         public IRubyObject get() {
             IRubyObject errorInfo = runtime.getGlobalVariables().get("$!");
             IRubyObject backtrace = errorInfo.isNil() ? runtime.getNil() : errorInfo.callMethod(errorInfo.getRuntime().getCurrentContext(), "backtrace");
@@ -497,6 +526,7 @@ public class RubyGlobal {
             return backtrace;
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (runtime.getGlobalVariables().get("$!").isNil()) {
                 throw runtime.newArgumentError("$! not set.");
@@ -511,10 +541,12 @@ public class RubyGlobal {
             super(runtime, name, null);
         }
 
+        @Override
         public IRubyObject get() {
             return RuntimeHelpers.getLastLine(runtime, runtime.getCurrentContext());
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             RuntimeHelpers.setLastLine(runtime, runtime.getCurrentContext(), value);
             return value;
@@ -526,6 +558,7 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (value == get()) {
                 return value;
@@ -540,6 +573,7 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (value == get()) {
                 return value;
@@ -551,24 +585,16 @@ public class RubyGlobal {
                 // we set anything assigned to stdout/stderr to sync
                 io.getHandler().setSync(true);
             }
-            if (! value.respondsTo("write")) {
+
+            if (!value.respondsTo("write")) {
                 throw runtime.newTypeError(name() + " must have write method, " +
                                     value.getType().getName() + " given");
             }
-            
-            if ("$stderr".equals(name())) {
-                runtime.defineVariable(new OutputGlobalVariable(runtime, "$deferr", value));
-            }
-            
-            if ("$stdout".equals(name())) {
-                runtime.defineVariable(new OutputGlobalVariable(runtime, "$defout", value));
-                runtime.defineVariable(new OutputGlobalVariable(runtime, "$>", value));
-            }
 
             return super.set(value);
         }
     }
-    
+
     private static class LoadPath extends ReadonlyGlobalVariable {
         public LoadPath(Ruby runtime, String name) {
             super(runtime, name, null);
@@ -577,6 +603,7 @@ public class RubyGlobal {
         /**
          * @see org.jruby.runtime.GlobalVariable#get()
          */
+        @Override
         public IRubyObject get() {
             return runtime.getLoadService().getLoadPath();
         }
@@ -590,6 +617,7 @@ public class RubyGlobal {
         /**
          * @see org.jruby.runtime.GlobalVariable#get()
          */
+        @Override
         public IRubyObject get() {
             return runtime.getLoadService().getLoadedFeatures();
         }
