@@ -37,9 +37,12 @@ import java.io.PrintStream;
 import java.io.UnsupportedEncodingException;
 import java.util.ArrayList;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 
+import java.util.Set;
+import java.util.StringTokenizer;
 import org.jruby.ast.executable.Script;
 import org.jruby.exceptions.MainExitException;
 import org.jruby.runtime.Constants;
@@ -49,9 +52,40 @@ import org.jruby.util.JRubyFile;
 import org.jruby.util.KCode;
 import org.jruby.util.NormalizedFile;
 import org.jruby.util.SafePropertyAccessor;
+import org.objectweb.asm.Opcodes;
 
 public class RubyInstanceConfig {
 
+    /**
+     * The max count of active methods eligible for JIT-compilation.
+     */
+    private static final int JIT_MAX_METHODS_LIMIT = 4096;
+
+    /**
+     * The max size of JIT-compiled methods (full class size) allowed.
+     */
+    private static final int JIT_MAX_SIZE_LIMIT = Integer.MAX_VALUE;
+
+    /**
+     * The JIT threshold to the specified method invocation count.
+     */
+    private static final int JIT_THRESHOLD = 50;
+    
+    /** The version to use for generated classes. Set to current JVM version by default */
+    public static final int JAVA_VERSION;
+    
+    /**
+     * Default size for chained compilation.
+     */
+    private static final int CHAINED_COMPILE_LINE_COUNT_DEFAULT = 500;
+    
+    /**
+     * The number of lines at which a method, class, or block body is split into
+     * chained methods (to dodge 64k method-size limit in JVM).
+     */
+    public static final int CHAINED_COMPILE_LINE_COUNT
+            = SafePropertyAccessor.getInt("jruby.compile.chainsize", CHAINED_COMPILE_LINE_COUNT_DEFAULT);
+
     public enum CompileMode {
         JIT, FORCE, OFF;
 
@@ -93,6 +127,7 @@ public class RubyInstanceConfig {
     private final int jitLogEvery;
     private final int jitThreshold;
     private final int jitMax;
+    private final int jitMaxSize;
     private final boolean samplingEnabled;
     private CompatVersion compatVersion;
 
@@ -103,6 +138,7 @@ public class RubyInstanceConfig {
 
     // from CommandlineParser
     private List<String> loadPaths = new ArrayList<String>();
+    private Set<String> excludedMethods = new HashSet<String>();
     private StringBuffer inlineScript = new StringBuffer();
     private boolean hasInlineScript = false;
     private String scriptFileName = null;
@@ -118,6 +154,7 @@ public class RubyInstanceConfig {
     private Boolean verbose = Boolean.FALSE;
     private boolean debug = false;
     private boolean showVersion = false;
+    private boolean showBytecode = false;
     private boolean showCopyright = false;
     private boolean endOfArguments = false;
     private boolean shouldRunInterpreter = true;
@@ -130,6 +167,7 @@ public class RubyInstanceConfig {
     private String recordSeparator = "\n";
     private boolean shouldCheckSyntax = false;
     private String inputFieldSeparator = null;
+    private boolean managementEnabled = true;
 
     private int safeLevel = 0;
 
@@ -152,6 +190,7 @@ public class RubyInstanceConfig {
     public static final boolean THREADLESS_COMPILE_ENABLED
             = FASTEST_COMPILE_ENABLED
             || SafePropertyAccessor.getBoolean("jruby.compile.threadless");
+    public static final boolean LAZYHANDLES_COMPILE = SafePropertyAccessor.getBoolean("jruby.compile.lazyHandles", false);
     public static final boolean INDEXED_METHODS
             = SafePropertyAccessor.getBoolean("jruby.indexed.methods");
     public static final boolean FORK_ENABLED
@@ -168,7 +207,13 @@ public class RubyInstanceConfig {
     public static final boolean NATIVE_NET_PROTOCOL
             = SafePropertyAccessor.getBoolean("jruby.native.net.protocol", false);
 
+    public static boolean FULL_TRACE_ENABLED
+            = SafePropertyAccessor.getBoolean("jruby.debug.fullTrace", false);
+
+    public static final String COMPILE_EXCLUDE
+            = SafePropertyAccessor.getProperty("jruby.jit.exclude");
     public static boolean nativeEnabled = true;
+    
 
     public static interface LoadServiceCreator {
         LoadService create(Ruby runtime);
@@ -184,12 +229,24 @@ public class RubyInstanceConfig {
 
 
     static {
+        String specVersion = null;
         try {
+            specVersion = System.getProperty("jruby.bytecode.version");
+            if (specVersion == null) {
+                specVersion = System.getProperty("java.specification.version");
+            }
             if (System.getProperty("jruby.native.enabled") != null) {
                 nativeEnabled = Boolean.getBoolean("jruby.native.enabled");
             }
         } catch (SecurityException se) {
             nativeEnabled = false;
+            specVersion = "1.5";
+        }
+        
+        if (specVersion.equals("1.5")) {
+            JAVA_VERSION = Opcodes.V1_5;
+        } else {
+            JAVA_VERSION = Opcodes.V1_6;
         }
     }
 
@@ -220,10 +277,19 @@ public class RubyInstanceConfig {
             jitLogEvery = 0;
             jitThreshold = -1;
             jitMax = 0;
+            jitMaxSize = -1;
+            managementEnabled = false;
         } else {
             String threshold = SafePropertyAccessor.getProperty("jruby.jit.threshold");
             String max = SafePropertyAccessor.getProperty("jruby.jit.max");
-
+            String maxSize = SafePropertyAccessor.getProperty("jruby.jit.maxsize");
+            
+            if (COMPILE_EXCLUDE != null) {
+                String[] elements = COMPILE_EXCLUDE.split(",");
+                for (String element : elements) excludedMethods.add(element);
+            }
+            
+            managementEnabled = SafePropertyAccessor.getBoolean("jruby.management.enabled", true);
             runRubyInProcess = SafePropertyAccessor.getBoolean("jruby.launch.inproc", true);
             boolean jitProperty = SafePropertyAccessor.getProperty("jruby.jit.enabled") != null;
             if (jitProperty) {
@@ -247,8 +313,12 @@ public class RubyInstanceConfig {
             jitLoggingVerbose = SafePropertyAccessor.getBoolean("jruby.jit.logging.verbose");
             String logEvery = SafePropertyAccessor.getProperty("jruby.jit.logEvery");
             jitLogEvery = logEvery == null ? 0 : Integer.parseInt(logEvery);
-            jitThreshold = threshold == null ? 50 : Integer.parseInt(threshold);
-            jitMax = max == null ? 4096 : Integer.parseInt(max);
+            jitThreshold = threshold == null ?
+                    JIT_THRESHOLD : Integer.parseInt(threshold);
+            jitMax = max == null ?
+                    JIT_MAX_METHODS_LIMIT : Integer.parseInt(max);
+            jitMaxSize = maxSize == null ?
+                    JIT_MAX_SIZE_LIMIT : Integer.parseInt(maxSize);
         }
 
         // default ClassCache using jitMax as a soft upper bound
@@ -272,7 +342,7 @@ public class RubyInstanceConfig {
     }
 
     public String getBasicUsageHelp() {
-        StringBuffer sb = new StringBuffer();
+        StringBuilder sb = new StringBuilder();
         sb
                 .append("Usage: jruby [switches] [--] [programfile] [arguments]\n")
                 .append("  -0[octal]       specify record separator (\0, if no argument)\n")
@@ -287,7 +357,8 @@ public class RubyInstanceConfig {
                 .append("  -Idirectory     specify $LOAD_PATH directory (may be used more than once)\n")
                 .append("  -J[java option] pass an option on to the JVM (e.g. -J-Xmx512m)\n")
                 .append("                    use --properties to list JRuby properties\n")
-                .append("  -Kkcode         specifies KANJI (Japanese) code-set\n")
+                .append("                    run 'java -help' for a list of other Java options\n")
+                .append("  -Kkcode         specifies code-set (e.g. -Ku for Unicode\n")
                 .append("  -l              enable line ending processing\n")
                 .append("  -n              assume 'while gets(); ... end' loop around your script\n")
                 .append("  -p              assume loop like -n but print line also like sed\n")
@@ -304,13 +375,20 @@ public class RubyInstanceConfig {
                 .append("  --debug         sets the execution mode most suitable for debugger functionality\n")
                 .append("  --jdb           runs JRuby process under JDB\n")
                 .append("  --properties    List all configuration Java properties (pass -J-Dproperty=value)\n")
+                .append("  --sample        run with profiling using the JVM's sampling profiler\n")
+                .append("  --client        use the non-optimizing \"client\" JVM (improves startup; default)\n")
+                .append("  --server        use the optimizing \"server\" JVM (improves perf)\n")
+                .append("  --manage        enable remote JMX management and monitoring of the VM and JRuby\n")
+                .append("  --1.8           specify Ruby 1.8.x compatibility (default)\n")
+                .append("  --1.9           specify Ruby 1.9.x compatibility\n")
+                .append("  --bytecode      show the JVM bytecode produced by compiling specified code\n")
                 .append("  --version       print the version\n");
 
         return sb.toString();
     }
 
     public String getExtendedHelp() {
-        StringBuffer sb = new StringBuffer();
+        StringBuilder sb = new StringBuilder();
         sb
                 .append("These flags are for extended JRuby options.\n")
                 .append("Specify them by passing -X<option>\n")
@@ -326,7 +404,7 @@ public class RubyInstanceConfig {
     }
 
     public String getPropertyHelp() {
-        StringBuffer sb = new StringBuffer();
+        StringBuilder sb = new StringBuilder();
         sb
                 .append("These properties can be used to alter runtime behavior for perf or compatibility.\n")
                 .append("Specify them by passing -J-D<property>=<value>\n")
@@ -345,18 +423,27 @@ public class RubyInstanceConfig {
                 .append("       (EXPERIMENTAL) Turn on compilation without polling for \"unsafe\" thread events. Default is false\n")
                 .append("    jruby.compile.fastops=true|false\n")
                 .append("       (EXPERIMENTAL) Turn on fast operators for Fixnum. Default is false\n")
+                .append("    jruby.compile.chainsize=<line count>\n")
+                .append("       Set the number of lines at which compiled bodies are \"chained\". Default is " + CHAINED_COMPILE_LINE_COUNT_DEFAULT + "\n")
+                .append("    jruby.compile.lazyHandles=true|false\n")
+                .append("       Generate method bindings (handles) for compiled methods lazily. Default is false.")
                 .append("\nJIT SETTINGS:\n")
                 .append("    jruby.jit.threshold=<invocation count>\n")
-                .append("       Set the JIT threshold to the specified method invocation count. Default is 20\n")
+                .append("       Set the JIT threshold to the specified method invocation count. Default is " + JIT_THRESHOLD + ".\n")
                 .append("    jruby.jit.max=<method count>\n")
                 .append("       Set the max count of active methods eligible for JIT-compilation.\n")
-                .append("       Default is 2048 per runtime. A value of 0 disables JIT, -1 disables max.\n")
+                .append("       Default is " + JIT_MAX_METHODS_LIMIT + " per runtime. A value of 0 disables JIT, -1 disables max.\n")
+                .append("    jruby.jit.maxsize=<jitted method size (full .class)>\n")
+                .append("       Set the maximum full-class byte size allowed for jitted methods. Default is Integer.MAX_VALUE\n")
                 .append("    jruby.jit.logging=true|false\n")
                 .append("       Enable JIT logging (reports successful compilation). Default is false\n")
                 .append("    jruby.jit.logging.verbose=true|false\n")
                 .append("       Enable verbose JIT logging (reports failed compilation). Default is false\n")
                 .append("    jruby.jit.logEvery=<method count>\n")
                 .append("       Log a message every n methods JIT compiled. Default is 0 (off).\n")
+                .append("    jruby.jit.exclude=<ClsOrMod,ClsOrMod::method_name,-::method_name>\n")
+                .append("       Exclude methods from JIT by class/module short name, c/m::method_name,\n")
+                .append("       or -::method_name for anon/singleton classes/modules. Comma-delimited.\n")
                 .append("\nNATIVE SUPPORT:\n")
                 .append("    jruby.native.enabled=true|false\n")
                 .append("       Enable/disable native extensions (like JNA for non-Java APIs; Default is true\n")
@@ -382,30 +469,36 @@ public class RubyInstanceConfig {
                 .append("    jruby.objectspace.enabled=true|false\n")
                 .append("       Enable or disable ObjectSpace.each_object (default is disabled)\n")
                 .append("    jruby.launch.inproc=true|false\n")
-                .append("       Set in-process launching of e.g. system('ruby ...'). Default is true\n");
+                .append("       Set in-process launching of e.g. system('ruby ...'). Default is true\n")
+                .append("    jruby.bytecode.version=1.5|1.6\n")
+                .append("       Set bytecode version for JRuby to generate. Default is current JVM version.\n")
+                .append("    jruby.management.enabled=true|false\n")
+                .append("       Set whether JMX management is enabled. Default is true.\n")
+                .append("    jruby.debug.fullTrace=true|false\n")
+                .append("       Set whether full traces are enabled (c-call/c-return). Default is false.\n");
 
         return sb.toString();
     }
 
     public String getVersionString() {
-        StringBuffer buf = new StringBuffer("ruby ");
+        String ver = Constants.RUBY_VERSION;
         switch (compatVersion) {
         case RUBY1_8:
-            buf.append(Constants.RUBY_VERSION);
+            ver = Constants.RUBY_VERSION;
             break;
         case RUBY1_9:
-            buf.append(Constants.RUBY1_9_VERSION);
+            ver = Constants.RUBY1_9_VERSION;
             break;
         }
-        buf
-                .append(" (")
-                .append(Constants.COMPILE_DATE + " rev " + Constants.REVISION)
-                .append(") [")
-                .append(SafePropertyAccessor.getProperty("os.arch", "unknown") + "-jruby" + Constants.VERSION)
-                .append("]")
-                .append("\n");
 
-        return buf.toString();
+        String fullVersion = String.format(
+                "jruby %s (ruby %s patchlevel %s) (%s rev %s) [%s-java]\n",
+                Constants.VERSION, ver, Constants.RUBY_PATCHLEVEL,
+                Constants.COMPILE_DATE, Constants.REVISION,
+                SafePropertyAccessor.getProperty("os.arch", "unknown")
+                );
+
+        return fullVersion;
     }
 
     public String getCopyrightString() {
@@ -448,6 +541,10 @@ public class RubyInstanceConfig {
         return jitMax;
     }
 
+    public int getJitMaxSize() {
+        return jitMaxSize;
+    }
+
     public boolean isRunRubyInProcess() {
         return runRubyInProcess;
     }
@@ -794,9 +891,11 @@ public class RubyInstanceConfig {
 //                    case 'x' :
 //                        break;
                 case 'X':
-                    String extendedOption = grabValue("jruby: missing extended option, listing available options\n" + getExtendedHelp());
+                    String extendedOption = grabOptionalValue();
 
-                    if (extendedOption.equals("-O")) {
+                    if (extendedOption == null) {
+                        throw new MainExitException(0, "jruby: missing extended option, listing available options\n" + getExtendedHelp());
+                    } else if (extendedOption.equals("-O")) {
                         objectSpaceEnabled = false;
                     } else if (extendedOption.equals("+O")) {
                         objectSpaceEnabled = true;
@@ -836,6 +935,7 @@ public class RubyInstanceConfig {
                         break FOR;
                     } else if (argument.equals("--debug")) {
                         compileMode = CompileMode.OFF;
+                        FULL_TRACE_ENABLED = true;
                         System.setProperty("jruby.reflection", "true");
                         break FOR;
                     } else if (argument.equals("--jdb")) {
@@ -853,6 +953,9 @@ public class RubyInstanceConfig {
                     } else if (argument.equals("--version")) {
                         setShowVersion(true);
                         break FOR;
+                    } else if (argument.equals("--bytecode")) {
+                        setShowBytecode(true);
+                        break FOR;
                     } else {
                         if (argument.equals("--")) {
                             // ruby interpreter compatibilty
@@ -872,13 +975,18 @@ public class RubyInstanceConfig {
             if (scriptName.equals("irb")) {
                 scriptName = "jirb";
             }
-            try {
-                String jrubyHome = JRubyFile.create(System.getProperty("user.dir"), JRubyFile.getFileProperty("jruby.home")).getCanonicalPath();
-                scriptFileName = JRubyFile.create(jrubyHome + JRubyFile.separator + "bin", scriptName).getCanonicalPath();
-            } catch (IOException io) {
-                MainExitException mee = new MainExitException(1, "jruby: Can't determine script filename");
-                mee.setUsageError(true);
-                throw mee;
+
+            scriptFileName = scriptName;
+
+            if (!new File(scriptFileName).exists()) {
+                try {
+                    String jrubyHome = JRubyFile.create(System.getProperty("user.dir"), JRubyFile.getFileProperty("jruby.home")).getCanonicalPath();
+                    scriptFileName = JRubyFile.create(jrubyHome + JRubyFile.separator + "bin", scriptName).getCanonicalPath();
+                } catch (IOException io) {
+                    MainExitException mee = new MainExitException(1, "jruby: Can't determine script filename");
+                    mee.setUsageError(true);
+                    throw mee;
+                }
             }
 
             // route 'gem' through ruby code in case we're running out of the complete jar
@@ -917,11 +1025,7 @@ public class RubyInstanceConfig {
     }
 
     public byte[] inlineScript() {
-        try {
-            return inlineScript.toString().getBytes("UTF-8");
-        } catch (UnsupportedEncodingException e) {
-            return inlineScript.toString().getBytes();
-        }
+        return inlineScript.toString().getBytes();
     }
 
     public List<String> requiredLibraries() {
@@ -1033,6 +1137,10 @@ public class RubyInstanceConfig {
     public boolean isShowVersion() {
         return showVersion;
     }
+    
+    public boolean isShowBytecode() {
+        return showBytecode;
+    }
 
     public boolean isShowCopyright() {
         return showCopyright;
@@ -1041,6 +1149,10 @@ public class RubyInstanceConfig {
     protected void setShowVersion(boolean showVersion) {
         this.showVersion = showVersion;
     }
+    
+    protected void setShowBytecode(boolean showBytecode) {
+        this.showBytecode = showBytecode;
+    }
 
     protected void setShowCopyright(boolean showCopyright) {
         this.showCopyright = showCopyright;
@@ -1097,4 +1209,13 @@ public class RubyInstanceConfig {
     public Map getOptionGlobals() {
         return optionGlobals;
     }
+    
+    public boolean isManagementEnabled() {
+        return managementEnabled;
+    }
+    
+    public Set getExcludedMethods() {
+        return excludedMethods;
+    }
+    
 }
