@@ -81,6 +81,14 @@ public class RubyNumeric extends RubyObject {
     };
 
     public static double DBL_EPSILON=2.2204460492503131e-16;
+
+    private static IRubyObject convertToNum(double val, Ruby runtime) {
+
+        if (val >= (double) RubyFixnum.MAX || val < (double) RubyFixnum.MIN) {
+            return RubyBignum.newBignum(runtime, val);
+        }
+        return RubyFixnum.newFixnum(runtime, (long) val);
+    }
     
     public RubyNumeric(Ruby runtime, RubyClass metaClass) {
         super(runtime, metaClass);
@@ -157,7 +165,7 @@ public class RubyNumeric extends RubyObject {
         }
         if (arg.isNil()) {
             throw arg.getRuntime().newTypeError("no implicit conversion from nil to integer");
-    }
+        }
 
         if (arg instanceof RubyFloat) {
             double aFloat = ((RubyFloat) arg).getDoubleValue();
@@ -183,11 +191,7 @@ public class RubyNumeric extends RubyObject {
         if (Double.isNaN(val)) {
             throw runtime.newFloatDomainError("NaN");
         }
-
-        if (val >= (double) RubyFixnum.MAX || val < (double) RubyFixnum.MIN) {
-            return RubyBignum.newBignum(runtime, val);
-        }
-        return RubyFixnum.newFixnum(runtime, (long) val);
+        return convertToNum(val,runtime);
     }
 
     /** rb_num2dbl and NUM2DBL
@@ -286,28 +290,30 @@ public class RubyNumeric extends RubyObject {
     public static RubyInteger str2inum(Ruby runtime, RubyString str, int base, boolean strict) {
         if (base != 0 && (base < 2 || base > 36)) {
             throw runtime.newArgumentError("illegal radix " + base);
-            }
+        }
         ByteList bytes = str.getByteList();
         try {
-            return runtime.newFixnum(Convert.byteListToLong(bytes,base,strict));
-
+            return runtime.newFixnum(Convert.byteListToLong(bytes, base, strict));
         } catch (InvalidIntegerException e) {
-            if (strict) {
-                throw runtime.newArgumentError("invalid value for Integer: "
-                        + str.callMethod(runtime.getCurrentContext(), "inspect").toString());
-            }
-            return RubyFixnum.zero(runtime);
+            return str2inumIIE(strict, runtime, str);
         } catch (NumberTooLargeException e) {
+            return str2inumNTLE(strict, runtime, str, bytes, base);
+        }
+    }
+
+    private static RubyInteger str2inumIIE(boolean strict, Ruby runtime, RubyString str) throws RaiseException {
+        if (strict) {
+            throw runtime.newArgumentError("invalid value for Integer: " + str.callMethod(runtime.getCurrentContext(), "inspect").toString());
+        }
+        return RubyFixnum.zero(runtime);
+    }
+    
+    private static RubyInteger str2inumNTLE(boolean strict, Ruby runtime, RubyString str, ByteList bytes, int base) {
         try {
-                BigInteger bi = Convert.byteListToBigInteger(bytes,base,strict);
-                return new RubyBignum(runtime,bi);
-            } catch (InvalidIntegerException e2) {
-                if(strict) {
-                    throw runtime.newArgumentError("invalid value for Integer: "
-                            + str.callMethod(runtime.getCurrentContext(), "inspect").toString());
-                }
-                return RubyFixnum.zero(runtime);
-            }
+            BigInteger bi = Convert.byteListToBigInteger(bytes, base, strict);
+            return new RubyBignum(runtime, bi);
+        } catch (InvalidIntegerException e2) {
+            return str2inumIIE(strict, runtime, str);
         }
     }
 
@@ -442,16 +448,20 @@ public class RubyNumeric extends RubyObject {
      */
     protected final IRubyObject coerceRelOp(ThreadContext context, String method, IRubyObject other) {
         RubyArray ary = doCoerce(context, other, false);
-        if (ary != null) {
-            IRubyObject result = (ary.eltInternal(0)).callMethod(context, method,ary.eltInternal(1));
-            if (!result.isNil()) {
-                return result;
-            }
+        if (ary == null) {
+            return RubyComparable.cmperr(this, other);
         }
+
+        return unwrapCoerced(context, method, other, ary);
+    }
     
-        RubyComparable.cmperr(this, other); // MRI also does it that way       
-        return null; // not reachd as in MRI
-    } 
+    private final IRubyObject unwrapCoerced(ThreadContext context, String method, IRubyObject other, RubyArray ary) {
+        IRubyObject result = (ary.eltInternal(0)).callMethod(context, method, ary.eltInternal(1));
+        if (result.isNil()) {
+            return RubyComparable.cmperr(this, other);
+        }
+        return result;
+    }
         
     public RubyNumeric asNumeric() {
         return this;
@@ -465,7 +475,7 @@ public class RubyNumeric extends RubyObject {
     /** num_sadded
      *
      */
-    @JRubyMethod(name = "singleton_method_added", required = 1)
+    @JRubyMethod(name = "singleton_method_added")
     public IRubyObject sadded(IRubyObject name) {
         throw getRuntime().newTypeError("can't define singleton method " + name + " for " + getType().getName());
     } 
@@ -474,7 +484,7 @@ public class RubyNumeric extends RubyObject {
      *
      */
     @Override
-    @JRubyMethod(name = "initialize_copy", required = 1, visibility = Visibility.PRIVATE)
+    @JRubyMethod(name = "initialize_copy", visibility = Visibility.PRIVATE)
     public IRubyObject initialize_copy(IRubyObject arg) {
         throw getRuntime().newTypeError("can't copy " + getType().getName());
     }
@@ -482,7 +492,7 @@ public class RubyNumeric extends RubyObject {
     /** num_coerce
      *
      */
-    @JRubyMethod(name = "coerce", required = 1)
+    @JRubyMethod(name = "coerce")
     public IRubyObject coerce(IRubyObject other) {
         if (getClass() == other.getClass()) return getRuntime().newArray(other, this);
 
@@ -513,7 +523,7 @@ public class RubyNumeric extends RubyObject {
     /** num_cmp
      *
      */
-    @JRubyMethod(name = "<=>", required = 1)
+    @JRubyMethod(name = "<=>")
     public IRubyObject op_cmp(IRubyObject other) {
         if (this == other) { // won't hurt fixnums
             return RubyFixnum.zero(getRuntime());
@@ -524,7 +534,7 @@ public class RubyNumeric extends RubyObject {
     /** num_eql
      *
      */
-    @JRubyMethod(name = "eql?", required = 1)
+    @JRubyMethod(name = "eql?")
     public IRubyObject eql_p(ThreadContext context, IRubyObject other) {
         if (getClass() != other.getClass()) return getRuntime().getFalse();
         return equalInternal(context, this, other) ? getRuntime().getTrue() : getRuntime().getFalse();
@@ -533,7 +543,7 @@ public class RubyNumeric extends RubyObject {
     /** num_quo
      *
      */
-    @JRubyMethod(name = "quo", required = 1)
+    @JRubyMethod(name = "quo")
     public IRubyObject quo(ThreadContext context, IRubyObject other) {
         return callMethod(context, "/", other);
     }
@@ -541,7 +551,7 @@ public class RubyNumeric extends RubyObject {
     /** num_div
      * 
      */
-    @JRubyMethod(name = "div", required = 1)
+    @JRubyMethod(name = "div")
     public IRubyObject div(ThreadContext context, IRubyObject other) {
         return callMethod(context, "/", other).convertToFloat().floor();
     }
@@ -549,7 +559,7 @@ public class RubyNumeric extends RubyObject {
     /** num_divmod
      * 
      */
-    @JRubyMethod(name = "divmod", required = 1)
+    @JRubyMethod(name = "divmod")
     public IRubyObject divmod(ThreadContext context, IRubyObject other) {
         return RubyArray.newArray(getRuntime(), div(context, other), modulo(context, other));
     }
@@ -557,7 +567,7 @@ public class RubyNumeric extends RubyObject {
     /** num_modulo
      *
      */
-    @JRubyMethod(name = "modulo", required = 1)
+    @JRubyMethod(name = "modulo")
     public IRubyObject modulo(ThreadContext context, IRubyObject other) {
         return callMethod(context, "%", other);
     }
@@ -565,7 +575,7 @@ public class RubyNumeric extends RubyObject {
     /** num_remainder
      *
      */
-    @JRubyMethod(name = "remainder", required = 1)
+    @JRubyMethod(name = "remainder")
     public IRubyObject remainder(ThreadContext context, IRubyObject dividend) {
         IRubyObject z = callMethod(context, "%", dividend);
         IRubyObject x = this;
@@ -662,19 +672,21 @@ public class RubyNumeric extends RubyObject {
     
     @JRubyMethod(name = "step", required = 1, optional = 1, frame = true)
     public IRubyObject step(ThreadContext context, IRubyObject[] args, Block block) {
-        IRubyObject to;
-        IRubyObject step;
-        
-        if(args.length == 1){ 
-            to = args[0];
-            step = RubyFixnum.one(getRuntime());
-        } else if (args.length == 2) {
-            to = args[0];
-            step = args[1];
-        }else{
-            throw getRuntime().newTypeError("wrong number of arguments");
+        switch (args.length) {
+        case 0: throw context.getRuntime().newArgumentError(0, 1);
+        case 1: return step(context, args[0], block);
+        case 2: return step(context, args[0], args[1], block);
+        default: throw context.getRuntime().newArgumentError(args.length, 2);
         }
-        
+    }
+    
+    @JRubyMethod(name = "step", frame = true)
+    public IRubyObject step(ThreadContext context, IRubyObject arg0, Block block) {
+        return step(context, arg0, RubyFixnum.one(context.getRuntime()), block);
+    }
+    
+    @JRubyMethod(name = "step", frame = true)
+    public IRubyObject step(ThreadContext context, IRubyObject to, IRubyObject step, Block block) {
         if (this instanceof RubyFixnum && to instanceof RubyFixnum && step instanceof RubyFixnum) {
             long value = getLongValue();
             long end = ((RubyFixnum) to).getLongValue();
@@ -772,5 +784,4 @@ public class RubyNumeric extends RubyObject {
             return this;
         }
     }
-    
 }
