@@ -48,6 +48,8 @@ import org.joni.Region;
 import org.joni.Syntax;
 import org.joni.WarnCallback;
 import org.joni.encoding.Encoding;
+import org.joni.exception.JOniException;
+
 import static org.jruby.anno.FrameField.*;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.anno.JRubyClass;
@@ -76,14 +78,6 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     private Regex pattern;
     private ByteList str;
 
-    private final static byte[] PIPE = new byte[]{'|'};
-    private final static byte[] DASH = new byte[]{'-'};
-    private final static byte[] R_PAREN = new byte[]{')'};
-    private final static byte[] COLON = new byte[]{':'};
-    private final static byte[] M_CHAR = new byte[]{'m'};
-    private final static byte[] I_CHAR = new byte[]{'i'};
-    private final static byte[] X_CHAR = new byte[]{'x'};
-
     private static final int REGEXP_LITERAL_F =     1 << 11;
     private static final int REGEXP_KCODE_DEFAULT = 1 << 12;
 
@@ -131,6 +125,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
         runtime.setRegexp(regexpClass);
         regexpClass.index = ClassIndex.REGEXP;
         regexpClass.kindOf = new RubyModule.KindOf() {
+            @Override
                 public boolean isKindOf(IRubyObject obj, RubyModule type) {
                     return obj instanceof RubyRegexp;
                 }
@@ -189,10 +184,12 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     }
 
     @JRubyMethod(name = "kcode")
-    public IRubyObject kcode() {
-        return (!isKCodeDefault() && kcode != null) ? getRuntime().newString(kcode.name()) : getRuntime().getNil();
+    public IRubyObject kcode(ThreadContext context) {
+        return (!isKCodeDefault() && kcode != null) ? 
+            context.getRuntime().newString(kcode.name()) : context.getRuntime().getNil();
     }
 
+    @Override
     public int getNativeTypeIndex() {
         return ClassIndex.REGEXP;
     }
@@ -206,6 +203,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     }
 
     @JRubyMethod(name = "hash")
+    @Override
     public RubyFixnum hash() {
         check();
         int hashval = (int)pattern.getOptions();
@@ -219,33 +217,32 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     }
 
     @JRubyMethod(name = {"==", "eql?"}, required = 1)
-    public IRubyObject op_equal(IRubyObject other) {
-        if(this == other) return getRuntime().getTrue();
-        if(!(other instanceof RubyRegexp)) return getRuntime().getFalse();
+    @Override
+    public IRubyObject op_equal(ThreadContext context, IRubyObject other) {
+        if(this == other) return context.getRuntime().getTrue();
+        if(!(other instanceof RubyRegexp)) return context.getRuntime().getFalse();
         RubyRegexp otherRegex = (RubyRegexp)other;
         
         check();
         otherRegex.check();
         
-        if(str.equal(otherRegex.str) && kcode == otherRegex.kcode &&
-                                      pattern.getOptions() == otherRegex.pattern.getOptions()) {
-            return getRuntime().getTrue();
-        }
-        return getRuntime().getFalse();
+        return context.getRuntime().newBoolean(str.equal(otherRegex.str) && 
+                kcode == otherRegex.kcode && pattern.getOptions() == otherRegex.pattern.getOptions());
     }
 
     @JRubyMethod(name = "~", reads = {LASTLINE, BACKREF}, writes = BACKREF)
     public IRubyObject op_match2(ThreadContext context) {
+        Ruby runtime = context.getRuntime();
         IRubyObject line = context.getCurrentFrame().getLastLine();
         if(!(line instanceof RubyString)) {
-            context.getCurrentFrame().setBackRef(getRuntime().getNil());
-            return getRuntime().getNil();
+            context.getCurrentFrame().setBackRef(runtime.getNil());
+            return runtime.getNil();
         }
         int start = search(context, (RubyString)line, 0, false);
         if(start < 0) {
-            return getRuntime().getNil();
+            return runtime.getNil();
         } else {
-            return getRuntime().newFixnum(start);
+            return runtime.newFixnum(start);
         }
     }
 
@@ -254,14 +251,15 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
      */
     @JRubyMethod(name = "===", required = 1, writes = BACKREF)
     public IRubyObject eqq(ThreadContext context, IRubyObject str) {
+        Ruby runtime = context.getRuntime();        
         if(!(str instanceof RubyString)) str = str.checkStringType();
 
         if (str.isNil()) {
-            context.getCurrentFrame().setBackRef(getRuntime().getNil());
-            return getRuntime().getFalse();
+            context.getCurrentFrame().setBackRef(runtime.getNil());
+            return runtime.getFalse();
         }
         int start = search(context, (RubyString)str, 0, false);
-        return (start < 0) ? getRuntime().getFalse() : getRuntime().getTrue();
+        return (start < 0) ? runtime.getFalse() : runtime.getTrue();
     }
 
     private static final int REGEX_QUOTED = 1;
@@ -302,8 +300,8 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
         throw getRuntime().newRegexpError(err + ": " + rb_reg_desc(s,start, len,flags));
     }
 
-    private final StringBuffer rb_reg_desc(byte[] s, int start, int len, int flags) {
-        StringBuffer sb = new StringBuffer("/");
+    private final StringBuilder rb_reg_desc(byte[] s, int start, int len, int flags) {
+        StringBuilder sb = new StringBuilder("/");
         rb_reg_expr_str(sb, s, start, len);
         sb.append("/");
 
@@ -317,7 +315,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
         return sb;
     }
 
-    private final void rb_reg_expr_str(StringBuffer sb, byte[] s, int start, int len) {
+    private final void rb_reg_expr_str(StringBuilder sb, byte[] s, int start, int len) {
         int p,pend;
         boolean need_escape = false;
         p = start;
@@ -365,6 +363,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     /** rb_reg_init_copy
      */
     @JRubyMethod(name = "initialize_copy", required = 1)
+    @Override
     public IRubyObject initialize_copy(IRubyObject re) {
         if(this == re) return this;
 
@@ -372,7 +371,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
 
         if (getMetaClass().getRealClass() != re.getMetaClass().getRealClass()) {
             throw getRuntime().newTypeError("wrong argument type");
-	    }
+        }
 
         RubyRegexp regexp = (RubyRegexp)re;
         regexp.check();
@@ -514,7 +513,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     /** rb_reg_search
      */
     public int search(ThreadContext context, RubyString str, int pos, boolean reverse) {
-        Ruby runtime = getRuntime();
+        Ruby runtime = context.getRuntime();
         Frame frame = context.getCurrentRubyFrame();
 
         ByteList value = str.getByteList();
@@ -522,46 +521,47 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
             frame.setBackRef(runtime.getNil());
             return -1;
         }
-        
+
+        return performSearch(reverse, pos, value, frame, runtime, context, str);
+    }
+
+    private int performSearch(boolean reverse, int pos, ByteList value, Frame frame, Ruby runtime, ThreadContext context, RubyString str) {
         check();
-        
-        int range = reverse ? -pos : value.realSize - pos;
-        
-        Matcher matcher = pattern.matcher(value.bytes, value.begin, value.begin + value.realSize);
-        
-        int result = matcher.search(value.begin + pos, 
-                                    value.begin + pos + range,
-                                    Option.NONE);
-        
+
+        int realSize = value.realSize;
+        int begin = value.begin;
+        int range = reverse ? -pos : realSize - pos;
+
+        Matcher matcher = pattern.matcher(value.bytes, begin, begin + realSize);
+
+        int result = matcher.search(begin + pos, begin + pos + range, Option.NONE);
+
         if (result < 0) {
             frame.setBackRef(runtime.getNil());
             return result;
         }
-        
-        updateBackRef(str, frame, matcher);
+
+        updateBackRef(context, str, frame, matcher);
 
         return result;
     }
     
-    final RubyMatchData updateBackRef(RubyString str, Frame frame, Matcher matcher) {
+    final RubyMatchData updateBackRef(ThreadContext context, RubyString str, Frame frame, Matcher matcher) {
+        Ruby runtime = context.getRuntime();
         IRubyObject backref = frame.getBackRef();
         final RubyMatchData match;
         if (backref == null || backref.isNil() || ((RubyMatchData)backref).used()) {
-            match = new RubyMatchData(getRuntime());
+            match = new RubyMatchData(runtime);
         } else {
             match = (RubyMatchData)backref;
-            if(getRuntime().getSafeLevel() >= 3) {
-                match.setTaint(true);
-            } else {
-                match.setTaint(false);
-            }
+            match.setTaint(runtime.getSafeLevel() >= 3);
         }
         
         match.regs = matcher.getRegion(); // lazy, null when no groups defined
         match.begin = matcher.getBegin();
         match.end = matcher.getEnd();
         
-        match.str = (RubyString)str.strDup().freeze();
+        match.str = (RubyString)str.strDup(runtime).freeze(context);
         match.pattern = pattern;
 
         frame.setBackRef(match);
@@ -575,18 +575,19 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
      * 
      */
     @JRubyMethod(name = "=~", required = 1, reads = BACKREF, writes = BACKREF)
+    @Override
     public IRubyObject op_match(ThreadContext context, IRubyObject str) {
         int start;
         if(str.isNil()) {
-            context.getCurrentFrame().setBackRef(getRuntime().getNil());
+            context.getCurrentFrame().setBackRef(context.getRuntime().getNil());
             return str;
         }
         
         start = search(context, str.convertToString(), 0, false);
 
-        if(start < 0) return getRuntime().getNil();
+        if (start < 0) return context.getRuntime().getNil();
 
-        return RubyFixnum.newFixnum(getRuntime(), start);
+        return RubyFixnum.newFixnum(context.getRuntime(), start);
     }
 
     /** rb_reg_match_m
@@ -594,11 +595,10 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
      */
     @JRubyMethod(name = "match", required = 1, reads = BACKREF)
     public IRubyObject match_m(ThreadContext context, IRubyObject str) {
-        if(op_match(context, str).isNil()) {
-            return getRuntime().getNil();
-        }
+        if (op_match(context, str).isNil()) return context.getRuntime().getNil();
+
         IRubyObject result =  context.getCurrentFrame().getBackRef();
-        if(result instanceof RubyMatchData) {
+        if (result instanceof RubyMatchData) {
             ((RubyMatchData)result).use();
         }
         return result;
@@ -702,12 +702,10 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     }
 
     @JRubyMethod(name = "casefold?")
-    public IRubyObject casefold_p() {
+    public IRubyObject casefold_p(ThreadContext context) {
         check();
-        if ((pattern.getOptions() & RE_OPTION_IGNORECASE) != 0) {
-            return getRuntime().getTrue();
-        }
-        return getRuntime().getFalse();
+        
+        return context.getRuntime().newBoolean((pattern.getOptions() & RE_OPTION_IGNORECASE) != 0);
     }
 
     /** rb_reg_source
@@ -715,10 +713,11 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
      */
     @JRubyMethod(name = "source")
     public IRubyObject source() {
+        Ruby runtime = getRuntime();
         check();
-        RubyString str = RubyString.newStringShared(getRuntime(), this.str);
+        RubyString str = RubyString.newStringShared(runtime, this.str);
         if(isTaint()) {
-            str.taint();
+            str.taint(runtime.getCurrentContext());
         }
         return str;
     }
@@ -731,6 +730,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
      *
      */
     @JRubyMethod(name = "inspect")
+    @Override
     public IRubyObject inspect() {
         check();
         return getRuntime().newString(ByteList.create(rb_reg_desc(str.bytes, str.begin, str.realSize, pattern.getOptions()).toString()));
@@ -739,130 +739,124 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
     private final static int EMBEDDABLE = RE_OPTION_MULTILINE|RE_OPTION_IGNORECASE|RE_OPTION_EXTENDED;
 
     @JRubyMethod(name = "to_s")
+    @Override
     public IRubyObject to_s() {
         RubyString ss = getRuntime().newString("(?");
         check();
+
         int options = pattern.getOptions();
-        int p = str.begin;
-        int l = str.realSize;
-        byte[] _str = str.bytes;
-
-
+        int ptr = str.begin;
+        int len = str.realSize;
+        byte[] bytes = str.bytes;
         again: do {
-            if(l >= 4 && _str[p] == '(' && _str[p+1] == '?') {
+            if (len >= 4 && bytes[ptr] == '(' && bytes[ptr + 1] == '?') {
                 boolean err = true;
-                p += 2;
-                if((l -= 2) > 0) {
+                ptr += 2;
+                if ((len -= 2) > 0) {
                     do {
-                        if(_str[p] == 'm') {
+                        if (bytes[ptr] == 'm') {
                             options |= RE_OPTION_MULTILINE;
-                        } else if(_str[p] == 'i') {
+                        } else if (bytes[ptr] == 'i') {
                             options |= RE_OPTION_IGNORECASE;
-                        } else if(_str[p] == 'x') {
+                        } else if (bytes[ptr] == 'x') {
                             options |= RE_OPTION_EXTENDED;
                         } else {
                             break;
                         }
-                        p++;
-                    } while(--l > 0);
+                        ptr++;
+                    } while(--len > 0);
                 }
-                if(l > 1 && _str[p] == '-') {
-                    ++p;
-                    --l;
+                if (len > 1 && bytes[ptr] == '-') {
+                    ++ptr;
+                    --len;
                     do {
-                        if(_str[p] == 'm') {
+                        if (bytes[ptr] == 'm') {
                             options &= ~RE_OPTION_MULTILINE;
-                        } else if(_str[p] == 'i') {
+                        } else if (bytes[ptr] == 'i') {
                             options &= ~RE_OPTION_IGNORECASE;
-                        } else if(_str[p] == 'x') {
+                        } else if (bytes[ptr] == 'x') {
                             options &= ~RE_OPTION_EXTENDED;
                         } else {
                             break;
                         }
-                        p++;
-                    } while(--l > 0);
+                        ptr++;
+                    } while(--len > 0);
                 }
-                if(_str[p] == ')') {
-                    --l;
-                    ++p;
+
+                if (bytes[ptr] == ')') {
+                    --len;
+                    ++ptr;
                     continue again;
                 }
-                if(_str[p] == ':' && _str[p+l-1] == ')') {
+
+                if (bytes[ptr] == ':' && bytes[ptr + len - 1] == ')') {
                     try {
-                        new Regex(_str,++p,l-=2,Option.DEFAULT,kcode.getEncoding(),Syntax.DEFAULT);
+                        new Regex(bytes, ++ptr, ptr + (len-=2) ,Option.DEFAULT, kcode.getEncoding(), Syntax.DEFAULT);
                         err = false;
-                    } catch(Exception e) {
+                    } catch (JOniException e) {
                         err = true;
                     }
                 }
-                if(err) {
+
+                if (err) {
                     options = (int)pattern.getOptions();
-                    p = str.begin;
-                    l = str.realSize;
+                    ptr = str.begin;
+                    len = str.realSize;
                 }
             }
+            if ((options & RE_OPTION_MULTILINE) !=0 ) ss.cat((byte)'m');
+            if ((options & RE_OPTION_IGNORECASE) !=0 ) ss.cat((byte)'i');
+            if ((options & RE_OPTION_EXTENDED) !=0 ) ss.cat((byte)'x');
 
-            if((options & RE_OPTION_MULTILINE)!=0) ss.cat(M_CHAR);
-            if((options & RE_OPTION_IGNORECASE)!=0) ss.cat(I_CHAR);
-            if((options & RE_OPTION_EXTENDED)!=0) ss.cat(X_CHAR);
-
-            if((options&EMBEDDABLE) != EMBEDDABLE) {
-                ss.cat(DASH);
-                if((options & RE_OPTION_MULTILINE)==0) ss.cat(M_CHAR);
-                if((options & RE_OPTION_IGNORECASE)==0) ss.cat(I_CHAR);
-                if((options & RE_OPTION_EXTENDED)==0) ss.cat(X_CHAR);
+            if ((options & EMBEDDABLE) != EMBEDDABLE) {
+                ss.cat((byte)'-');
+                if ((options & RE_OPTION_MULTILINE) == 0) ss.cat((byte)'m');
+                if ((options & RE_OPTION_IGNORECASE) == 0) ss.cat((byte)'i');
+                if ((options & RE_OPTION_EXTENDED) == 0) ss.cat((byte)'x');
             }
-            ss.cat(COLON);
-            rb_reg_expr_str(ss,p,l);
-            ss.cat(R_PAREN);
+            ss.cat((byte)':');
+            rb_reg_expr_str(ss, ptr, len);
+            ss.cat((byte)')');
             ss.infectBy(this);
             return ss;
         } while(true);
     }
 
-    private final boolean ISPRINT(byte c) {
-        return ISPRINT((char)(c&0xFF));
-    }
-
-    private final boolean ISPRINT(char c) {
-        return (' ' == c || (!Character.isWhitespace(c) && !Character.isISOControl(c)));
-    }
-
-    private final void rb_reg_expr_str(RubyString ss, int s, int l) {
+    private final void rb_reg_expr_str(RubyString ss, int s, int len) {
         int p = s;
-        int pend = l;
+        int pend = p + len;
         boolean need_escape = false;
-        while(p<pend) {
-            if(str.bytes[p] == '/' || (!ISPRINT(str.bytes[p]) && kcode.getEncoding().length(str.bytes[p]) == 1)) {
+        Encoding enc = kcode.getEncoding();
+        while (p < pend) {
+            if (str.bytes[p] == '/' || (!enc.isPrint(str.bytes[p] & 0xff) && enc.length(str.bytes[p]) == 1)) {
                 need_escape = true;
                 break;
             }
-            p += kcode.getEncoding().length(str.bytes[p]);
+            p += enc.length(str.bytes[p]);
         }
-        if(!need_escape) {
-            ss.cat(str.bytes,s,l);
+        if (!need_escape) {
+            ss.cat(str.bytes, s, len);
         } else {
             p = s; 
-            while(p<pend) {
-                if(str.bytes[p] == '\\') {
-                    int n = kcode.getEncoding().length(str.bytes[p+1]) + 1;
-                    ss.cat(str.bytes,p,n);
+            while (p<pend) {
+                if (str.bytes[p] == '\\') {
+                    int n = enc.length(str.bytes[p+1]) + 1;
+                    ss.cat(str.bytes, p, n);
                     p += n;
                     continue;
-                } else if(str.bytes[p] == '/') {
-                    char c = '\\';
-                    ss.cat((byte)c);
-                    ss.cat(str.bytes,p,1);
-                } else if(kcode.getEncoding().length(str.bytes[p]) != 1) {
-                    ss.cat(str.bytes,p,kcode.getEncoding().length(str.bytes[p]));
-                    p += kcode.getEncoding().length(str.bytes[p]);
+                } else if (str.bytes[p] == '/') {
+                    ss.cat((byte)'\\');
+                    ss.cat(str.bytes, p, 1);
+                } else if (enc.length(str.bytes[p]) != 1) {
+                    ss.cat(str.bytes, p, enc.length(str.bytes[p]));
+                    p += enc.length(str.bytes[p]);
                     continue;
-                } else if(ISPRINT(str.bytes[p])) {
+                } else if (enc.isPrint(str.bytes[p] & 0xff)) {
                     ss.cat(str.bytes,p,1);
-                } else if(!Character.isWhitespace(str.bytes[p])) {
-                    ss.cat(ByteList.create(Integer.toString(str.bytes[p]&0377,8)));
+                } else if (!enc.isSpace(str.bytes[p] & 0xff)) {
+                    ss.cat(ByteList.create(Integer.toString(str.bytes[p] & 0377, 8)));
                 } else {
-                    ss.cat(str.bytes,p,1);
+                    ss.cat(str.bytes, p, 1);
                 }
                 p++;
             }
@@ -993,9 +987,9 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
             end = m.regs.end[nth];
         }
         
-        if(start == -1) return match.getRuntime().getNil();
+        if (start == -1) return match.getRuntime().getNil();
 
-        RubyString str = m.str.makeShared(start, end - start);
+        RubyString str = m.str.makeShared(match.getRuntime(), start, end - start);
         str.infectBy(match);
         return str;
     }
@@ -1054,7 +1048,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
         
         if(beg == -1) match.getRuntime().getNil(); 
 
-        RubyString str = m.str.makeShared(0, beg);
+        RubyString str = m.str.makeShared(match.getRuntime(), 0, beg);
         str.infectBy(match);
         return str;
     }
@@ -1075,7 +1069,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
             end = m.regs.end[0];
         }
         
-        RubyString str = m.str.makeShared(end, m.str.getByteList().realSize - end);
+        RubyString str = m.str.makeShared(match.getRuntime(), end, m.str.getByteList().realSize - end);
         str.infectBy(match);
         return str;
     }
@@ -1117,9 +1111,9 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
             RubyString source = recv.getRuntime().newString();
             IRubyObject[] _args = new IRubyObject[3];
 
-            for(int i = 0; i < args.length; i++) {
-                if(0 < i) {
-                    source.cat(PIPE);
+            for (int i = 0; i < args.length; i++) {
+                if (0 < i) {
+                    source.cat((byte)'|');
                 }
                 IRubyObject v = TypeConverter.convertToTypeWithCheck(args[i], recv.getRuntime().getRegexp(), 0, "to_regexp");
                 if (!v.isNil()) {
@@ -1176,7 +1170,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
      * 
      */
     @JRubyMethod(name = "named_captures", compat = CompatVersion.RUBY1_9)
-    public IRubyObject named_captures() {
+    public IRubyObject named_captures(ThreadContext context) {
         RubyHash hash = RubyHash.newHash(getRuntime());
         if (pattern.numberOfNames() == 0) return hash;
 
@@ -1186,7 +1180,7 @@ public class RubyRegexp extends RubyObject implements ReOptions, WarnCallback {
             RubyArray ary = getRuntime().newArray(backrefs.length);
 
             for (int backref : backrefs) ary.append(RubyFixnum.newFixnum(getRuntime(), backref));
-            hash.fastASet(RubyString.newStringShared(getRuntime(), e.name, e.nameP, e.nameEnd - e.nameP).freeze(), ary);
+            hash.fastASet(RubyString.newStringShared(getRuntime(), e.name, e.nameP, e.nameEnd - e.nameP).freeze(context), ary);
         }
         return hash;
     }    
