@@ -56,7 +56,6 @@ import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 
 import java.util.concurrent.ExecutionException;
-import java.util.concurrent.TimeoutException;
 import java.util.concurrent.locks.ReentrantLock;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.anno.JRubyClass;
@@ -87,6 +86,7 @@ public class RubyThread extends RubyObject {
 
     private final ThreadService threadService;
     private volatile boolean isStopped = false;
+    private volatile boolean isDead = false;
     public Object stopLock = new Object();
     
     private volatile boolean killed = false;
@@ -99,9 +99,6 @@ public class RubyThread extends RubyObject {
     protected RubyThread(Ruby runtime, RubyClass type) {
         super(runtime, type);
         this.threadService = runtime.getThreadService();
-        // set to default thread group
-        RubyThreadGroup defaultThreadGroup = (RubyThreadGroup)runtime.getThreadGroup().fastGetConstant("Default");
-        defaultThreadGroup.add(this, Block.NULL_BLOCK);
         finalResult = runtime.getNil();
     }
     
@@ -126,6 +123,9 @@ public class RubyThread extends RubyObject {
         rubyThread.threadImpl = new NativeThread(rubyThread, Thread.currentThread());
         runtime.getThreadService().setMainThread(rubyThread);
         
+        // set to default thread group
+        runtime.getDefaultThreadGroup().addDirectly(rubyThread);
+        
         threadClass.setMarshal(ObjectMarshal.NOT_MARSHALABLE_MARSHAL);
         
         return threadClass;
@@ -175,20 +175,32 @@ public class RubyThread extends RubyObject {
         
         context.preAdoptThread();
         
+        // set to default thread group
+        runtime.getDefaultThreadGroup().addDirectly(rubyThread);
+        
         return rubyThread;
     }
     
     @JRubyMethod(name = "initialize", rest = true, frame = true, visibility = Visibility.PRIVATE)
     public IRubyObject initialize(IRubyObject[] args, Block block) {
-        if (!block.isGiven()) throw getRuntime().newThreadError("must be called with a block");
+        Ruby runtime = getRuntime();
+        if (!block.isGiven()) throw runtime.newThreadError("must be called with a block");
 
         if (RubyInstanceConfig.POOLING_ENABLED) {
             threadImpl = new FutureThread(this, new RubyRunnable(this, args, block));
         } else {
             threadImpl = new NativeThread(this, new RubyNativeThread(this, args, block));
         }
+        
+        // set to default thread group
+        runtime.getDefaultThreadGroup().addDirectly(this);
+        
         threadImpl.start();
         
+        // We yield here to hopefully permit the target thread to schedule
+        // MRI immediately schedules it, so this is close but not exact
+        Thread.yield();
+        
         return this;
     }
     
@@ -205,12 +217,6 @@ public class RubyThread extends RubyObject {
         return rubyThread;
     }
     
-    private void ensureCurrent(ThreadContext context) {
-        if (this != context.getThread()) {
-            throw new RuntimeException("internal thread method called from another thread");
-        }
-    }
-    
     private void ensureNotCurrent() {
         if (this == getRuntime().getCurrentContext().getThread()) {
             throw new RuntimeException("internal thread method called from another thread");
@@ -220,6 +226,7 @@ public class RubyThread extends RubyObject {
     public synchronized void cleanTerminate(IRubyObject result) {
         finalResult = result;
         isStopped = true;
+        isDead = true;
     }
 
     public void pollThreadEvents() {
@@ -230,20 +237,12 @@ public class RubyThread extends RubyObject {
         // check for criticalization *before* locking ourselves
         threadService.waitForCritical();
 
-        ensureCurrent(context);
-
-        if (DEBUG) System.out.println("thread " + Thread.currentThread() + " before");
-        if (killed) throw new ThreadKill();
-
-        if (DEBUG) System.out.println("thread " + Thread.currentThread() + " after");
-        if (receivedException != null) {
-            // clear this so we don't keep re-throwing
-            IRubyObject raiseException = receivedException;
-            receivedException = null;
-            RubyModule kernelModule = getRuntime().getKernel();
-            if (DEBUG) System.out.println("thread " + Thread.currentThread() + " before propagating exception: " + killed);
-            kernelModule.callMethod(context, "raise", raiseException);
-        }
+        if (killed) throwThreadKill();
+        if (receivedException != null) receivedAnException(context);
+    }
+    
+    private void throwThreadKill() {
+        throw new ThreadKill();
     }
 
     /**
@@ -339,10 +338,10 @@ public class RubyThread extends RubyObject {
 
     @JRubyMethod(name = "alive?")
     public RubyBoolean alive_p() {
-        return threadImpl.isAlive() ? getRuntime().getTrue() : getRuntime().getFalse();
+        return !isDead && threadImpl.isAlive() ? getRuntime().getTrue() : getRuntime().getFalse();
     }
 
-    @JRubyMethod(name = "join", optional = 1)
+    @JRubyMethod(name = "join", optional = 1, backtrace = true)
     public IRubyObject join(IRubyObject[] args) {
         long timeoutMillis = Long.MAX_VALUE;
         if (args.length > 0) {
@@ -353,7 +352,7 @@ public class RubyThread extends RubyObject {
             // than or equal to zero returns immediately; returns nil
             timeoutMillis = (long)(1000.0D * args[0].convertToFloat().getValue());
             if (timeoutMillis <= 0) {
-	        // TODO: not sure that we should skip caling join() altogether.
+	        // TODO: not sure that we should skip calling join() altogether.
 		// Thread.join() has some implications for Java Memory Model, etc.
 	        if (threadImpl.isAlive()) {
 		   return getRuntime().getNil();
@@ -440,7 +439,7 @@ public class RubyThread extends RubyObject {
     @JRubyMethod(name = "inspect")
     public IRubyObject inspect() {
         // FIXME: There's some code duplication here with RubyObject#inspect
-        StringBuffer part = new StringBuffer();
+        StringBuilder part = new StringBuilder();
         String cname = getMetaClass().getRealClass().getName();
         part.append("#<").append(cname).append(":0x");
         part.append(Integer.toHexString(System.identityHashCode(this)));
@@ -658,7 +657,7 @@ public class RubyThread extends RubyObject {
     }
     
     public void sleep(long millis) throws InterruptedException {
-        ensureCurrent(getRuntime().getCurrentContext());
+        assert this == getRuntime().getCurrentContext().getThread();
         synchronized (stopLock) {
             pollThreadEvents();
             try {
@@ -684,7 +683,7 @@ public class RubyThread extends RubyObject {
         } else if (exitingException != null) {
             return getRuntime().getNil();
         } else {
-            return getRuntime().newBoolean(false);
+            return getRuntime().getFalse();
         }
     }
 
@@ -693,6 +692,9 @@ public class RubyThread extends RubyObject {
     	// need to reexamine this
         RubyThread currentThread = getRuntime().getCurrentContext().getThread();
         
+        // If the killee thread is the same as the killer thread, just die
+        if (currentThread == this) throwThreadKill();
+        
         try {
             if (DEBUG) System.out.println("thread " + Thread.currentThread() + " trying to kill");
             while (!(currentThread.lock.tryLock() && this.lock.tryLock())) {
@@ -759,10 +761,10 @@ public class RubyThread extends RubyObject {
         if (runtime.getSystemExit().isInstance(rubyException)) {
             threadService.getMainThread().raise(new IRubyObject[] {rubyException}, Block.NULL_BLOCK);
         } else if (abortOnException(runtime)) {
-            // FIXME: printError explodes on some nullpointer
-            //getRuntime().getRuntime().printError(exception.getException());
+            runtime.printError(rubyException);
             RubyException systemExit = RubySystemExit.newInstance(runtime, 1);
             systemExit.message = rubyException.message;
+            systemExit.set_backtrace(rubyException.backtrace());
             threadService.getMainThread().raise(new IRubyObject[] {systemExit}, Block.NULL_BLOCK);
             return;
         } else if (runtime.getDebug().isTrue()) {
@@ -788,11 +790,15 @@ public class RubyThread extends RubyObject {
             SelectableChannel selectable = (SelectableChannel)channel;
             
             try {
+                io.addBlockingThread(this);
                 currentSelector = selectable.provider().openSelector();
             
                 SelectionKey key = selectable.register(currentSelector, SelectionKey.OP_ACCEPT);
 
                 int result = currentSelector.select();
+                
+                // check for thread events, in case we've been woken up to die
+                pollThreadEvents();
 
                 if (result == 1) {
                     Set<SelectionKey> keySet = currentSelector.selectedKeys();
@@ -814,6 +820,7 @@ public class RubyThread extends RubyObject {
                     }
                 }
                 currentSelector = null;
+                io.removeBlockingThread(this);
             }
         } else {
             // can't select, just have to do a blocking call
@@ -834,4 +841,15 @@ public class RubyThread extends RubyObject {
     public void afterBlockingCall() {
         isStopped = false;
     }
+
+    private void receivedAnException(ThreadContext context) {
+        // clear this so we don't keep re-throwing
+        IRubyObject raiseException = receivedException;
+        receivedException = null;
+        RubyModule kernelModule = getRuntime().getKernel();
+        if (DEBUG) {
+            System.out.println("thread " + Thread.currentThread() + " before propagating exception: " + killed);
+        }
+        kernelModule.callMethod(context, "raise", raiseException);
+    }
 }
