@@ -48,6 +48,7 @@ import org.jruby.RubyString;
 import org.jruby.anno.JRubyClass;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.runtime.ObjectAllocator;
+import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.io.BadDescriptorException;
 import org.jruby.util.io.ChannelStream;
@@ -76,6 +77,10 @@ public class RubyBasicSocket extends RubyIO {
     }
     
     protected void initSocket(ChannelDescriptor descriptor) {
+        // make sure descriptor is registered
+        registerDescriptor(descriptor);
+        
+        // continue with normal initialization
         openFile = new OpenFile();
         
         try {
@@ -89,21 +94,62 @@ public class RubyBasicSocket extends RubyIO {
     }
 
     @Override
-    public IRubyObject close_write() {
-        try {
-            ((SocketChannel)openFile.getWriteStream().getDescriptor().getChannel()).socket().shutdownOutput();
-            openFile.getWriteStream().fclose();
-        } catch (BadDescriptorException ex) {
-            throw getRuntime().newErrnoEBADFError();
-        } catch (IOException e) {
-            throw getRuntime().newIOError(e.getMessage());
+    public IRubyObject close_write(ThreadContext context) {
+        if (getRuntime().getSafeLevel() >= 4 && isTaint()) {
+            throw getRuntime().newSecurityError("Insecure: can't close");
+        }
+        
+        if (openFile.getPipeStream() == null && openFile.isReadable()) {
+            throw getRuntime().newIOError("closing non-duplex IO for writing");
+        }
+        
+        if (!openFile.isReadable()) {
+            close();
+        } else {
+            Channel socketChannel = openFile.getMainStream().getDescriptor().getChannel();
+            if (socketChannel instanceof SocketChannel
+                    || socketChannel instanceof DatagramChannel) {
+                try {
+                    asSocket().shutdownOutput();
+                } catch (IOException e) {
+                    throw getRuntime().newIOError(e.getMessage());
+                }
+            }
+            openFile.setPipeStream(null);
+            openFile.setMode(openFile.getMode() & ~OpenFile.WRITABLE);
         }
         return getRuntime().getNil();
     }
+    
+    @Override
+    public IRubyObject close_read(ThreadContext context) {
+        Ruby runtime = context.getRuntime();
+        if (runtime.getSafeLevel() >= 4 && isTaint()) {
+            throw getRuntime().newSecurityError("Insecure: can't close");
+        }
+        
+        if (!openFile.isWritable()) {
+            close();
+        } else {
+            Channel socketChannel = openFile.getMainStream().getDescriptor().getChannel();
+            if (socketChannel instanceof SocketChannel
+                    || socketChannel instanceof DatagramChannel) {
+                try {
+                    asSocket().shutdownInput();
+                } catch (IOException e) {
+                    throw runtime.newIOError(e.getMessage());
+                }
+            }
+            openFile.setMainStream(openFile.getPipeStream());
+            openFile.setPipeStream(null);
+            openFile.setMode(openFile.getMode() & ~OpenFile.READABLE);
+        }
+        return runtime.getNil();
+    }
 
     @JRubyMethod(name = "send", rest = true)
-    public IRubyObject write_send(IRubyObject[] args) {
-        return syswrite(args[0]);
+    public IRubyObject write_send(ThreadContext context, IRubyObject[] args) {
+        return syswrite(context, args[0]);
     }
     
     @JRubyMethod(rest = true)
@@ -354,7 +400,7 @@ public class RubyBasicSocket extends RubyIO {
     }
 
     private IRubyObject number(long s) {
-        StringBuffer result = new StringBuffer();
+        StringBuilder result = new StringBuilder();
         result.append((char) ((s>>24) &0xff)).append((char) ((s>>16) &0xff));
         result.append((char) ((s >> 8) & 0xff)).append((char) (s & 0xff));
         return getRuntime().newString(result.toString());
@@ -491,8 +537,8 @@ public class RubyBasicSocket extends RubyIO {
     }
 
     @JRubyMethod(optional = 1)
-    public IRubyObject shutdown(IRubyObject[] args) {
-        if (getRuntime().getSafeLevel() >= 4 && tainted_p().isFalse()) {
+    public IRubyObject shutdown(ThreadContext context, IRubyObject[] args) {
+        if (getRuntime().getSafeLevel() >= 4 && tainted_p(context).isFalse()) {
             throw getRuntime().newSecurityError("Insecure: can't shutdown socket");
         }
         
