@@ -39,9 +39,6 @@ import org.jruby.ast.ArgsNode;
 import org.jruby.ast.ListNode;
 import org.jruby.ast.Node;
 import org.jruby.ast.executable.Script;
-import org.jruby.compiler.JITCompiler;
-import org.jruby.evaluator.AssignmentVisitor;
-import org.jruby.evaluator.ASTInterpreter;
 import org.jruby.exceptions.JumpException;
 import org.jruby.exceptions.RaiseException;
 import org.jruby.internal.runtime.JumpTarget;
@@ -129,7 +126,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
         Ruby runtime = context.getRuntime();
 
         if (runtime.getInstanceConfig().getCompileMode().shouldJIT()) {
-            JITCompiler.runJIT(this, runtime, context, name);
+            runtime.getJITCompiler().runJIT(this, context, name);
         }
         
         if (jitCompiledScript != null && !runtime.hasEventHooks()) {
@@ -140,7 +137,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, args, block);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -171,11 +168,13 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
                 traceCall(context, runtime, name);
             }
 
-            return ASTInterpreter.eval(runtime, context, body, self, block);
+            return body.interpret(runtime, context, self, block);
         } catch (JumpException.ReturnJump rj) {
-            return handleReturn(rj);
+            return handleReturn(context, rj);
         } catch (JumpException.RedoJump rj) {
             return handleRedo(runtime);
+        } catch (StackOverflowError sfe) {
+            throw runtime.newSystemStackError("stack level too deep");
         } finally {
             if (runtime.hasEventHooks()) {
                 traceReturn(context, runtime, name);
@@ -194,7 +193,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
                 
                 return jitCompiledScript.__file__(context, self, args, Block.NULL_BLOCK);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -215,7 +214,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, Block.NULL_BLOCK);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -236,7 +235,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, block);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -257,7 +256,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, arg0, Block.NULL_BLOCK);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -278,7 +277,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, arg0, block);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -299,7 +298,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, arg0, arg1, Block.NULL_BLOCK);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -320,7 +319,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, arg0, arg1, block);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -341,7 +340,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, arg0, arg1, arg2, Block.NULL_BLOCK);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -362,7 +361,7 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
 
                 return jitCompiledScript.__file__(context, self, arg0, arg1, arg2, block);
             } catch (JumpException.ReturnJump rj) {
-                return handleReturn(rj);
+                return handleReturn(context, rj);
             } catch (JumpException.RedoJump rj) {
                 return handleRedo(runtime);
             } finally {
@@ -386,17 +385,6 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
         }
         jitCallConfig.post(context);
     }
-    
-    private IRubyObject handleReturn(JumpException.ReturnJump rj) {
-        if (rj.getTarget() == this) {
-            return (IRubyObject) rj.getValue();
-        }
-        throw rj;
-    }
-
-    private IRubyObject handleRedo(Ruby runtime) throws RaiseException {
-        throw runtime.newLocalJumpError("redo", runtime.getNil(), "unexpected redo");
-    }
 
     private void prepareArguments(ThreadContext context, Ruby runtime, IRubyObject[] args) {
         if (requiredArgsCount > args.length) {
@@ -435,13 +423,13 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
             int j = 0;
             for (int i = requiredArgsCount; i < args.length && j < optArgs.size(); i++, j++) {
                 // in-frame EvalState should already have receiver set as self, continue to use it
-                AssignmentVisitor.assign(runtime, context, context.getFrameSelf(), optArgs.get(j), args[i], Block.NULL_BLOCK, true);
+                optArgs.get(j).assign(runtime, context, context.getFrameSelf(), args[i], Block.NULL_BLOCK, true);
                 givenArgsCount++;
             }
    
             // assign the default values, adding to the end of allArgs
             for (int i = 0; j < optArgs.size(); i++, j++) {
-                ASTInterpreter.eval(runtime, context, optArgs.get(j), context.getFrameSelf(), Block.NULL_BLOCK);
+                optArgs.get(j).interpret(runtime, context, context.getFrameSelf(), Block.NULL_BLOCK);
             }
         }
         
@@ -473,15 +461,14 @@ public final class DefaultMethod extends DynamicMethod implements JumpTarget {
     }
 
     private void traceReturn(ThreadContext context, Ruby runtime, String name) {
-        Frame frame = context.getPreviousFrame();
-
-        runtime.callEventHooks(context, EventHook.RUBY_EVENT_RETURN, frame.getFile(), frame.getLine(), name, getImplementationClass());
+        runtime.callEventHooks(context, EventHook.RUBY_EVENT_RETURN, context.getFile(), context.getLine(), name, getImplementationClass());
     }
     
     private void traceCall(ThreadContext context, Ruby runtime, String name) {
         runtime.callEventHooks(context, EventHook.RUBY_EVENT_CALL, position.getFile(), position.getStartLine(), name, getImplementationClass());
     }
 
+    @Override
     public Arity getArity() {
         return argsNode.getArity();
     }
