@@ -37,6 +37,7 @@ import java.util.Comparator;
 import java.util.HashMap;
 import java.util.List;
 import org.jruby.Ruby;
+import org.jruby.RubyInstanceConfig;
 import org.jruby.parser.StaticScope;
 import org.objectweb.asm.ClassWriter;
 import org.objectweb.asm.MethodVisitor;
@@ -122,10 +123,6 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
     private final static String COMPILED_CALL_SIG_THREE = sig(IRubyObject.class,
             params(ThreadContext.class, IRubyObject.class, RubyModule.class, String.class, IRubyObject.class, IRubyObject.class, IRubyObject.class));
     
-    /** The super constructor signature for compile Ruby method handles. */
-    private final static String COMPILED_SUPER_SIG = 
-            sig(Void.TYPE, RubyModule.class, Arity.class, Visibility.class, StaticScope.class, Object.class, CallConfiguration.class);
-    
     /** The super constructor signature for Java-based method handles. */
     private final static String JAVA_SUPER_SIG = sig(Void.TYPE, params(RubyModule.class, Visibility.class));
     
@@ -179,6 +176,19 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         }
     }
 
+    /**
+     * Use code generation to provide a method handle for a compiled Ruby method.
+     * 
+     * @see org.jruby.internal.runtime.methods.MethodFactory#getCompiledMethod
+     */
+    public DynamicMethod getCompiledMethodLazily(
+            RubyModule implementationClass, String method, Arity arity, 
+            Visibility visibility, StaticScope scope, Object scriptObject, CallConfiguration callConfig) {
+        return new CompiledMethod.LazyCompiledMethod(implementationClass, method, arity, visibility, scope, scriptObject, callConfig,
+                new InvocationMethodFactory(classLoader));
+    }
+            
+
     /**
      * Use code generation to provide a method handle for a compiled Ruby method.
      * 
@@ -214,11 +224,7 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
                         // check arity
                         mv.aload(1);
                         mv.invokevirtual(p(ThreadContext.class), "getRuntime", sig(Ruby.class));
-                        if (scope.getRequiredArgs() == 0) {
-                            mv.getstatic(p(IRubyObject.class), "NULL_ARRAY", ci(IRubyObject[].class));
-                        } else {
-                            mv.aload(5);
-                        }
+                        mv.aload(5);
                         mv.pushInt(scope.getRequiredArgs());
                         mv.pushInt(scope.getRequiredArgs());
                         mv.invokestatic(p(Arity.class), "checkArgumentCount", sig(int.class, Ruby.class, IRubyObject[].class, int.class, int.class));
@@ -274,10 +280,6 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
                         }
                     }
 
-                    // store null for result var
-                    mv.aconst_null();
-                    mv.astore(8);
-
                     Label tryBegin = new Label();
                     Label tryEnd = new Label();
                     Label doFinally = new Label();
@@ -286,10 +288,14 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
                     Label catchReturnJump = new Label();
                     Label catchRedoJump = new Label();
 
-                    mv.trycatch(tryBegin, tryEnd, catchReturnJump, p(JumpException.ReturnJump.class));
+                    if (callConfig != CallConfiguration.FRAME_ONLY) {
+                        mv.trycatch(tryBegin, tryEnd, catchReturnJump, p(JumpException.ReturnJump.class));
+                    }
                     mv.trycatch(tryBegin, tryEnd, catchRedoJump, p(JumpException.RedoJump.class));
                     mv.trycatch(tryBegin, tryEnd, doFinally, null);
-                    mv.trycatch(catchReturnJump, doReturnFinally, doFinally, null);
+                    if (callConfig != CallConfiguration.FRAME_ONLY) {
+                        mv.trycatch(catchReturnJump, doReturnFinally, doFinally, null);
+                    }
                     mv.trycatch(catchRedoJump, doRedoFinally, doFinally, null);
                     mv.label(tryBegin);
                     {
@@ -322,20 +328,24 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
                     }
 
                     // return jump handling
-                    mv.label(catchReturnJump);
-                    {
-                        mv.aload(0);
-                        mv.swap();
-                        mv.invokevirtual(COMPILED_SUPER_CLASS, "handleReturnJump", sig(IRubyObject.class, JumpException.ReturnJump.class));
-                        mv.label(doReturnFinally);
-                        
-                        // finally
-                        if (!callConfig.isNoop()) {
-                            invokeCallConfigPost(mv, COMPILED_SUPER_CLASS, callConfig);
+                    if (callConfig != CallConfiguration.FRAME_ONLY) {
+                        mv.label(catchReturnJump);
+                        {
+                            mv.aload(0);
+                            mv.swap();
+                            mv.aload(1);
+                            mv.swap();
+                            mv.invokevirtual(COMPILED_SUPER_CLASS, "handleReturn", sig(IRubyObject.class, ThreadContext.class, JumpException.ReturnJump.class));
+                            mv.label(doReturnFinally);
+
+                            // finally
+                            if (!callConfig.isNoop()) {
+                                invokeCallConfigPost(mv, COMPILED_SUPER_CLASS, callConfig);
+                            }
+
+                            // return result if we're still good
+                            mv.areturn();
                         }
-                        
-                        // return result if we're still good
-                        mv.areturn();
                     }
 
                     // redo jump handling
@@ -375,9 +385,9 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
                     generatedClass = endCall(cw,mv,mname);
                 }
 
-                return (DynamicMethod)generatedClass
-                        .getConstructor(RubyModule.class, Arity.class, Visibility.class, StaticScope.class, Object.class, CallConfiguration.class)
-                        .newInstance(implementationClass, arity, visibility, scope, scriptObject, callConfig);
+                CompiledMethod compiledMethod = (CompiledMethod)generatedClass.newInstance();
+                compiledMethod.init(implementationClass, arity, visibility, scope, scriptObject, callConfig);
+                return compiledMethod;
             } catch(Exception e) {
                 e.printStackTrace();
                 throw implementationClass.getRuntime().newLoadError(e.getMessage());
@@ -397,11 +407,7 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         
         if (DEBUG) out.println("Binding multiple: " + desc1.declaringClassName + "." + javaMethodName);
         
-        String generatedClassName = CodegenUtils.getAnnotatedBindingClassName(javaMethodName, desc1.declaringClassName, desc1.isStatic, desc1.actualRequired, desc1.optional, true);
-        String generatedClassPath = generatedClassName.replace('.', '/');
-        
         synchronized (classLoader) {
-
             try {
                 Class c = getAnnotatedMethodClass(descs);
                 int min = Integer.MAX_VALUE;
@@ -464,12 +470,17 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
             // simple path, no multimethod
             return getAnnotatedMethodClass(descs.get(0));
         }
+        
         JavaMethodDescriptor desc1 = descs.get(0);
         String javaMethodName = desc1.name;
         
         if (DEBUG) out.println("Binding multiple: " + desc1.declaringClassName + "." + javaMethodName);
         
         String generatedClassName = CodegenUtils.getAnnotatedBindingClassName(javaMethodName, desc1.declaringClassName, desc1.isStatic, desc1.actualRequired, desc1.optional, true);
+        if (RubyInstanceConfig.FULL_TRACE_ENABLED) {
+            // in debug mode we append _DBG to class name to force it to regenerate (or use pre-generated debug version)
+            generatedClassName += "_DBG";
+        }
         String generatedClassPath = generatedClassName.replace('.', '/');
         
         synchronized (classLoader) {
@@ -544,7 +555,11 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
                         }
                         break;
                     case 3:
-                        superClass = p(JavaMethod.JavaMethodOneOrTwoOrThree.class);
+                        if (block) {
+                            superClass = p(JavaMethod.JavaMethodOneOrTwoOrThreeBlock.class);
+                        } else {
+                            superClass = p(JavaMethod.JavaMethodOneOrTwoOrThree.class);
+                        }
                         break;
                     }
                     break;
@@ -637,6 +652,10 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         String javaMethodName = desc.name;
         
         String generatedClassName = CodegenUtils.getAnnotatedBindingClassName(javaMethodName, desc.declaringClassName, desc.isStatic, desc.actualRequired, desc.optional, false);
+        if (RubyInstanceConfig.FULL_TRACE_ENABLED) {
+            // in debug mode we append _DBG to class name to force it to regenerate (or use pre-generated debug version)
+            generatedClassName += "_DBG";
+        }
         String generatedClassPath = generatedClassName.replace('.', '/');
         
         synchronized (classLoader) {
@@ -927,18 +946,12 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
     }
 
     private ClassWriter createCompiledCtor(String namePath, String sup) throws Exception {
-        ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
-        cw.visit(V1_4, ACC_PUBLIC + ACC_SUPER, namePath, null, sup, null);
-        MethodVisitor mv = cw.visitMethod(ACC_PUBLIC, "<init>", COMPILED_SUPER_SIG, null, null);
+        ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS | ClassWriter.COMPUTE_FRAMES);
+        cw.visit(RubyInstanceConfig.JAVA_VERSION, ACC_PUBLIC + ACC_SUPER, namePath, null, sup, null);
+        MethodVisitor mv = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
         mv.visitCode();
         mv.visitVarInsn(ALOAD, 0);
-        mv.visitVarInsn(ALOAD, 1);
-        mv.visitVarInsn(ALOAD, 2);
-        mv.visitVarInsn(ALOAD, 3);
-        mv.visitVarInsn(ALOAD, 4);
-        mv.visitVarInsn(ALOAD, 5);
-        mv.visitVarInsn(ALOAD, 6);
-        mv.visitMethodInsn(INVOKESPECIAL, sup, "<init>", COMPILED_SUPER_SIG);
+        mv.visitMethodInsn(INVOKESPECIAL, sup, "<init>", "()V");
         Label line = new Label();
         mv.visitLineNumber(0, line);
         mv.visitInsn(RETURN);
@@ -948,8 +961,8 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
     }
 
     private ClassWriter createJavaMethodCtor(String namePath, String sup) throws Exception {
-        ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
-        cw.visit(V1_4, ACC_PUBLIC + ACC_SUPER, namePath, null, sup, null);
+        ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS | ClassWriter.COMPUTE_FRAMES);
+        cw.visit(RubyInstanceConfig.JAVA_VERSION, ACC_PUBLIC + ACC_SUPER, namePath, null, sup, null);
         MethodVisitor mv = cw.visitMethod(ACC_PUBLIC, "<init>", JAVA_SUPER_SIG, null, null);
         mv.visitCode();
         mv.visitVarInsn(ALOAD, 0);
@@ -965,8 +978,8 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
     }
 
     private ClassWriter createIndexedJavaMethodCtor(String namePath, String sup) throws Exception {
-        ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
-        cw.visit(V1_4, ACC_PUBLIC + ACC_SUPER, namePath, null, sup, null);
+        ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS | ClassWriter.COMPUTE_FRAMES);
+        cw.visit(RubyInstanceConfig.JAVA_VERSION, ACC_PUBLIC + ACC_SUPER, namePath, null, sup, null);
         MethodVisitor mv = cw.visitMethod(ACC_PUBLIC, "<init>", JAVA_INDEXED_SUPER_SIG, null, null);
         mv.visitCode();
         mv.visitVarInsn(ALOAD, 0);
@@ -982,30 +995,6 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         return cw;
     }
 
-    private void handleRedo(Label tryRedoJump, SkinnyMethodAdapter mv, Label tryFinally) {
-        mv.label(tryRedoJump);
-
-        // clear the redo
-        mv.pop();
-
-        // get runtime, dup it
-        mv.aload(1);
-        mv.invokevirtual(p(ThreadContext.class), "getRuntime", sig(Ruby.class));
-        mv.invokevirtual(p(Ruby.class), "newRedoLocalJumpError", sig(RaiseException.class));
-        mv.go_to(tryFinally);
-    }
-
-    private void handleReturn(Label catchReturnJump, SkinnyMethodAdapter mv, Label doFinally, Label normalExit, String typePath) {
-        mv.label(catchReturnJump);
-
-        mv.aload(0);
-        mv.swap();
-        mv.invokevirtual(typePath, "handleReturnJump", sig(IRubyObject.class, JumpException.ReturnJump.class));
-
-        mv.astore(8);
-        mv.go_to(normalExit);
-    }
-
     private void invokeCallConfigPost(SkinnyMethodAdapter mv, String superClass, CallConfiguration callConfig) {
         if (callConfig != CallConfiguration.NO_FRAME_NO_SCOPE) {
             mv.aload(0);
@@ -1225,7 +1214,7 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
     protected Class endClass(ClassWriter cw, String name) {
         cw.visitEnd();
         byte[] code = cw.toByteArray();
-        CheckClassAdapter.verify(new ClassReader(code), false, new PrintWriter(System.err));
+        if (DEBUG) CheckClassAdapter.verify(new ClassReader(code), false, new PrintWriter(System.err));
          
         return classLoader.defineClass(name, code);
     }
@@ -1331,6 +1320,10 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         if (!callConfig.isNoop()) {
             invokeCallConfigPre(method, superClass, specificArity, block, callConfig);
         }
+        
+        if (RubyInstanceConfig.FULL_TRACE_ENABLED) {
+            invokeCCallTrace(method);
+        }
 
         Label tryBegin = new Label();
         Label tryEnd = new Label();
@@ -1339,11 +1332,6 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         Label catchRedoJump = new Label();
 
         if (!callConfig.isNoop()) {
-            // only need return/redo handling for block-receiving methods
-            if (block) {
-                method.trycatch(tryBegin, tryEnd, catchRedoJump, p(JumpException.RedoJump.class));
-                method.trycatch(catchRedoJump, doRedoFinally, doFinally, null);
-            }
             method.trycatch(tryBegin, tryEnd, doFinally, null);
         }
         
@@ -1367,6 +1355,10 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         
         // normal finally and exit
         {
+            if (RubyInstanceConfig.FULL_TRACE_ENABLED) {
+                invokeCReturnTrace(method);
+            }
+            
             if (!callConfig.isNoop()) {
                 invokeCallConfigPost(method, superClass, callConfig);
             }
@@ -1377,33 +1369,13 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
         
         // these are only needed if we have a non-noop call config
         if (!callConfig.isNoop()) {
-            // we only need redo jump handling if we've got a block, since it would only come from yields
-            if (block) {
-                // redo jump handling
-                method.label(catchRedoJump);
-                {
-                    // clear the redo
-                    method.pop();
-
-                    // get runtime, create jump error, and throw it
-                    method.aload(1);
-                    method.invokevirtual(p(ThreadContext.class), "getRuntime", sig(Ruby.class));
-                    method.invokevirtual(p(Ruby.class), "newRedoLocalJumpError", sig(RaiseException.class));
-
-                    // finally
-                    method.label(doRedoFinally);
-                    if (!callConfig.isNoop()) {
-                        invokeCallConfigPost(method, superClass, callConfig);
-                    }
-
-                    // throw redo error if we're still good
-                    method.athrow();
-                }
-            }
-
             // finally handling for abnormal exit
             {
                 method.label(doFinally);
+                
+                if (RubyInstanceConfig.FULL_TRACE_ENABLED) {
+                    invokeCReturnTrace(method);
+                }
 
                 //call post method stuff (exception raised)
                 if (!callConfig.isNoop()) {
@@ -1415,4 +1387,18 @@ public class InvocationMethodFactory extends MethodFactory implements Opcodes {
             }
         }
     }
+    
+    private void invokeCCallTrace(SkinnyMethodAdapter method) {
+        method.aload(0); // method itself
+        method.aload(1); // ThreadContext
+        method.aload(4); // invoked name
+        method.invokevirtual(p(JavaMethod.class), "callTrace", sig(void.class, ThreadContext.class, String.class));
+    }
+    
+    private void invokeCReturnTrace(SkinnyMethodAdapter method) {
+        method.aload(0); // method itself
+        method.aload(1); // ThreadContext
+        method.aload(4); // invoked name
+        method.invokevirtual(p(JavaMethod.class), "returnTrace", sig(void.class, ThreadContext.class, String.class));
+    }
 }
