@@ -31,6 +31,7 @@ import org.jruby.internal.runtime.JumpTarget;
 import org.jruby.parser.StaticScope;
 import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
+import org.jruby.runtime.EventHook;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
@@ -43,8 +44,7 @@ public abstract class JavaMethod extends DynamicMethod implements JumpTarget, Cl
     private Class[] argumentTypes;
     private String javaName;
     private boolean isSingleton;
-    protected final int methodIndex;
-    protected final StaticScope staticScope;
+    protected StaticScope staticScope;
     
     public static abstract class JavaMethodNoBlock extends JavaMethod {
         public JavaMethodNoBlock(RubyModule implementationClass, Visibility visibility) {
@@ -428,6 +428,34 @@ public abstract class JavaMethod extends DynamicMethod implements JumpTarget, Cl
         }
     }
     
+    public static abstract class JavaMethodOneOrTwoOrThreeBlock extends JavaMethod {
+        public JavaMethodOneOrTwoOrThreeBlock(RubyModule implementationClass, Visibility visibility) {
+            super(implementationClass, visibility);
+        }
+        public JavaMethodOneOrTwoOrThreeBlock(RubyModule implementationClass, Visibility visibility, CallConfiguration callConfig, StaticScope staticScope, Arity arity) {
+            super(implementationClass, visibility, callConfig, staticScope, arity);
+        }
+        public JavaMethodOneOrTwoOrThreeBlock(RubyModule implementationClass, Visibility visibility, int methodIndex) {
+            super(implementationClass, visibility, methodIndex);
+        }
+        
+        public abstract IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, Block block);
+        
+        public abstract IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, IRubyObject arg1, Block block);
+        
+        public abstract IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, IRubyObject arg1, IRubyObject arg2, Block block);
+        
+        public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block) {
+            switch (args.length) {
+            case 0: throw context.getRuntime().newArgumentError(0, 1);
+            case 1: return call(context, self, clazz, name, args[0], block);
+            case 2: return call(context, self, clazz, name, args[0], args[1], block);
+            case 3: return call(context, self, clazz, name, args[0], args[1], args[2], block);
+            default: throw context.getRuntime().newArgumentError(args.length, 3);
+            }
+        }
+    }
+    
     public static abstract class JavaMethodTwo extends JavaMethod {
         public JavaMethodTwo(RubyModule implementationClass, Visibility visibility) {
             super(implementationClass, visibility);
@@ -542,13 +570,11 @@ public abstract class JavaMethod extends DynamicMethod implements JumpTarget, Cl
 
     public JavaMethod(RubyModule implementationClass, Visibility visibility) {
         super(implementationClass, visibility, CallConfiguration.FRAME_ONLY);
-        this.methodIndex = -1;
         this.staticScope = null;
     }
 
     public JavaMethod(RubyModule implementationClass, Visibility visibility, CallConfiguration callConfig, StaticScope staticScope, Arity arity) {
         super(implementationClass, visibility, callConfig);
-        this.methodIndex = -1;
         this.staticScope = staticScope;
         this.arity = arity;
         this.arityValue = arity.getValue();
@@ -556,9 +582,17 @@ public abstract class JavaMethod extends DynamicMethod implements JumpTarget, Cl
 
     public JavaMethod(RubyModule implementationClass, Visibility visibility, int methodIndex) {
         super(implementationClass, visibility, CallConfiguration.FRAME_ONLY);
-        this.methodIndex = methodIndex;
         this.staticScope = null;
     }
+    
+    protected JavaMethod() {}
+    
+    public void init(RubyModule implementationClass, Arity arity, Visibility visibility, StaticScope staticScope, CallConfiguration callConfig) {
+        this.staticScope = staticScope;
+        this.arity = arity;
+        this.arityValue = arity.getValue();
+        super.init(implementationClass, visibility, callConfig);
+    }
 
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block);
     
@@ -611,12 +645,12 @@ public abstract class JavaMethod extends DynamicMethod implements JumpTarget, Cl
         context.postMethodBacktraceAndScope();
     }
     
-    protected IRubyObject handleReturnJump(ReturnJump rj) {
-        if (rj.getTarget() == this) {
-            return (IRubyObject)rj.getValue();
-        } else {
-            throw rj;
-        }
+    protected final void callTrace(ThreadContext context, String name) {
+        context.trace(EventHook.RUBY_EVENT_C_CALL, name, getImplementationClass());
+    }
+    
+    protected final void returnTrace(ThreadContext context, String name) {
+        context.trace(EventHook.RUBY_EVENT_C_CALL, name, getImplementationClass());
     }
     
     public void setArity(Arity arity) {
