@@ -8,6 +8,7 @@ import org.jruby.RubyClass;
 import org.jruby.RubyException;
 import org.jruby.RubyFixnum;
 import org.jruby.RubyHash;
+import org.jruby.RubyInstanceConfig;
 import org.jruby.RubyKernel;
 import org.jruby.RubyLocalJumpError;
 import org.jruby.RubyMatchData;
@@ -29,6 +30,7 @@ import org.jruby.parser.LocalStaticScope;
 import org.jruby.parser.StaticScope;
 import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
+import org.jruby.runtime.BlockBody;
 import org.jruby.runtime.CallSite;
 import org.jruby.runtime.CallType;
 import org.jruby.runtime.CallbackFactory;
@@ -58,33 +60,32 @@ public class RuntimeHelpers {
         return factory.getBlockCallback(closureMethod, scriptObject);
     }
     
-    public static Block createBlock(ThreadContext context, IRubyObject self, int arity, 
-            String[] staticScopeNames, CompiledBlockCallback callback, boolean hasMultipleArgsHead, int argsNodeType, boolean light) {
+    public static BlockBody createCompiledBlockBody(ThreadContext context, Object scriptObject, String closureMethod, int arity, 
+            String[] staticScopeNames, boolean hasMultipleArgsHead, int argsNodeType, boolean light) {
         StaticScope staticScope = 
             new BlockStaticScope(context.getCurrentScope().getStaticScope(), staticScopeNames);
         staticScope.determineModule();
         
         if (light) {
-            return CompiledBlockLight.newCompiledClosureLight(
-                    context,
-                    self,
-                    Arity.createArity(arity),
-                    staticScope,
-                    callback,
-                    hasMultipleArgsHead,
-                    argsNodeType);
+            return CompiledBlockLight.newCompiledBlockLight(
+                    Arity.createArity(arity), staticScope,
+                    createBlockCallback(context.getRuntime(), scriptObject, closureMethod),
+                    hasMultipleArgsHead, argsNodeType);
         } else {
-            return CompiledBlock.newCompiledClosure(
-                    context,
-                    self,
-                    Arity.createArity(arity),
-                    staticScope,
-                    callback,
-                    hasMultipleArgsHead,
-                    argsNodeType);
+            return CompiledBlock.newCompiledBlock(
+                    Arity.createArity(arity), staticScope,
+                    createBlockCallback(context.getRuntime(), scriptObject, closureMethod),
+                    hasMultipleArgsHead, argsNodeType);
         }
     }
     
+    public static Block createBlock(ThreadContext context, IRubyObject self, BlockBody body) {
+        return CompiledBlock.newCompiledClosure(
+                context,
+                self,
+                body);
+    }
+    
     public static IRubyObject runBeginBlock(ThreadContext context, IRubyObject self, String[] staticScopeNames, CompiledBlockCallback callback) {
         StaticScope staticScope = 
             new BlockStaticScope(context.getCurrentScope().getStaticScope(), staticScopeNames);
@@ -92,7 +93,7 @@ public class RuntimeHelpers {
         
         context.preScopedBody(DynamicScope.newDynamicScope(staticScope, context.getCurrentScope()));
         
-        Block block = CompiledBlock.newCompiledClosure(context, self, Arity.createArity(0), staticScope, callback, false, Block.ZERO_ARGS);
+        Block block = CompiledBlock.newCompiledClosure(context, self, Arity.createArity(0), staticScope, callback, false, BlockBody.ZERO_ARGS);
         
         try {
             block.yield(context, null);
@@ -229,10 +230,8 @@ public class RuntimeHelpers {
         if (methodIndex == MethodIndex.METHOD_MISSING) {
             return RubyKernel.method_missing(context, self, args, block);
         }
-
-        IRubyObject[] newArgs = new IRubyObject[args.length + 1];
-        System.arraycopy(args, 0, newArgs, 1, args.length);
-        newArgs[0] = context.getRuntime().newSymbol(name);
+        
+        IRubyObject[] newArgs = prepareMethodMissingArgs(args, context, name);
 
         return receiver.callMethod(context, "method_missing", newArgs, block);
     }
@@ -294,7 +293,7 @@ public class RuntimeHelpers {
         return (RubyArray) value;
     }
 
-    public static RubyArray ensureMultipleAssignableRubyArray(Ruby runtime, IRubyObject value, boolean masgnHasHead) {
+    public static RubyArray ensureMultipleAssignableRubyArray(IRubyObject value, Ruby runtime, boolean masgnHasHead) {
         if (!(value instanceof RubyArray)) {
             value = ArgsUtil.convertToRubyArray(runtime, value, masgnHasHead);
         }
@@ -446,37 +445,50 @@ public class RuntimeHelpers {
         if (block.getProcObject() != null) {
             blockArg = block.getProcObject();
         } else {
-            blockArg = runtime.newProc(Block.Type.PROC, block);
+            blockArg = runtime.newBlockPassProc(Block.Type.PROC, block);
             blockArg.getBlock().type = Block.Type.PROC;
         }
         
         return blockArg;
     }
     
-    public static Block getBlockFromBlockPassBody(IRubyObject proc, Block currentBlock) {
-        Ruby runtime = proc.getRuntime();
-
+    public static Block getBlockFromBlockPassBody(Ruby runtime, IRubyObject proc, Block currentBlock) {
         // No block from a nil proc
         if (proc.isNil()) return Block.NULL_BLOCK;
 
         // If not already a proc then we should try and make it one.
         if (!(proc instanceof RubyProc)) {
-            proc = TypeConverter.convertToType(proc, runtime.getProc(), 0, "to_proc", false);
-
-            if (!(proc instanceof RubyProc)) {
-                throw runtime.newTypeError("wrong argument type "
-                        + proc.getMetaClass().getName() + " (expected Proc)");
-            }
+            proc = coerceProc(proc, runtime);
         }
 
+        return getBlockFromProc(currentBlock, proc);
+    }
+
+    private static IRubyObject coerceProc(IRubyObject proc, Ruby runtime) throws RaiseException {
+        proc = TypeConverter.convertToType(proc, runtime.getProc(), 0, "to_proc", false);
+
+        if (!(proc instanceof RubyProc)) {
+            throw runtime.newTypeError("wrong argument type " + proc.getMetaClass().getName() + " (expected Proc)");
+        }
+        return proc;
+    }
+
+    private static Block getBlockFromProc(Block currentBlock, IRubyObject proc) {
         // TODO: Add safety check for taintedness
         if (currentBlock != null && currentBlock.isGiven()) {
             RubyProc procObject = currentBlock.getProcObject();
             // The current block is already associated with proc.  No need to create a new one
-            if (procObject != null && procObject == proc) return currentBlock;
+            if (procObject != null && procObject == proc) {
+                return currentBlock;
+            }
         }
 
-        return ((RubyProc) proc).getBlock();
+        return ((RubyProc) proc).getBlock();       
+    }
+    
+    public static Block getBlockFromBlockPassBody(IRubyObject proc, Block currentBlock) {
+        return getBlockFromBlockPassBody(proc.getRuntime(), proc, currentBlock);
+
     }
     
     public static IRubyObject backref(ThreadContext context) {
@@ -532,19 +544,17 @@ public class RuntimeHelpers {
         return context.returnJump(result);
     }
     
-    public static IRubyObject breakJumpInWhile(JumpException.BreakJump bj, Block aBlock) {
+    public static IRubyObject breakJumpInWhile(JumpException.BreakJump bj, Block aBlock, ThreadContext context) {
         // JRUBY-530, while case
-        if (bj.getTarget() == aBlock.getBody()) {
-            bj.setTarget(null);
-            
-            throw bj;
+        if (bj.getTarget() == context.getFrameJumpTarget()) {
+            return (IRubyObject) bj.getValue();
         }
 
-        return (IRubyObject) bj.getValue();
+        throw bj;
     }
     
-    public static IRubyObject breakJump(IRubyObject value) {
-        throw new JumpException.BreakJump(null, value);
+    public static IRubyObject breakJump(ThreadContext context, IRubyObject value) {
+        throw new JumpException.BreakJump(context.getFrameJumpTarget(), value);
     }
     
     public static IRubyObject breakLocalJumpError(Ruby runtime, IRubyObject value) {
@@ -666,6 +676,46 @@ public class RuntimeHelpers {
         return new IRubyObject[] {one, two, three, four, five};
     }
     
+    public static String[] constructStringArray(String one) {
+        return new String[] {one};
+    }
+    
+    public static String[] constructStringArray(String one, String two) {
+        return new String[] {one, two};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three) {
+        return new String[] {one, two, three};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three, String four) {
+        return new String[] {one, two, three, four};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three, String four, String five) {
+        return new String[] {one, two, three, four, five};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three, String four, String five, String six) {
+        return new String[] {one, two, three, four, five, six};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three, String four, String five, String six, String seven) {
+        return new String[] {one, two, three, four, five, six, seven};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three, String four, String five, String six, String seven, String eight) {
+        return new String[] {one, two, three, four, five, six, seven, eight};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three, String four, String five, String six, String seven, String eight, String nine) {
+        return new String[] {one, two, three, four, five, six, seven, eight, nine};
+    }
+    
+    public static String[] constructStringArray(String one, String two, String three, String four, String five, String six, String seven, String eight, String nine, String ten) {
+        return new String[] {one, two, three, four, five, six, seven, eight, nine, ten};
+    }
+    
     public static final int MAX_SPECIFIC_ARITY_HASH = 3;
     
     public static RubyHash constructHash(Ruby runtime, IRubyObject key1, IRubyObject value1) {
@@ -787,7 +837,8 @@ public class RuntimeHelpers {
     }
     
     public static IRubyObject postOpAsgnWithOrAnd(IRubyObject receiver, IRubyObject value, ThreadContext context, CallSite varAsgnSite) {
-        return varAsgnSite.call(context, receiver, value);
+        varAsgnSite.call(context, receiver, value);
+        return value;
     }
     
     public static IRubyObject opAsgnWithMethod(ThreadContext context, IRubyObject receiver, IRubyObject arg, CallSite varSite, CallSite opSite, CallSite opAsgnSite) {
@@ -940,8 +991,12 @@ public class RuntimeHelpers {
     private static DynamicMethod constructNormalMethod(String name, Visibility visibility, MethodFactory factory, RubyModule containingClass, String javaName, int arity, StaticScope scope, Object scriptObject, CallConfiguration callConfig) {
         DynamicMethod method;
 
-        if (name.equals("initialize") || visibility == Visibility.MODULE_FUNCTION) {
-            method = factory.getCompiledMethod(containingClass, javaName, Arity.createArity(arity), Visibility.PRIVATE, scope, scriptObject, callConfig);
+        if (name.equals("initialize") || name.equals("initialize_copy") || visibility == Visibility.MODULE_FUNCTION) {
+            visibility = Visibility.PRIVATE;
+        }
+        
+        if (RubyInstanceConfig.LAZYHANDLES_COMPILE) {
+            method = factory.getCompiledMethodLazily(containingClass, javaName, Arity.createArity(arity), visibility, scope, scriptObject, callConfig);
         } else {
             method = factory.getCompiledMethod(containingClass, javaName, Arity.createArity(arity), visibility, scope, scriptObject, callConfig);
         }
@@ -950,7 +1005,7 @@ public class RuntimeHelpers {
     }
 
     private static DynamicMethod constructSingletonMethod(MethodFactory factory, RubyClass rubyClass, String javaName, int arity, StaticScope scope, Object scriptObject, CallConfiguration callConfig) {
-        return factory.getCompiledMethod(rubyClass, javaName, Arity.createArity(arity), Visibility.PUBLIC, scope, scriptObject, callConfig);
+        return factory.getCompiledMethodLazily(rubyClass, javaName, Arity.createArity(arity), Visibility.PUBLIC, scope, scriptObject, callConfig);
     }
 
     private static StaticScope creatScopeForClass(ThreadContext context, String[] scopeNames, int required, int optional, int rest) {
@@ -999,4 +1054,28 @@ public class RuntimeHelpers {
         
         return rubyClass;
     }
+
+    private static IRubyObject[] prepareMethodMissingArgs(IRubyObject[] args, ThreadContext context, String name) {
+        IRubyObject[] newArgs = new IRubyObject[args.length + 1];
+        System.arraycopy(args, 0, newArgs, 1, args.length);
+        newArgs[0] = context.getRuntime().newSymbol(name);
+
+        return newArgs;
+    }
+    
+    public static IRubyObject arrayEntryOrNil(RubyArray array, IRubyObject nil, int index) {
+        if (index < array.getLength()) {
+            return array.entry(index);
+        } else {
+            return nil;
+        }
+    }
+    
+    public static RubyArray subarrayOrEmpty(RubyArray array, Ruby runtime, int index) {
+        if (index < array.getLength()) {
+            return createSubarray(array, index);
+        } else {
+            return RubyArray.newEmptyArray(runtime);
+        }
+    }
 }
