@@ -42,13 +42,28 @@ import org.jruby.ast.ArgumentNode;
 import org.jruby.ast.ArrayNode;
 import org.jruby.ast.AssignableNode;
 import org.jruby.ast.AttrAssignNode;
+import org.jruby.ast.AttrAssignOneArgNode;
+import org.jruby.ast.AttrAssignThreeArgNode;
+import org.jruby.ast.AttrAssignTwoArgNode;
 import org.jruby.ast.BackRefNode;
 import org.jruby.ast.BeginNode;
 import org.jruby.ast.BignumNode;
 import org.jruby.ast.BlockNode;
 import org.jruby.ast.BlockPassNode;
 import org.jruby.ast.BreakNode;
+import org.jruby.ast.CallNoArgBlockNode;
+import org.jruby.ast.CallNoArgBlockPassNode;
+import org.jruby.ast.CallNoArgNode;
 import org.jruby.ast.CallNode;
+import org.jruby.ast.CallOneArgNode;
+import org.jruby.ast.CallOneArgBlockNode;
+import org.jruby.ast.CallOneArgBlockPassNode;
+import org.jruby.ast.CallThreeArgBlockNode;
+import org.jruby.ast.CallThreeArgBlockPassNode;
+import org.jruby.ast.CallThreeArgNode;
+import org.jruby.ast.CallTwoArgBlockNode;
+import org.jruby.ast.CallTwoArgBlockPassNode;
+import org.jruby.ast.CallTwoArgNode;
 import org.jruby.ast.ClassVarAsgnNode;
 import org.jruby.ast.ClassVarDeclNode;
 import org.jruby.ast.ClassVarNode;
@@ -59,7 +74,19 @@ import org.jruby.ast.DRegexpNode;
 import org.jruby.ast.DStrNode;
 import org.jruby.ast.DotNode;
 import org.jruby.ast.EvStrNode;
+import org.jruby.ast.FCallNoArgBlockNode;
+import org.jruby.ast.FCallNoArgBlockPassNode;
+import org.jruby.ast.FCallNoArgNode;
 import org.jruby.ast.FCallNode;
+import org.jruby.ast.FCallOneArgBlockNode;
+import org.jruby.ast.FCallOneArgBlockPassNode;
+import org.jruby.ast.FCallOneArgNode;
+import org.jruby.ast.FCallThreeArgBlockNode;
+import org.jruby.ast.FCallThreeArgBlockPassNode;
+import org.jruby.ast.FCallThreeArgNode;
+import org.jruby.ast.FCallTwoArgBlockNode;
+import org.jruby.ast.FCallTwoArgBlockPassNode;
+import org.jruby.ast.FCallTwoArgNode;
 import org.jruby.ast.FalseNode;
 import org.jruby.ast.FixnumNode;
 import org.jruby.ast.FlipNode;
@@ -70,6 +97,7 @@ import org.jruby.ast.IArgumentNode;
 import org.jruby.ast.IfNode;
 import org.jruby.ast.InstAsgnNode;
 import org.jruby.ast.InstVarNode;
+import org.jruby.ast.IterNode;
 import org.jruby.ast.ListNode;
 import org.jruby.ast.LocalAsgnNode;
 import org.jruby.ast.Match2Node;
@@ -245,7 +273,7 @@ public class ParserSupport {
             case Tokens.k__LINE__:
                 throw new SyntaxException(PID.INVALID_ASSIGNMENT, lhs.getPosition(), "Can't assign to __LINE__", "__LINE__");
             case Tokens.tIDENTIFIER:
-                return currentScope.assign(value != NilImplicitNode.NIL ? union(lhs, value) : lhs.getPosition(), (String) lhs.getValue(), value);
+                return currentScope.assign(value != NilImplicitNode.NIL ? union(lhs, value) : lhs.getPosition(), (String) lhs.getValue(), makeNullNil(value));
             case Tokens.tCONSTANT:
                 if (isInDef() || isInSingle()) {
                     throw new SyntaxException(PID.DYNAMIC_CONSTANT_ASSIGNMENT, lhs.getPosition(), "dynamic constant assignment");
@@ -287,7 +315,7 @@ public class ParserSupport {
             second = ((NewlineNode) second).getNextNode();
         }
         
-        if (second == null)	return first.getPosition();
+        if (second == null) return first.getPosition();
         if (first == null) return second.getPosition();
         
         return first.getPosition().union(second.getPosition());
@@ -302,7 +330,11 @@ public class ParserSupport {
     public Node addRootNode(Node topOfAST, ISourcePosition position) {
         position = topOfAST != null ? topOfAST.getPosition() : position;
 
-        if (result.getBeginNodes().size() == 0) return new RootNode(position, result.getScope(), topOfAST);
+        if (result.getBeginNodes().size() == 0) {
+            if (topOfAST == null) topOfAST = NilImplicitNode.NIL;
+            
+            return new RootNode(position, result.getScope(), topOfAST);
+        }
         
         BlockNode newTopOfAST = new BlockNode(position);
         for (Node beginNode: result.getBeginNodes()) {
@@ -322,13 +354,7 @@ public class ParserSupport {
 
         // Reduces overhead in interp by not set position every single line we encounter. 
         if (!configuration.hasExtraPositionInformation()) {
-            while (head instanceof NewlineNode) {
-                Node nextNode = ((NewlineNode) head).getNextNode();
-                
-                if (!(nextNode instanceof NewlineNode)) break;
-                
-                head = nextNode;
-            }
+            head = compactNewlines(head);
         }
 
         if (!(head instanceof BlockNode)) {
@@ -348,7 +374,7 @@ public class ParserSupport {
     public Node getOperatorCallNode(Node firstNode, String operator) {
         checkExpression(firstNode);
 
-        return new CallNode(firstNode.getPosition(), firstNode, operator, null);
+        return new CallNoArgNode(firstNode.getPosition(), firstNode, operator);
     }
     
     public Node getOperatorCallNode(Node firstNode, String operator, Node secondNode) {
@@ -358,13 +384,13 @@ public class ParserSupport {
     public Node getOperatorCallNode(Node firstNode, String operator, Node secondNode, ISourcePosition defaultPosition) {
         if (defaultPosition != null) {
             firstNode = checkForNilNode(firstNode, defaultPosition);
-        	secondNode = checkForNilNode(secondNode, defaultPosition);
+            secondNode = checkForNilNode(secondNode, defaultPosition);
         }
         
         checkExpression(firstNode);
         checkExpression(secondNode);
         
-        return new CallNode(union(firstNode.getPosition(), secondNode.getPosition()), firstNode, operator, new ArrayNode(secondNode.getPosition(), secondNode));
+        return new CallOneArgNode(union(firstNode.getPosition(), secondNode.getPosition()), firstNode, operator, new ArrayNode(secondNode.getPosition(), secondNode));
     }
 
     public Node getMatchNode(Node firstNode, Node secondNode) {
@@ -387,7 +413,7 @@ public class ParserSupport {
     public Node aryset(Node receiver, Node index) {
         checkExpression(receiver);
 
-        return new AttrAssignNode(receiver.getPosition(), receiver, "[]=", index);
+        return new_attrassign(receiver.getPosition(), receiver, "[]=", index);
     }
 
     /**
@@ -400,7 +426,7 @@ public class ParserSupport {
     public Node attrset(Node receiver, String name) {
         checkExpression(receiver);
 
-        return new AttrAssignNode(receiver.getPosition(), receiver, name + "=", null);
+        return new_attrassign(receiver.getPosition(), receiver, name + "=", null);
     }
 
     public void backrefAssignError(Node node) {
@@ -415,7 +441,13 @@ public class ParserSupport {
     }
 
     public Node arg_add(ISourcePosition position, Node node1, Node node2) {
-        if (node1 == null) return new ArrayNode(node2 == null ? position : node2.getPosition(), node2);
+        if (node1 == null) {
+            if (node2 == null) {
+                return new ArrayNode(position, NilImplicitNode.NIL);
+            } else {
+                return new ArrayNode(node2.getPosition(), node2);
+            }
+        }
         if (node1 instanceof ArrayNode) return ((ArrayNode) node1).add(node2);
         
         return new ArgsPushNode(position, node1, node2);
@@ -436,7 +468,7 @@ public class ParserSupport {
         } else if (lhs instanceof IArgumentNode) {
             IArgumentNode invokableNode = (IArgumentNode) lhs;
             
-            invokableNode.setArgsNode(arg_add(lhs.getPosition(), invokableNode.getArgsNode(), rhs));
+            return invokableNode.setArgsNode(arg_add(lhs.getPosition(), invokableNode.getArgsNode(), rhs));
         }
         
         return newNode;
@@ -501,6 +533,18 @@ public class ParserSupport {
             warnings.warning(ID.VOID_VALUE_EXPRESSION, node.getPosition(), "void value expression");
         }
     }
+
+    private Node compactNewlines(Node head) {
+        while (head instanceof NewlineNode) {
+            Node nextNode = ((NewlineNode) head).getNextNode();
+
+            if (!(nextNode instanceof NewlineNode)) {
+                break;
+            }
+            head = nextNode;
+        }
+        return head;
+    }
     
     private boolean isExpression(Node node) {
         do {
@@ -643,9 +687,16 @@ public class ParserSupport {
 
         return false;
     }
+    
+    private Node makeNullNil(Node node) {
+        return node == null ? NilImplicitNode.NIL : node;
+    }
 
     private Node cond0(Node node) {
         checkAssignmentInCondition(node);
+        
+        Node leftNode = null;
+        Node rightNode = null;
 
         switch(node.nodeId) {
         case DREGEXPNODE: {
@@ -654,11 +705,15 @@ public class ParserSupport {
             return new Match2Node(position, node, new GlobalVarNode(position, "$_"));
         }
         case ANDNODE:
-            return new AndNode(node.getPosition(), cond0(((AndNode) node).getFirstNode()), 
-                    cond0(((AndNode) node).getSecondNode()));
+            leftNode = cond0(((AndNode) node).getFirstNode());
+            rightNode = cond0(((AndNode) node).getSecondNode());
+            
+            return new AndNode(node.getPosition(), makeNullNil(leftNode), makeNullNil(rightNode));
         case ORNODE:
-            return new OrNode(node.getPosition(), cond0(((OrNode) node).getFirstNode()), 
-                    cond0(((OrNode) node).getSecondNode()));
+            leftNode = cond0(((OrNode) node).getFirstNode());
+            rightNode = cond0(((OrNode) node).getSecondNode());
+            
+            return new OrNode(node.getPosition(), makeNullNil(leftNode), makeNullNil(rightNode));
         case DOTNODE: {
             DotNode dotNode = (DotNode) node;
             if (dotNode.isLiteral()) return node; 
@@ -666,7 +721,7 @@ public class ParserSupport {
             String label = String.valueOf("FLIP" + node.hashCode());
             currentScope.getLocalScope().addVariable(label);
             int slot = currentScope.isDefined(label);
-
+            
             return new FlipNode(node.getPosition(),
                     getFlipConditionNode(((DotNode) node).getBeginNode()),
                     getFlipConditionNode(((DotNode) node).getEndNode()),
@@ -682,7 +737,7 @@ public class ParserSupport {
     }
 
     public Node getConditionNode(Node node) {
-        if (node == null) return null;
+        if (node == null) return NilImplicitNode.NIL;
 
         if (node instanceof NewlineNode) {
             return new NewlineNode(node.getPosition(), cond0(((NewlineNode) node).getNextNode()));
@@ -706,17 +761,29 @@ public class ParserSupport {
 
         return node;
     }
-
-    public AndNode newAndNode(Node left, Node right) {
-        checkExpression(left);
-        
-        return new AndNode(union(left, right), left, right);
+    
+    public SplatNode newSplatNode(ISourcePosition position, Node node) {
+        return new SplatNode(position, makeNullNil(node));
+    }
+    
+    public ArrayNode newArrayNode(ISourcePosition position, Node firstNode) {
+        return new ArrayNode(position, makeNullNil(firstNode));
     }
 
-    public OrNode newOrNode(Node left, Node right) {
+    public AndNode newAndNode(ISourcePosition position, Node left, Node right) {
         checkExpression(left);
         
-        return new OrNode(union(left, right), left, right);
+        if (left == null && right == null) return new AndNode(position, makeNullNil(left), makeNullNil(right));
+        
+        return new AndNode(union(left, right), makeNullNil(left), makeNullNil(right));
+    }
+
+    public OrNode newOrNode(ISourcePosition position, Node left, Node right) {
+        checkExpression(left);
+
+        if (left == null && right == null) return new OrNode(position, makeNullNil(left), makeNullNil(right));
+        
+        return new OrNode(union(left, right), makeNullNil(left), makeNullNil(right));
     }
 
     public Node getReturnArgsNode(Node node) {
@@ -727,37 +794,164 @@ public class ParserSupport {
         }
         return node;
     }
-
-    public Node new_call(Node receiver, Token name, Node args, Node iter) {
-        if (args == null) {
-            return new CallNode(union(receiver, name), receiver,(String) name.getValue(), null, iter);
+    
+    public Node new_attrassign(ISourcePosition position, Node receiver, String name, Node args) {
+        if (!(args instanceof ArrayNode)) return new AttrAssignNode(position, receiver, name, args);
+        
+        ArrayNode argsNode = (ArrayNode) args;
+        
+        switch (argsNode.size()) {
+            case 1:
+                return new AttrAssignOneArgNode(position, receiver, name, argsNode);
+            case 2:
+                return new AttrAssignTwoArgNode(position, receiver, name, argsNode);
+            case 3:
+                return new AttrAssignThreeArgNode(position, receiver, name, argsNode);
+            default:
+                return new AttrAssignNode(position, receiver, name, argsNode);
         }
-
+    }
+    
+    private Node new_call_noargs(Node receiver, Token name, IterNode iter) {
+        ISourcePosition position = union(receiver, name);
+        
+        if (receiver == null) receiver = NilImplicitNode.NIL;
+        
+        if (iter != null) return new CallNoArgBlockNode(position, receiver, (String) name.getValue(), iter);
+        
+        return new CallNoArgNode(position, receiver, (String) name.getValue());
+    }
+    
+    private Node new_call_complexargs(Node receiver, Token name, Node args, Node iter) {
         if (args instanceof BlockPassNode) {
             // Block and block pass passed in at same time....uh oh
             if (iter != null) {
                 throw new SyntaxException(PID.BLOCK_ARG_AND_BLOCK_GIVEN, iter.getPosition(), "Both block arg and actual block given.");
             }
-                
-            return new CallNode(union(receiver, args), receiver, (String) name.getValue(), 
-                    ((BlockPassNode) args).getArgsNode(), args);
+
+            return new_call_blockpass(receiver, name, (BlockPassNode) args);
         }
             
-        return new CallNode(union(receiver, args), receiver,(String) name.getValue(), args, iter);
+        return new CallNode(union(receiver, args), receiver,(String) name.getValue(), args, iter);        
+    }
+    
+    private Node new_call_blockpass(Node receiver, Token operation, BlockPassNode blockPass) {
+        ISourcePosition position = union(receiver, blockPass);
+        String name = (String) operation.getValue();
+        Node args = blockPass.getArgsNode();
+        
+        if (args == null) return new CallNoArgBlockPassNode(position, receiver, name, args, blockPass);
+        if (!(args instanceof ArrayNode)) return new CallNode(position, receiver, name, args, blockPass);
+        
+        switch (((ArrayNode) args).size()) {
+            case 0:  // foo()
+                return new CallNoArgBlockPassNode(position, receiver, name, args, blockPass);
+            case 1:
+                return new CallOneArgBlockPassNode(position, receiver, name, (ArrayNode) args, blockPass);
+            case 2:
+                return new CallTwoArgBlockPassNode(position, receiver, name, (ArrayNode) args, blockPass);
+            case 3:
+                return new CallThreeArgBlockPassNode(position, receiver, name, (ArrayNode) args, blockPass);
+            default:
+                return new CallNode(position, receiver, name, args, blockPass);
+        } 
     }
 
-    public Node new_fcall(Token operation, Node args, Node iter) {
-        String name = (String) operation.getValue();
+    public Node new_call(Node receiver, Token name, Node argsNode, Node iter) {
+        if (argsNode == null) return new_call_noargs(receiver, name, (IterNode) iter);
+        if (!(argsNode instanceof ArrayNode)) return new_call_complexargs(receiver, name, argsNode, iter);
         
-        if (args == null) return new FCallNode(operation.getPosition(), name, args, iter);
+        ArrayNode args = (ArrayNode) argsNode;
+
+        switch (args.size()) {
+            case 0:
+                if (iter != null) return new CallNoArgBlockNode(union(receiver, args), receiver, (String) name.getValue(), args, (IterNode) iter);
+                    
+                return new CallNoArgNode(union(receiver, args), receiver, args, (String) name.getValue());
+            case 1:
+                if (iter != null) return new CallOneArgBlockNode(union(receiver, args), receiver, (String) name.getValue(), args, (IterNode) iter);
+                
+                return new CallOneArgNode(union(receiver, args), receiver, (String) name.getValue(), args);
+            case 2:
+                if (iter != null) return new CallTwoArgBlockNode(union(receiver, args), receiver, (String) name.getValue(), args, (IterNode) iter);
+                
+                return new CallTwoArgNode(union(receiver, args), receiver, (String) name.getValue(), args);
+            case 3:
+                if (iter != null) return new CallThreeArgBlockNode(union(receiver, args), receiver, (String) name.getValue(), args, (IterNode) iter);
+                
+                return new CallThreeArgNode(union(receiver, args), receiver, (String) name.getValue(), args);
+            default:
+                return new CallNode(union(receiver, args), receiver, (String) name.getValue(), args, iter);
+        }
+    }
+    
+    private Node new_fcall_noargs(Token operation, IterNode iter) {
+        if (iter != null) return new FCallNoArgBlockNode(operation.getPosition(), (String) operation.getValue(), iter);
+        return new FCallNoArgNode(operation.getPosition(), (String) operation.getValue());
+    }
+    
+    private Node new_fcall_simpleargs(Token operation, ArrayNode args, Node iter) {
+        String name = (String) operation.getValue();
+            
+        switch (args.size()) {
+            case 0:  // foo()
+                if (iter != null) return new FCallNoArgBlockNode(union(operation, args), name, args, (IterNode) iter);
+                    
+                return new FCallNoArgNode(union(operation, args), args, name);
+            case 1:
+                if (iter != null) return new FCallOneArgBlockNode(union(operation, args), name, args, (IterNode) iter);
+                
+                return new FCallOneArgNode(union(operation, args), name, args);
+            case 2:
+                if (iter != null) return new FCallTwoArgBlockNode(union(operation, args), name, args, (IterNode) iter);
+                
+                return new FCallTwoArgNode(union(operation, args), name, args);
+            case 3:
+                if (iter != null) return new FCallThreeArgBlockNode(union(operation, args), name, args, (IterNode) iter);
+                
+                return new FCallThreeArgNode(union(operation, args), name, args);
+            default:
+                return new FCallNode(union(operation, args), name, args, iter);
+        }
+    }
+    
+    private Node new_fcall_blockpass(Token operation, BlockPassNode blockPass) {
+        ISourcePosition position = union(operation, blockPass);
+        String name = (String) operation.getValue();
+        Node args = blockPass.getArgsNode();
+        
+        if (args == null) return new FCallNoArgBlockPassNode(position, name, args, blockPass);
+        if (!(args instanceof ArrayNode)) return new FCallNode(position, name, args, blockPass);
+        
+        switch (((ArrayNode) args).size()) {
+            case 0:  // foo()
+                return new FCallNoArgBlockPassNode(position, name, args, blockPass);
+            case 1:
+                return new FCallOneArgBlockPassNode(position, name, (ArrayNode) args, blockPass);
+            case 2:
+                return new FCallTwoArgBlockPassNode(position, name, (ArrayNode) args, blockPass);
+            case 3:
+                return new FCallThreeArgBlockPassNode(position, name, (ArrayNode) args, blockPass);
+            default:
+                return new FCallNode(position, name, args, blockPass);
+        }        
+    }
+    
+    public Node new_fcall(Token operation, Node args, Node iter) {
+        if (args == null) return new_fcall_noargs(operation, (IterNode) iter);
+        if (args instanceof ArrayNode) return new_fcall_simpleargs(operation, (ArrayNode) args, iter);
+
+        String name = (String) operation.getValue();
 
         if (args instanceof BlockPassNode) {
             if (iter != null) {
                 throw new SyntaxException(PID.BLOCK_ARG_AND_BLOCK_GIVEN, iter.getPosition(), "Both block arg and actual block given.");
             }
-            return new FCallNode(union(operation, args), name, ((BlockPassNode) args).getArgsNode(), args);
+
+            return new_fcall_blockpass(operation, (BlockPassNode) args);
         }
         
+
         return new FCallNode(union(operation, args), name, args, iter);
     }
 
@@ -910,7 +1104,7 @@ public class ParserSupport {
         } else {
             state = false;
         }
-
+        
         return new YieldNode(position, node, state);
     }
     
