@@ -54,12 +54,18 @@ public abstract class CallSite {
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2);
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3);
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject[] args);
-    // with block
+    // with block pass
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, Block block);
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, Block block);
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, Block block);
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3, Block block);
     public abstract IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject[] args, Block block);
+    // with block literal (iter)
+    public abstract IRubyObject callIter(ThreadContext context, IRubyObject self, Block block);
+    public abstract IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject arg1, Block block);
+    public abstract IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, Block block);
+    public abstract IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3, Block block);
+    public abstract IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject[] args, Block block);
 
     public static class InlineCachingCallSite extends CallSite implements CacheMap.CacheSite {
         private static class CacheEntry {
@@ -84,167 +90,181 @@ public abstract class CallSite {
         protected IRubyObject cacheAndCall(RubyClass selfType, Block block, IRubyObject[] args, ThreadContext context, IRubyObject self) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, args, context.getFrameSelf(), callType, block);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, args, block);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, args, block);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, args, block);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, IRubyObject[] args, ThreadContext context, IRubyObject self) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, args, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+            if (methodMissing(method, context)) {
+                return callMethodMising(context, self, method, args);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, args);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, args);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, ThreadContext context, IRubyObject self) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, IRubyObject.NULL_ARRAY, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, Block block, ThreadContext context, IRubyObject self) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, IRubyObject.NULL_ARRAY, context.getFrameSelf(), callType, block);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, block);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, block);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, block);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, ThreadContext context, IRubyObject self, IRubyObject arg) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[] {arg}, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, arg);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, arg);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, arg);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, Block block, ThreadContext context, IRubyObject self, IRubyObject arg) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[] {arg}, context.getFrameSelf(), callType, block);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, arg, block);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, arg, block);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, arg, block);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[] {arg1,arg2}, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, arg1, arg2);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, arg1, arg2);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, arg1, arg2);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, Block block, ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[] {arg1,arg2}, context.getFrameSelf(), callType, block);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, arg1, arg2, block);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, arg1, arg2, block);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, arg1, arg2, block);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[] {arg1,arg2,arg3}, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, arg1, arg2,arg3);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, arg1, arg2, arg3);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, arg1, arg2, arg3);
         }
 
         protected IRubyObject cacheAndCall(RubyClass selfType, Block block, ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3) {
             DynamicMethod method = selfType.searchMethod(methodName);
 
-            if (method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
-                return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[] {arg1,arg2,arg3}, context.getFrameSelf(), callType, block);
+            if (methodMissing(method, context)) {
+                return callMethodMissing(context, self, method, arg1, arg2, arg3, block);
             }
+            
+            updateCacheEntry(method, selfType);
 
-            IRubyObject result = method.call(context, self, selfType, methodName, arg1, arg2, arg3, block);
-            
-            if (misses < MAX_MISSES) {
-                updateCacheEntry(method, selfType);
-            }
-            
-            return result;
+            return method.call(context, self, selfType, methodName, arg1, arg2, arg3, block);
+        }
+
+        private IRubyObject callMethodMising(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject[] args) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, args, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, IRubyObject.NULL_ARRAY, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, Block block) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, IRubyObject.NULL_ARRAY, context.getFrameSelf(), callType, block);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject arg) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[]{arg}, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject[] args, Block block) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, args, context.getFrameSelf(), callType, block);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject arg, Block block) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[]{arg}, context.getFrameSelf(), callType, block);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject arg1, IRubyObject arg2) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[]{arg1, arg2}, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject arg1, IRubyObject arg2, Block block) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[]{arg1, arg2}, context.getFrameSelf(), callType, block);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[]{arg1, arg2, arg3}, context.getFrameSelf(), callType, Block.NULL_BLOCK);
+        }
+
+        private IRubyObject callMethodMissing(ThreadContext context, IRubyObject self, DynamicMethod method, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3, Block block) {
+            return RuntimeHelpers.callMethodMissing(context, self, method, methodName, new IRubyObject[]{arg1, arg2, arg3}, context.getFrameSelf(), callType, block);
+        }
+
+        private boolean methodMissing(DynamicMethod method, ThreadContext context) {
+            return method.isUndefined() || (!methodName.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType));
+        }
+
+        private RubyClass pollAndGetClass(ThreadContext context, IRubyObject self) {
+            context.callThreadPoll();
+
+            RubyClass selfType = self.getMetaClass();
+
+            return selfType;
         }
         
         private void updateCacheEntry(DynamicMethod method, RubyClass selfType) {
-            misses++;
-            cache = new CacheEntry(method, selfType);
-            selfType.getRuntime().getCacheMap().add(method, this);
+            if (misses < MAX_MISSES) {
+                misses++;
+                cache = new CacheEntry(method, selfType);
+                selfType.getRuntime().getCacheMap().add(method, this);
+            }
         }
         
         public void removeCachedMethod() {
@@ -252,9 +272,7 @@ public abstract class CallSite {
         }
         
         public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject[] args) {
-            context.callThreadPoll();
-
-            RubyClass selfType = self.getMetaClass();
+            RubyClass selfType = pollAndGetClass(context, self);
             
             CacheEntry myCache = cache;
             if (myCache.cachedType == selfType) {
@@ -265,11 +283,9 @@ public abstract class CallSite {
         }
         
         public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject[] args, Block block) {
-            context.callThreadPoll();
+            RubyClass selfType = pollAndGetClass(context, self);
 
             try {
-                RubyClass selfType = self.getMetaClass();
-
                 CacheEntry myCache = cache;
                 if (myCache.cachedType == selfType) {
                     return myCache.cachedMethod.call(context, self, selfType, methodName, args, block);
@@ -277,7 +293,7 @@ public abstract class CallSite {
 
                 return cacheAndCall(selfType, block, args, context, self);
             } catch (JumpException.BreakJump bj) {
-                return handleBreakJump(bj, block);
+                return handleBreakJump(context, bj);
             } catch (JumpException.RetryJump rj) {
                 throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
             } catch (StackOverflowError soe) {
@@ -285,10 +301,29 @@ public abstract class CallSite {
             }
         }
         
-        public IRubyObject call(ThreadContext context, IRubyObject self) {
-            context.callThreadPoll();
+        public IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject[] args, Block block) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
-            RubyClass selfType = self.getMetaClass();
+            try {
+                CacheEntry myCache = cache;
+                if (myCache.cachedType == selfType) {
+                    return myCache.cachedMethod.call(context, self, selfType, methodName, args, block);
+                }
+
+                return cacheAndCall(selfType, block, args, context, self);
+            } catch (JumpException.BreakJump bj) {
+                return handleBreakJump(context, bj);
+            } catch (JumpException.RetryJump rj) {
+                throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
+            } catch (StackOverflowError soe) {
+                throw context.getRuntime().newSystemStackError("stack level too deep");
+            } finally {
+                block.escape();
+            }
+        }
+        
+        public IRubyObject call(ThreadContext context, IRubyObject self) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
             CacheEntry myCache = cache;
             if (myCache.cachedType == selfType) {
@@ -299,11 +334,9 @@ public abstract class CallSite {
         }
         
         public IRubyObject call(ThreadContext context, IRubyObject self, Block block) {
-            context.callThreadPoll();
+            RubyClass selfType = pollAndGetClass(context, self);
 
             try {
-                RubyClass selfType = self.getMetaClass();
-
                 CacheEntry myCache = cache;
                 if (myCache.cachedType == selfType) {
                     return myCache.cachedMethod.call(context, self, selfType, methodName, block);
@@ -311,7 +344,7 @@ public abstract class CallSite {
 
                 return cacheAndCall(selfType, block, context, self);
             } catch (JumpException.BreakJump bj) {
-                return handleBreakJump(bj, block);
+                return handleBreakJump(context, bj);
             } catch (JumpException.RetryJump rj) {
                 throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
             } catch (StackOverflowError soe) {
@@ -319,10 +352,29 @@ public abstract class CallSite {
             }
         }
         
-        public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1) {
-            context.callThreadPoll();
+        public IRubyObject callIter(ThreadContext context, IRubyObject self, Block block) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
-            RubyClass selfType = self.getMetaClass();
+            try {
+                CacheEntry myCache = cache;
+                if (myCache.cachedType == selfType) {
+                    return myCache.cachedMethod.call(context, self, selfType, methodName, block);
+                }
+
+                return cacheAndCall(selfType, block, context, self);
+            } catch (JumpException.BreakJump bj) {
+                return handleBreakJump(context, bj);
+            } catch (JumpException.RetryJump rj) {
+                throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
+            } catch (StackOverflowError soe) {
+                throw context.getRuntime().newSystemStackError("stack level too deep");
+            } finally {
+                block.escape();
+            }
+        }
+        
+        public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
             CacheEntry myCache = cache;
             if (myCache.cachedType == selfType) {
@@ -333,11 +385,9 @@ public abstract class CallSite {
         }
         
         public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, Block block) {
-            context.callThreadPoll();
+            RubyClass selfType = pollAndGetClass(context, self);
 
             try {
-                RubyClass selfType = self.getMetaClass();
-
                 CacheEntry myCache = cache;
                 if (myCache.cachedType == selfType) {
                     return myCache.cachedMethod.call(context, self, selfType, methodName, arg1, block);
@@ -345,7 +395,7 @@ public abstract class CallSite {
 
                 return cacheAndCall(selfType, block, context, self, arg1);
             } catch (JumpException.BreakJump bj) {
-                return handleBreakJump(bj, block);
+                return handleBreakJump(context, bj);
             } catch (JumpException.RetryJump rj) {
                 throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
             } catch (StackOverflowError soe) {
@@ -353,10 +403,29 @@ public abstract class CallSite {
             }
         }
         
-        public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2) {
-            context.callThreadPoll();
+        public IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject arg1, Block block) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
-            RubyClass selfType = self.getMetaClass();
+            try {
+                CacheEntry myCache = cache;
+                if (myCache.cachedType == selfType) {
+                    return myCache.cachedMethod.call(context, self, selfType, methodName, arg1, block);
+                }
+
+                return cacheAndCall(selfType, block, context, self, arg1);
+            } catch (JumpException.BreakJump bj) {
+                return handleBreakJump(context, bj);
+            } catch (JumpException.RetryJump rj) {
+                throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
+            } catch (StackOverflowError soe) {
+                throw context.getRuntime().newSystemStackError("stack level too deep");
+            } finally {
+                block.escape();
+            }
+        }
+        
+        public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
             CacheEntry myCache = cache;
             if (myCache.cachedType == selfType) {
@@ -367,11 +436,9 @@ public abstract class CallSite {
         }
         
         public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, Block block) {
-            context.callThreadPoll();
+            RubyClass selfType = pollAndGetClass(context, self);
 
             try {
-                RubyClass selfType = self.getMetaClass();
-
                 CacheEntry myCache = cache;
                 if (myCache.cachedType == selfType) {
                     return myCache.cachedMethod.call(context, self, selfType, methodName, arg1, arg2, block);
@@ -379,7 +446,7 @@ public abstract class CallSite {
 
                 return cacheAndCall(selfType, block, context, self, arg1, arg2);
             } catch (JumpException.BreakJump bj) {
-                return handleBreakJump(bj, block);
+                return handleBreakJump(context, bj);
             } catch (JumpException.RetryJump rj) {
                 throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
             } catch (StackOverflowError soe) {
@@ -387,10 +454,29 @@ public abstract class CallSite {
             }
         }
         
-        public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3) {
-            context.callThreadPoll();
+        public IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, Block block) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
-            RubyClass selfType = self.getMetaClass();
+            try {
+                CacheEntry myCache = cache;
+                if (myCache.cachedType == selfType) {
+                    return myCache.cachedMethod.call(context, self, selfType, methodName, arg1, arg2, block);
+                }
+
+                return cacheAndCall(selfType, block, context, self, arg1, arg2);
+            } catch (JumpException.BreakJump bj) {
+                return handleBreakJump(context, bj);
+            } catch (JumpException.RetryJump rj) {
+                throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
+            } catch (StackOverflowError soe) {
+                throw context.getRuntime().newSystemStackError("stack level too deep");
+            } finally {
+                block.escape();
+            }
+        }
+        
+        public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
             CacheEntry myCache = cache;
             if (myCache.cachedType == selfType) {
@@ -401,11 +487,9 @@ public abstract class CallSite {
         }
         
         public IRubyObject call(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3, Block block) {
-            context.callThreadPoll();
+            RubyClass selfType = pollAndGetClass(context, self);
 
             try {
-                RubyClass selfType = self.getMetaClass();
-
                 CacheEntry myCache = cache;
                 if (myCache.cachedType == selfType) {
                     return myCache.cachedMethod.call(context, self, selfType, methodName, arg1, arg2, arg3, block);
@@ -413,17 +497,38 @@ public abstract class CallSite {
 
                 return cacheAndCall(selfType, block, context, self, arg1, arg2, arg3);
             } catch (JumpException.BreakJump bj) {
-                return handleBreakJump(bj, block);
+                return handleBreakJump(context, bj);
             } catch (JumpException.RetryJump rj) {
                 throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
             } catch (StackOverflowError soe) {
                 throw context.getRuntime().newSystemStackError("stack level too deep");
             }
         }
+        
+        public IRubyObject callIter(ThreadContext context, IRubyObject self, IRubyObject arg1, IRubyObject arg2, IRubyObject arg3, Block block) {
+            RubyClass selfType = pollAndGetClass(context, self);
 
-        private IRubyObject handleBreakJump(BreakJump bj, Block block) throws BreakJump {
+            try {
+                CacheEntry myCache = cache;
+                if (myCache.cachedType == selfType) {
+                    return myCache.cachedMethod.call(context, self, selfType, methodName, arg1, arg2, arg3, block);
+                }
+
+                return cacheAndCall(selfType, block, context, self, arg1, arg2, arg3);
+            } catch (JumpException.BreakJump bj) {
+                return handleBreakJump(context, bj);
+            } catch (JumpException.RetryJump rj) {
+                throw context.getRuntime().newLocalJumpError("retry", context.getRuntime().getNil(), "retry outside of rescue not yet supported");
+            } catch (StackOverflowError soe) {
+                throw context.getRuntime().newSystemStackError("stack level too deep");
+            } finally {
+                block.escape();
+            }
+        }
+
+        private IRubyObject handleBreakJump(ThreadContext context, BreakJump bj) throws BreakJump {
             // consume and rethrow or just keep rethrowing?
-            if (block.getBody() == bj.getTarget()) {
+            if (context.getFrameJumpTarget() == bj.getTarget()) {
                 return (IRubyObject) bj.getValue();
             }
             throw bj;
