@@ -197,35 +197,35 @@ public final class ThreadContext {
         return scopeStack[scopeIndex - 1];
     }
     
-    private void expandFramesIfNecessary(int newMax) {
-        if (newMax == frameStack.length) {
-            int newSize = frameStack.length * 2;
-            Frame[] newFrameStack = new Frame[newSize];
-            
-            System.arraycopy(frameStack, 0, newFrameStack, 0, frameStack.length);
-            
-            for (int i = frameStack.length; i < newSize; i++) {
-                newFrameStack[i] = new Frame();
-            }
-            
-            frameStack = newFrameStack;
+    private void expandFramesIfNecessary() {
+        int newSize = frameStack.length * 2;
+        frameStack = fillNewFrameStack(new Frame[newSize], newSize);
+    }
+
+    private Frame[] fillNewFrameStack(Frame[] newFrameStack, int newSize) {
+        System.arraycopy(frameStack, 0, newFrameStack, 0, frameStack.length);
+
+        for (int i = frameStack.length; i < newSize; i++) {
+            newFrameStack[i] = new Frame();
         }
+        
+        return newFrameStack;
     }
     
     private void expandParentsIfNecessary() {
-        if (parentIndex + 1 == parentStack.length) {
-            int newSize = parentStack.length * 2;
-            RubyModule[] newParentStack = new RubyModule[newSize];
-            
-            System.arraycopy(parentStack, 0, newParentStack, 0, parentStack.length);
-            
-            parentStack = newParentStack;
-        }
+        int newSize = parentStack.length * 2;
+        RubyModule[] newParentStack = new RubyModule[newSize];
+
+        System.arraycopy(parentStack, 0, newParentStack, 0, parentStack.length);
+
+        parentStack = newParentStack;
     }
     
     public void pushScope(DynamicScope scope) {
         scopeStack[++scopeIndex] = scope;
-        expandScopesIfNecessary();
+        if (scopeIndex + 1 == scopeStack.length) {
+            expandScopesIfNecessary();
+        }
     }
     
     public void popScope() {
@@ -233,14 +233,12 @@ public final class ThreadContext {
     }
     
     private void expandScopesIfNecessary() {
-        if (scopeIndex + 1 == scopeStack.length) {
-            int newSize = scopeStack.length * 2;
-            DynamicScope[] newScopeStack = new DynamicScope[newSize];
-            
-            System.arraycopy(scopeStack, 0, newScopeStack, 0, scopeStack.length);
-            
-            scopeStack = newScopeStack;
-        }
+        int newSize = scopeStack.length * 2;
+        DynamicScope[] newScopeStack = new DynamicScope[newSize];
+
+        System.arraycopy(scopeStack, 0, newScopeStack, 0, scopeStack.length);
+
+        scopeStack = newScopeStack;
     }
     
     public RubyThread getThread() {
@@ -272,18 +270,18 @@ public final class ThreadContext {
     
     //////////////////// CATCH MANAGEMENT ////////////////////////
     private void expandCatchIfNecessary() {
-        if (catchIndex + 1 == catchStack.length) {
-            int newSize = catchStack.length * 2;
-            CatchTarget[] newCatchStack = new CatchTarget[newSize];
-            
-            System.arraycopy(catchStack, 0, newCatchStack, 0, catchStack.length);
-            catchStack = newCatchStack;
-        }
+        int newSize = catchStack.length * 2;
+        CatchTarget[] newCatchStack = new CatchTarget[newSize];
+
+        System.arraycopy(catchStack, 0, newCatchStack, 0, catchStack.length);
+        catchStack = newCatchStack;
     }
     
     public void pushCatch(CatchTarget catchTarget) {
         catchStack[++catchIndex] = catchTarget;
-        expandCatchIfNecessary();
+        if (catchIndex + 1 == catchStack.length) {
+            expandCatchIfNecessary();
+        }
     }
     
     public void popCatch() {
@@ -302,24 +300,33 @@ public final class ThreadContext {
     private void pushFrameCopy() {
         Frame currentFrame = getCurrentFrame();
         frameStack[++frameIndex].updateFrame(currentFrame);
-        expandFramesIfNecessary(frameIndex + 1);
+        if (frameIndex + 1 == frameStack.length) {
+            expandFramesIfNecessary();
+        }
     }
     
     private Frame pushFrameCopy(Frame frame) {
         frameStack[++frameIndex].updateFrame(frame);
-        expandFramesIfNecessary(frameIndex + 1);
+        if (frameIndex + 1 == frameStack.length) {
+            expandFramesIfNecessary();
+        }
         return frameStack[frameIndex];
     }
     
     private Frame pushFrame(Frame frame) {
         frameStack[++frameIndex] = frame;
-        expandFramesIfNecessary(frameIndex + 1);
+        if (frameIndex + 1 == frameStack.length) {
+            expandFramesIfNecessary();
+        }
         return frame;
     }
     
     private void pushCallFrame(RubyModule clazz, String name, 
                                IRubyObject self, Block block, JumpTarget jumpTarget) {
-        pushFrame(clazz, name, self, block, jumpTarget);        
+        frameStack[++frameIndex].updateFrame(clazz, self, name, block, file, line, jumpTarget);
+        if (frameIndex + 1 == frameStack.length) {
+            expandFramesIfNecessary();
+        }
     }
     
     private void pushBacktraceFrame(String name) {
@@ -328,18 +335,16 @@ public final class ThreadContext {
     
     private void pushFrame(String name) {
         frameStack[++frameIndex].updateFrame(name, file, line);
-        expandFramesIfNecessary(frameIndex + 1);
-    }
-
-    private void pushFrame(RubyModule clazz, String name, 
-                               IRubyObject self, Block block, JumpTarget jumpTarget) {
-        frameStack[++frameIndex].updateFrame(clazz, self, name, block, file, line, jumpTarget);
-        expandFramesIfNecessary(frameIndex + 1);
+        if (frameIndex + 1 == frameStack.length) {
+            expandFramesIfNecessary();
+        }
     }
     
     private void pushFrame() {
         frameStack[++frameIndex].updateFrame(file, line);
-        expandFramesIfNecessary(frameIndex + 1);
+        if (frameIndex + 1 == frameStack.length) {
+            expandFramesIfNecessary();
+        }
     }
     
     private void popFrame() {
@@ -347,11 +352,13 @@ public final class ThreadContext {
         frameIndex--;
         setFile(frame.getFile());
         setLine(frame.getLine());
+        
+        frame.clear();
     }
         
-    private void popFrameReal() {
+    private void popFrameReal(Frame oldFrame) {
         Frame frame = frameStack[frameIndex];
-        frameStack[frameIndex] = new Frame();
+        frameStack[frameIndex] = oldFrame;
         frameIndex--;
         setFile(frame.getFile());
         setLine(frame.getLine());
@@ -374,13 +381,14 @@ public final class ThreadContext {
     }
     
     public Frame getNextFrame() {
-        expandFramesIfNecessary(frameIndex + 1);
+        if (frameIndex + 1 == frameStack.length) {
+            expandFramesIfNecessary();
+        }
         return frameStack[frameIndex + 1];
     }
     
     public Frame getPreviousFrame() {
-        int size = frameIndex + 1;
-        return size <= 1 ? null : frameStack[size - 2];
+        return frameIndex < 1 ? null : frameStack[frameIndex - 1];
     }
     
     public int getFrameCount() {
@@ -399,8 +407,8 @@ public final class ThreadContext {
         return getCurrentFrame().getJumpTarget();
     }
     
+    @Deprecated
     public void setFrameJumpTarget(JumpTarget target) {
-        getCurrentFrame().setJumpTarget(target);
     }
     
     public RubyModule getFrameKlazz() {
@@ -454,13 +462,19 @@ public final class ThreadContext {
         if ((calls++ & 0xFF) == 0) pollThreadEvents();
     }
     
+    public void trace(int event, String name, RubyModule implClass) {
+        runtime.callEventHooks(this, event, file, line, name, implClass);
+    }
+    
     public void pushRubyClass(RubyModule currentModule) {
         // FIXME: this seems like a good assertion, but it breaks compiled code and the code seems
         // to run without it...
         //assert currentModule != null : "Can't push null RubyClass";
         
         parentStack[++parentIndex] = currentModule;
-        expandParentsIfNecessary();
+        if (parentIndex + 1 == parentStack.length) {
+            expandParentsIfNecessary();
+        }
     }
     
     public RubyModule popRubyClass() {
@@ -507,28 +521,7 @@ public final class ThreadContext {
      * Used by the evaluator and the compiler to look up a constant by name
      */
     public IRubyObject getConstant(String internedName) {
-        StaticScope scope = getCurrentScope().getStaticScope();
-        RubyClass object = runtime.getObject();
-        IRubyObject result;
-        
-        // flipped from while to do to search current class first
-        do {
-            RubyModule klass = scope.getModule();
-            
-            // Not sure how this can happen
-            //if (NIL_P(klass)) return rb_const_get(CLASS_OF(self), id);
-            if ((result = klass.fastFetchConstant(internedName)) != null) {
-                if (result != RubyObject.UNDEF) {
-                    return result;
-                }
-                klass.deleteConstant(internedName);
-                if (runtime.getLoadService().autoload(klass.getName() + "::" + internedName) == null) break;
-                continue;
-            }
-            scope = scope.getPreviousCRefScope();
-        } while (scope != null && scope.getModule() != object);
-        
-        return getCurrentScope().getStaticScope().getModule().fastGetConstant(internedName);
+        return getCurrentScope().getStaticScope().getConstant(runtime, internedName, runtime.getObject());
     }
     
     /**
@@ -580,7 +573,7 @@ public final class ThreadContext {
             return;
         }
         
-        StringBuffer buf = new StringBuffer(60);
+        StringBuilder buf = new StringBuilder(60);
         buf.append(frame.getFile()).append(':').append(frame.getLine() + 1);
         
         if (previousFrame.getName() != null) {
@@ -598,6 +591,17 @@ public final class ThreadContext {
      * @return an Array with the backtrace
      */
     public static IRubyObject createBacktraceFromFrames(Ruby runtime, Frame[] backtraceFrames) {
+        return createBacktraceFromFrames(runtime, backtraceFrames, true);
+    }
+    
+    /**
+     * Create an Array with backtrace information.
+     * @param runtime
+     * @param level
+     * @param nativeException
+     * @return an Array with the backtrace
+     */
+    public static IRubyObject createBacktraceFromFrames(Ruby runtime, Frame[] backtraceFrames, boolean cropAtEval) {
         RubyArray backtrace = runtime.newArray();
         
         if (backtraceFrames == null || backtraceFrames.length <= 0) return backtrace;
@@ -607,7 +611,7 @@ public final class ThreadContext {
         for (int i = traceSize - 1; i > 0; i--) {
             Frame frame = backtraceFrames[i];
             // We are in eval with binding break out early
-            if (frame.isBindingFrame()) break;
+            if (cropAtEval && frame.isBindingFrame()) break;
 
             addBackTraceElement(backtrace, frame, backtraceFrames[i - 1]);
         }
@@ -772,9 +776,8 @@ public final class ThreadContext {
     }
     
     public void preNodeEval(RubyModule rubyClass, IRubyObject self, String name) {
-        setFile(name);
         pushRubyClass(rubyClass);
-        pushCallFrame(null, name, self, Block.NULL_BLOCK, null);
+        pushCallFrame(null, null, self, Block.NULL_BLOCK, null);
         // set visibility to private, since toplevel of scripts always started out private
         setCurrentVisibility(Visibility.PRIVATE);
     }
@@ -832,7 +835,8 @@ public final class ThreadContext {
         setWithinTrace(false);
     }
     
-    public void preForBlock(Binding binding, RubyModule klass) {
+    public Frame preForBlock(Binding binding, RubyModule klass) {
+        Frame lastFrame = getNextFrame();
         Frame f = binding.getFrame();
         f.setFile(file);
         f.setLine(line);
@@ -840,9 +844,11 @@ public final class ThreadContext {
         getCurrentFrame().setVisibility(binding.getVisibility());
         pushScope(binding.getDynamicScope());
         pushRubyClass((klass != null) ? klass : binding.getKlass());
+        return lastFrame;
     }
     
-    public void preYieldSpecificBlock(Binding binding, StaticScope scope, RubyModule klass) {
+    public Frame preYieldSpecificBlock(Binding binding, StaticScope scope, RubyModule klass) {
+        Frame lastFrame = getNextFrame();
         Frame f = pushFrame(binding.getFrame());
         f.setFile(file);
         f.setLine(line);
@@ -850,9 +856,11 @@ public final class ThreadContext {
         // new scope for this invocation of the block, based on parent scope
         pushScope(DynamicScope.newDynamicScope(scope, binding.getDynamicScope()));
         pushRubyClass((klass != null) ? klass : binding.getKlass());
+        return lastFrame;
     }
     
-    public void preYieldLightBlock(Binding binding, DynamicScope emptyScope, RubyModule klass) {
+    public Frame preYieldLightBlock(Binding binding, DynamicScope emptyScope, RubyModule klass) {
+        Frame lastFrame = getNextFrame();
         Frame f = pushFrame(binding.getFrame());
         f.setFile(file);
         f.setLine(line);
@@ -860,44 +868,49 @@ public final class ThreadContext {
         // just push the same empty scope, since we won't use one
         pushScope(emptyScope);
         pushRubyClass((klass != null) ? klass : binding.getKlass());
+        return lastFrame;
     }
     
-    public void preYieldNoScope(Binding binding, RubyModule klass) {
+    public Frame preYieldNoScope(Binding binding, RubyModule klass) {
+        Frame lastFrame = getNextFrame();
         Frame f = pushFrame(binding.getFrame());
         f.setFile(file);
         f.setLine(line);
         f.setVisibility(binding.getVisibility());
         pushRubyClass((klass != null) ? klass : binding.getKlass());
+        return lastFrame;
     }
     
-    public void preEvalWithBinding(Binding binding) {
+    public Frame preEvalWithBinding(Binding binding) {
+        Frame lastFrame = getNextFrame();
         Frame frame = binding.getFrame();
         frame.setIsBindingFrame(true);
         pushFrame(frame);
         getCurrentFrame().setVisibility(binding.getVisibility());
         pushRubyClass(binding.getKlass());
+        return lastFrame;
     }
     
-    public void postEvalWithBinding(Binding binding) {
+    public void postEvalWithBinding(Binding binding, Frame lastFrame) {
         binding.getFrame().setIsBindingFrame(false);
-        popFrameReal();
+        popFrameReal(lastFrame);
         popRubyClass();
     }
     
-    public void postYield(Binding binding) {
+    public void postYield(Binding binding, Frame lastFrame) {
         popScope();
-        popFrameReal();
+        popFrameReal(lastFrame);
         popRubyClass();
     }
     
-    public void postYieldLight(Binding binding) {
+    public void postYieldLight(Binding binding, Frame lastFrame) {
         popScope();
-        popFrameReal();
+        popFrameReal(lastFrame);
         popRubyClass();
     }
     
-    public void postYieldNoScope() {
-        popFrameReal();
+    public void postYieldNoScope(Frame lastFrame) {
+        popFrameReal(lastFrame);
         popRubyClass();
     }
     
