@@ -46,6 +46,7 @@ import java.util.jar.JarFile;
 import java.util.regex.Matcher;
 import java.util.regex.Pattern;
 
+import java.util.zip.ZipException;
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
 import org.jruby.RubyFile;
@@ -189,7 +190,7 @@ public class LoadService {
         if (path == null || path.length() == 0) return;
         
         synchronized(loadPath) {
-            loadPath.add(runtime.newString(path.replace('\\', '/')));
+            loadPath.append(runtime.newString(path.replace('\\', '/')));
         }
     }
 
@@ -281,8 +282,10 @@ public class LoadService {
         Script script = null;
         if (library == null) {
             String className = file;
-            if (file.lastIndexOf(".") != -1) className = file.substring(0, file.lastIndexOf("."));
-            className = className.replace('-', '_').replace('.', '_');
+            // Remove any relative prefix, e.g. "./foo/bar" becomes "foo/bar".
+            className = className.replaceFirst("^\\.\\/", "");
+            if (className.lastIndexOf(".") != -1) className = className.substring(0, className.lastIndexOf("."));
+            className = className.replace("-", "_minus_").replace('.', '_');
             int lastSlashIndex = className.lastIndexOf('/');
             if (lastSlashIndex > -1 &&
                     lastSlashIndex < className.length() - 1 &&
@@ -453,6 +456,11 @@ public class LoadService {
                             current = new JarFile(entry.substring(5,entry.indexOf("!/")));
                         }
                         jarFiles.put(entry,current);
+                    } catch (ZipException ignored) {
+                        if (runtime.getInstanceConfig().isVerbose()) {
+                            runtime.getErr().println("ZipException trying to access " + entry + ", stack trace follows:");
+                            ignored.printStackTrace(runtime.getErr());
+                        }
                     } catch (FileNotFoundException ignored) {
                     } catch (IOException e) {
                         throw runtime.newIOErrorFromException(e);
@@ -486,7 +494,10 @@ public class LoadService {
                             RubyFile.expandUserPath(runtime.getCurrentContext(), name));
                     if (current.isFile()) {
                         try {
-                            return new LoadServiceResource(current.toURI().toURL(), current.getPath());
+                            // relative paths without ./ on front get absolute path
+                            String resourcePath = name.startsWith("./") ? name : current.getPath();
+                            
+                            return new LoadServiceResource(current.toURI().toURL(), resourcePath);
                         } catch (MalformedURLException e) {
                             throw runtime.newIOErrorFromException(e);
                         }
@@ -569,7 +580,7 @@ public class LoadService {
             // Create package name, by splitting on / and joining all but the last elements with a ".", and downcasing them.
             String[] all = file.split("/");
 
-            StringBuffer finName = new StringBuffer();
+            StringBuilder finName = new StringBuilder();
             for(int i=0, j=(all.length-1); i<j; i++) {
                 finName.append(all[i].toLowerCase()).append(".");
                 
