@@ -62,17 +62,41 @@ import org.jruby.util.JRubyFile;
  */
 public class ChannelStream implements Stream, Finalizable {
     private final static boolean DEBUG = false;
-    private final static int BUFSIZE = 16 * 1024;
+    
+    /**
+     * The size of the read/write buffer allocated for this stream.
+     * 
+     * This size has been scaled back from its original 16k because although
+     * the larger buffer size results in raw File.open times being rather slow
+     * (due to the cost of instantiating a relatively large buffer). We should
+     * try to find a happy medium, or potentially pool buffers, or perhaps even
+     * choose a value based on platform(??), but for now I am reducing it along
+     * with changes for the "large read" patch from JRUBY-2657.
+     */
+    private final static int BUFSIZE = 4 * 1024;
+    
+    /**
+     * The size at which a single read should turn into a chunkier bulk read.
+     * Currently, this size is about 4x a normal buffer size.
+     * 
+     * This size was not really arrived at experimentally, and could potentially
+     * be increased. However, it seems like a "good size" and we should
+     * probably only adjust it if it turns out we would perform better with a
+     * larger buffer for large bulk reads.
+     */
+    private final static int BULK_READ_SIZE = 16 * 1024;
+    private final static ByteBuffer EMPTY_BUFFER = ByteBuffer.allocate(0);
     
     private Ruby runtime;
     protected ModeFlags modes;
     protected boolean sync = false;
     
-    protected ByteBuffer buffer; // r/w buffer
+    protected volatile ByteBuffer buffer; // r/w buffer
     protected boolean reading; // are we reading or writing?
     private ChannelDescriptor descriptor;
     private boolean blocking = true;
     protected int ungotc = -1;
+    private volatile boolean closedExplicitly = false;
 
     public ChannelStream(Ruby runtime, ChannelDescriptor descriptor, ModeFlags modes, FileDescriptor fileDescriptor) throws InvalidValueException {
         descriptor.checkNewModes(modes);
@@ -158,7 +182,12 @@ public class ChannelStream implements Stream, Finalizable {
     public boolean writeDataBuffered() {
         return !reading && buffer.position() > 0;
     }
-
+    private final int refillBuffer() throws IOException {
+        buffer.clear();
+        int n = ((ReadableByteChannel) descriptor.getChannel()).read(buffer);
+        buffer.flip();
+        return n;
+    }
     public synchronized ByteList fgets(ByteList separatorString) throws IOException, BadDescriptorException {
         checkReadable();
         ensureRead();
@@ -212,9 +241,7 @@ public class ChannelStream implements Stream, Finalizable {
                 
                 // no match, append remainder of buffer and continue with next block
                 buf.append(bytes, offset, buffer.remaining());
-                buffer.clear();
-                int read = ((ReadableByteChannel)descriptor.getChannel()).read(buffer);
-                buffer.flip();
+                int read = refillBuffer();
                 if (read == -1) break LineLoop;
             }
             
@@ -244,17 +271,84 @@ public class ChannelStream implements Stream, Finalizable {
         return buf;
     }
     
+    /**
+     * An version of read that reads all bytes up to and including a terminator byte.
+     * <p>
+     * If the terminator byte is found, it will be the last byte in the output buffer.
+     * </p>
+     *
+     * @param dst The output buffer.
+     * @param terminator The byte to terminate reading.
+     * @return The number of bytes read, or -1 if EOF is reached.
+     * 
+     * @throws java.io.IOException
+     * @throws org.jruby.util.io.BadDescriptorException
+     */
+    public synchronized int getline(ByteList dst, byte terminator) throws IOException, BadDescriptorException {
+        checkReadable();
+        ensureRead();
+        descriptor.checkOpen();
+        
+        int totalRead = 0;
+        boolean found = false;
+        if (ungotc != -1) {
+            dst.append((byte) ungotc);
+            found = ungotc == terminator;
+            ungotc = -1;
+            ++totalRead;
+        }
+        while (!found) {
+            final byte[] bytes = buffer.array();
+            final int begin = buffer.arrayOffset() + buffer.position();
+            final int end = begin + buffer.remaining();
+            int len = 0;
+            for (int i = begin; i < end && !found; ++i) {
+                found = bytes[i] == terminator;
+                ++len;
+            }
+            if (len > 0) {
+                dst.append(buffer, len);
+                totalRead += len;
+            }
+            if (!found) {
+                int n = refillBuffer();
+                if (n <= 0) {
+                    if (n < 0 && totalRead < 1) {
+                        return -1;
+                    }
+                    break;
+                }
+            }
+        }
+        return totalRead;
+    }
+    
     public synchronized ByteList readall() throws IOException, BadDescriptorException {
         if (descriptor.isSeekable()) {
             invalidateBuffer();
             FileChannel channel = (FileChannel)descriptor.getChannel();
             long left = channel.size() - channel.position();
-            if (left == 0) {
+            if (left <= 0) {
                 eof = true;
                 return null;
             }
-
-            return fread((int) left);
+            left += ungotc != -1 ? 1 : 0;
+            ByteList result = new ByteList((int) left);
+            ByteBuffer buf = ByteBuffer.wrap(result.unsafeBytes(), 
+                    result.begin(), (int) left);
+            if (ungotc != -1) {
+                buf.put((byte) ungotc);
+                ungotc = -1;
+            }
+            while (buf.hasRemaining()) {
+                int n = ((ReadableByteChannel) descriptor.getChannel()).read(buf);
+                if (n <= 0) {
+                    break;
+                }
+            }
+            eof = true;
+            result.length(buf.position());
+            return result;
         } else if (descriptor.isNull()) {
             return new ByteList(0);
         } else {
@@ -285,13 +379,14 @@ public class ChannelStream implements Stream, Finalizable {
      * @see org.jruby.util.IOHandler#close()
      */
     public synchronized void fclose() throws IOException, BadDescriptorException {
-        close(false); // not closing from finalise
+        closedExplicitly = true;
+        close(false); // not closing from finalize
     }
-    
+
     /**
      * Internal close, to safely work for finalizing.
      * @param finalizing true if this is in a finalizing context
-     * @throws IOException 
+     * @throws IOException
      * @throws BadDescriptorException
      */
     private void close(boolean finalizing) throws IOException, BadDescriptorException {
@@ -299,13 +394,15 @@ public class ChannelStream implements Stream, Finalizable {
             flushWrite();
 
             descriptor.close();
+            buffer = EMPTY_BUFFER;
 
-            if (DEBUG) getLogger("ChannelStream").info("Descriptor for fileno " + descriptor.getFileno() + " closed by stream");
+            if (DEBUG) getLogger("ChannelStream").info("Descriptor for fileno "
+                    + descriptor.getFileno() + " closed by stream");
         } finally {
             if (!finalizing) getRuntime().removeInternalFinalizer(this);
         }
     }
-    
+
     /**
      * Internal close, to safely work for finalizing.
      * @param finalizing true if this is in a finalizing context
@@ -429,15 +526,14 @@ public class ChannelStream implements Stream, Finalizable {
     public synchronized long fgetpos() throws IOException, PipeException, InvalidValueException, BadDescriptorException {
         // Correct position for read / write buffering (we could invalidate, but expensive)
         if (descriptor.isSeekable()) {
-            int ungotcCopy = ungotc;
-            lseek(0, SEEK_CUR);
-            ungotc = ungotcCopy;
-
             FileChannel fileChannel = (FileChannel)descriptor.getChannel();
-            if (ungotc != -1 && fileChannel.position() > 0) {
-                return fileChannel.position() - 1;
+            long pos = fileChannel.position();
+            // Adjust for buffered data
+            if (reading) {
+                pos -= buffer.remaining();
+                return pos - (pos > 0 && ungotc != -1 ? 1 : 0);
             } else {
-                return fileChannel.position();
+                return pos + buffer.position();
             }
         } else if (descriptor.isNull()) {
             return 0;
@@ -459,10 +555,10 @@ public class ChannelStream implements Stream, Finalizable {
         if (descriptor.isSeekable()) {
             FileChannel fileChannel = (FileChannel)descriptor.getChannel();
             ungotc = -1;
+            int adj = 0;
             if (reading) {
-                if (buffer.remaining() > 0) {
-                    fileChannel.position(fileChannel.position() - buffer.remaining());
-                }
+                // for SEEK_CUR, need to adjust for buffered data
+                adj = buffer.remaining();
                 buffer.clear();
                 buffer.flip();
             } else {
@@ -474,7 +570,7 @@ public class ChannelStream implements Stream, Finalizable {
                     fileChannel.position(offset);
                     break;
                 case SEEK_CUR:
-                    fileChannel.position(fileChannel.position() + offset);
+                    fileChannel.position(fileChannel.position() - adj + offset);
                     break;
                 case SEEK_END:
                     fileChannel.position(fileChannel.size() + offset);
@@ -562,6 +658,10 @@ public class ChannelStream implements Stream, Finalizable {
         // TODO this should entry into error handling somewhere
         int bytesRead = descriptor.read(number, byteList);
         
+        if (bytesRead == -1) {
+            eof = true;
+        }
+        
         return byteList;
     }
 
@@ -569,23 +669,43 @@ public class ChannelStream implements Stream, Finalizable {
         checkReadable();
         ensureRead();
         
-        ByteList result = new ByteList();
+        ByteList result = new ByteList(0);
+        
         int len = -1;
         if (buffer.hasRemaining()) { // already have some bytes buffered
             len = (number <= buffer.remaining()) ? number : buffer.remaining();
             result.append(buffer, len);
         }
+        boolean done = false;
+        //
+        // Avoid double-copying for reads that are larger than the buffer size
+        //
+        while ((number - result.length()) >= BUFSIZE) {
+            //
+            // limit each iteration to a max of BULK_READ_SIZE to avoid over-size allocations
+            //
+            int bytesToRead = Math.min(BULK_READ_SIZE, number - result.length());
+            int n = descriptor.read(bytesToRead, result);
+            if (n == -1) {
+                eof = true;
+                done = true;
+                break;
+            } else if (n == 0) {
+                done = true;
+                break;
+            }
+        }
         
-        ReadableByteChannel readChannel = (ReadableByteChannel)descriptor.getChannel();
-        int read = BUFSIZE;
-        while (result.length() != number) {
-            // try to read more
-            buffer.clear(); 
-            read = readChannel.read(buffer);
-            buffer.flip();
+        //
+        // Complete the request by filling the read buffer first
+        //
+        while (!done && result.length() != number) {
+            int read = refillBuffer();
             
-            if (read <= 0) {
-                // if we reach EOF or didn't read anything, bail out
+            if (read == -1) {
+                eof = true;
+                break;
+            } else if (read == 0) {
                 break;
             }
             
@@ -595,7 +715,11 @@ public class ChannelStream implements Stream, Finalizable {
             result.append(buffer, len);
         }
         
-        if (result.length() == 0 && number != 0) throw new EOFException();
+        if (result.length() == 0 && number != 0) {
+            if (eof) {
+                throw new EOFException();
+            }
+        }
         return result;
     }
     
@@ -603,13 +727,12 @@ public class ChannelStream implements Stream, Finalizable {
         ensureRead();
         
         if (!buffer.hasRemaining()) {
-            buffer.clear();
-            int read = descriptor.read(buffer);
-            buffer.flip();
-            
-            if (read <= 0) {
+            int len = refillBuffer();
+            if (len == -1) {
                 eof = true;
                 return -1;
+            } else if (len == 0) {
+                return -1;
             }
         }
         return buffer.get() & 0xFF;
@@ -704,15 +827,20 @@ public class ChannelStream implements Stream, Finalizable {
             if (posOverrun != 0) fileChannel.position(fileChannel.position() - posOverrun);
         }
     }
-    
+
     /**
      * Ensure close (especially flush) when we're finished with
      */
-    public void finalize() {
+    @Override
+    public synchronized void finalize() {
+        if (closedExplicitly) return;
+
         // FIXME: I got a bunch of NPEs when I didn't check for nulls here...HOW?!
-        if (descriptor != null && descriptor.isSeekable() && descriptor.isOpen()) closeForFinalize(); // close without removing from finalizers
+        if (descriptor != null && descriptor.isSeekable() && descriptor.isOpen()) {
+            closeForFinalize(); // close without removing from finalizers
+        }
     }
-    
+
     public int ready() throws IOException {
         return newInputStream().available();
     }
@@ -784,11 +912,57 @@ public class ChannelStream implements Stream, Finalizable {
         }
     }
 
-    public synchronized ByteList readpartial(int number) throws IOException, BadDescriptorException, EOFException {
+    public synchronized ByteList readnonblock(int number) throws IOException, BadDescriptorException, EOFException {
+        assert number >= 0;
+
+        if (number == 0) {
+            return null;
+        }
+
         if (descriptor.getChannel() instanceof SelectableChannel) {
+            SelectableChannel selectableChannel = (SelectableChannel)descriptor.getChannel();
+            synchronized (selectableChannel.blockingLock()) {
+                boolean oldBlocking = selectableChannel.isBlocking();
+                try {
+                    selectableChannel.configureBlocking(false);
+                    return readpartial(number);
+                } finally {
+                    selectableChannel.configureBlocking(oldBlocking);
+                }
+            }
+        } else {
+            return null;
+        }
+    }
+
+    public synchronized ByteList readpartial(int number) throws IOException, BadDescriptorException, EOFException {
+        assert number >= 0;
+
+        if (number == 0) {
+            return null;
+        }
+
+        if (descriptor.getChannel() instanceof SelectableChannel) {
+            // make sure that the ungotc is not forgotten
+            if (ungotc >= 0) {
+                number--;
+                if (number == 0 || !buffer.hasRemaining()) {
+                    ByteList result = new ByteList(new byte[] {(byte)ungotc}, false);
+                    ungotc = -1;
+                    return result;
+                }
+            }
+
             if (buffer.hasRemaining()) {
                 // already have some bytes buffered, just return those
-                return bufferedRead(buffer.remaining());
+
+                ByteList result = bufferedRead(Math.min(buffer.remaining(), number));
+
+                if (ungotc >= 0) {
+                    result.prepend((byte)ungotc);
+                    ungotc = -1;
+                }
+                return result;
             } else {
                 // otherwise, we try an unbuffered read to get whatever's available
                 return read(number);
